#!/usr/bin/env python
# encoding: utf-8
'''
PyCoALA -- Python Contact Area Localization and Analysis

PyCoALA is a Python program to identify and visualize contact areas between a receptor and 
a ligand from an ensemble of encounter complexes processed with the MCMap utility 
print-coor (contact or aa-contact mode). The contact count data will be visualized as a 
color-coded histogram plot. In addition, for an overview of the location and composition 
of such contact hot spots, residues of receptor and ligand involved most frequently in 
contacts with the partner can be mapped back onto a protein surface for visualization by a 
PyMOL script.(1) The program further supports the subtraction of two related simulation data 
sets to visualize the main areas and residues where the biggest differences arise both in a 
contact histogram plot and, with the help of PyMOL (1), on the protein surfaces. This tool provides 
a quick way of connecting heatmaps of contact data with the visualization of the main 
contributors on a protein surface.

It defines methods for reading a configuration file with the user-specified settings 
(readConfig), reading and processing MCMap print-coor input (processPrintCoorHistogram), 
plotting contact count histograms (plotHistogram) and mapping the contact hot spots back 
onto protein surfaces with the help of PyMOL scripts (mapContactsToProtein, 1).

Copyright (C) Heidelberg, 2017 Ina Poehner

(1) PyMOL. The PyMOL Molecular Graphics System, Schroedinger, LLC.

This program is free software: you can redistribute it and/or modify it under the terms of 
the GNU General Public License as published by the Free Software Foundation, either version 3 
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program.  
If not, see <https://www.gnu.org/licenses/>.

'''

from sys import argv
import os

# Own modules
from readConfig import readConfigFile
from processPrintCoorHistogram import readInput, normalizeContactDict, subtractTwoInputs
from plotHistogram import createNumPyArray, plotHistogram
from mapContactsToProtein import sortByCounts, writePymolScript

__updated__ = '2017-12-29'

def extractAndProcessInput(parameterDict):
    '''
    This function calls all relevant routines to read the contact count histogram input files, 
    collect the data, normalize and subtract it, if desired. Normalization is performed whenever total 
    orientations are given and required if subtraction will be done. The collected and preprocessed
    data is then handed to the functions for plotting histograms and creating PyMOL scripts.
    '''
    inputList = parameterDict.get('input')
    if parameterDict.get('subtraction'): # Use subtraction mode
        encounterList = parameterDict.get('orientations') # get the two orientations
        contactDict, receptorSequence, ligandSequence = readInput(inputList[0]) # Preprocess first input
        normalizedContacts = normalizeContactDict(contactDict, encounterList[0]) # Normalize first input
        contactDict2, receptorSequence2, ligandSequence2 = readInput(inputList[1]) # Preprocess second input
        normalizedContacts2 = normalizeContactDict(contactDict2, encounterList[1]) # Normalize second input
        parameterDict['normalized'] = True
        # Perform the subtraction: Normalized input 1 - normalized input 2
        combiDict = subtractTwoInputs(normalizedContacts, normalizedContacts2, receptorSequence, receptorSequence2, ligandSequence, ligandSequence2)
        # Call plotting and protein mapping functions
        return plotContactHistogram(combiDict, receptorSequence, ligandSequence, parameterDict), createProteinMapping(combiDict, parameterDict)
    else: # No subtraction mode
        contactDict, receptorSequence, ligandSequence = readInput(inputList[0])
        if parameterDict.get('orientations') != None: # Do a normalization of the data prior to further processing
            normalizedContacts = normalizeContactDict(contactDict, parameterDict.get('orientations')[0])
            parameterDict['normalized'] = True
            return plotContactHistogram(normalizedContacts, receptorSequence, ligandSequence, parameterDict), createProteinMapping(normalizedContacts, parameterDict)
        else:
            parameterDict['normalized'] = False
            return plotContactHistogram(contactDict, receptorSequence, ligandSequence, parameterDict), createProteinMapping(contactDict, parameterDict)
        
def plotContactHistogram(contactDict, receptorAA, ligandAA, parameterDict):
    '''
    This function calls all relevant routines for conversion of the data in NumPy arrays and plotting a
    color coded heatmap of the contact data. Plotting function uses different representations for standard
    and subtraction data representation as well as normalized vs. non-normalized input.
    '''
    countArray, receptorArray, ligandArray, receptorAA, ligandAA = createNumPyArray(contactDict, receptorAA, ligandAA)
    plotFile = parameterDict.get('plot')
    labelDict = parameterDict.get('labels')
    
    switchParameter = dict()
    switchParameter['subtraction'] = parameterDict.get('subtraction')
    switchParameter['logscale'] = parameterDict.get('logscale')
    switchParameter['normalized'] = parameterDict.get('normalized')
    plotHistogram(countArray, receptorArray, ligandArray, receptorAA, ligandAA, plotFile, switchParameter, labelDict)
    return

def createProteinMapping(countDict, parameterDict):
    '''
    This function calls all relevant routines to prepare and write a text report of a user-defined percentage of 
    top contacting residues in both receptor and ligand and writes PyMOL scripts to map the selected contacts back 
    on the protein surface for visualization.
    '''
    modeSwitch = parameterDict.get('subtraction')
    # Extract the relevant top contacts
    mappingDict = sortByCounts(countDict, modeSwitch, parameterDict.get('maxcontacts'), parameterDict.get('display'))
    # Write the scripts for backmapping to receptor and ligand surface
    writePymolScript(parameterDict.get('receptor'), mappingDict.get('receptor'), modeSwitch, parameterDict.get('receptorpml'))
    writePymolScript(parameterDict.get('ligand'), mappingDict.get('ligand'), modeSwitch, parameterDict.get('ligandpml'))
    return

# ===== Main program body =====
print('This is PyCoALA - a Python Contact Area Localization and Analysis tool')
# Read config file from command line
configFileName = argv[1]
if not os.path.exists(configFileName):
    raise FileNotFoundError('Config file not found. Make sure ' + configFileName + ' does exist.')
else:
        chosenParameters = readConfigFile(configFileName)

extractAndProcessInput(chosenParameters)
