'''
mapContactsToProtein is a module to provide PyMOL scripts for mapping contact areas on receptor
and ligand surfaces.

The preprocessed input is reorganized by the count values and a user-specified percentage of the total
reported contacts is further processed for visualization. Considered counts are reported as a text file
for reference. A PyMOL (1) script for both receptor and ligand will be written to show top contact residues 
on the respective protein surface. For standard mode with a single input, the contact hot spot residues will 
be colored by an amino acid type based scheme. For subtraction mode, amino acids with positive counts will be 
highlighted in blue and negative counts in red. Subtraction data is sorted by the number of counts and then 
organized into four bins to provide a gradient representation from residues with largest differences to those 
with smaller differences within the requested top percentage.

mapContactsToProtein is part of PyCoALA.

Copyright (C) Heidelberg, 2017 Ina Poehner

(1) PyMOL. The PyMOL Molecular Graphics System, Schrödinger, LLC.

This program is free software: you can redistribute it and/or modify it under the terms of 
the GNU General Public License as published by the Free Software Foundation, either version 3 
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program.  
If not, see <https://www.gnu.org/licenses/>.

'''

def sortByCounts(contactDict, modeSwitch, outFile, reportPercentage):
    '''
    Re-sort the data by the final count values from top count to minimal count. From the 
    reorganized data, the requested percentage of residues will be extracted in following steps.
    '''
    # get a list of contact tuples ordered by contact count values; use reverse to go from highest to lowest count    
    orderedContactTupleList = sorted(contactDict, key=contactDict.get, reverse=True)

    finalDict, pairsToReport = extractAAToMap(modeSwitch, orderedContactTupleList, reportPercentage)
    if modeSwitch:
        reportList = orderedContactTupleList[0:pairsToReport]
        counter = len(orderedContactTupleList)-pairsToReport
        while counter < len(orderedContactTupleList):
            reportList.append(orderedContactTupleList[counter])
            counter += 1
    elif not modeSwitch:
        reportList = orderedContactTupleList[0:pairsToReport]
    textReport(contactDict, reportList, outFile) 
    return finalDict

def textReport(countDict, reportTuples, outFile): 
    '''
    Provide a text report summary of the top (or top/lowest) counts, which will be used for the mapping 
    procedure for reference. If atomic resolution is used, the atom-wise counts are reported here. 
    '''
    fobj_textreport = open(outFile, "w")
    for contactTuple in reportTuples:
        receptorTuple = contactTuple[0]
        ligandTuple = contactTuple[1]
        for entry in receptorTuple:
            if isinstance(entry, int):
                fobj_textreport.write(str(entry) + ' \t ')
            else:
                fobj_textreport.write(entry + ' \t ')
        fobj_textreport.write(' - \t')
        for entry in ligandTuple:
            if isinstance(entry, int):
                fobj_textreport.write(str(entry) + ' \t ')
            else:
                fobj_textreport.write(entry + ' \t ')
        fobj_textreport.write(str(countDict.get(contactTuple)) + '\n')
    fobj_textreport.close()
    return

def extractAAToMap(modeSwitch, aminoAcidsList, percentage): 
    '''
    Function to decide on basis of the requested report percentage how many amino acids need 
    to be kept for reporting. From that, a dictionary for both receptor and ligand amino acids 
    to be visualized in the PyMOL scripts is created.
    '''
    totalPairs = len(aminoAcidsList) # length of list corresponds to available pairs with contact counts

    grabPairs = int(round(totalPairs * (percentage/100)))
    
    if grabPairs == 0: # If no contact pair would be considered, take only the top count
        grabPairs = 1
        
    mappingDict = createMappingDict(modeSwitch, aminoAcidsList, grabPairs)    
    return mappingDict, grabPairs

def createMappingDict(modeSwitch, aminoAcidsList, grabPairs):
    '''
    Depending on the mode, either, for standard mode, the top x percent of the total 
    counts or, for subtraction mode, both the top and the lowest x percent are collected.
    '''
    finalMappedDict = dict()
    receptorDict = dict()
    ligandDict = dict()
    
    pairsToMap = aminoAcidsList[:grabPairs]
    receptorAA, ligandAA = mapReceptorAndLigand(pairsToMap)
    
    if not modeSwitch: # Standard type based mode
        receptorDict['standard'] = receptorAA
        ligandDict['standard'] = ligandAA
    elif modeSwitch: # Subtraction mode
        receptorDict['positive'] = splitToBins(receptorAA)
        ligandDict['positive'] = splitToBins(ligandAA)
        swappedList = aminoAcidsList[::-1] # invert list to get lowest counts
        pairsToMap = []
        pairsToMap = swappedList[:grabPairs]
        receptorAA=[]
        ligandAA=[]
        receptorAA, ligandAA = mapReceptorAndLigand(pairsToMap)
        receptorDict['negative'] = splitToBins(receptorAA)
        ligandDict['negative'] = splitToBins(ligandAA)
        receptorDict = markDuplicates(receptorDict)
        ligandDict = markDuplicates(ligandDict)
    finalMappedDict['receptor'] = receptorDict
    finalMappedDict['ligand'] = ligandDict
    return finalMappedDict

def decompressGroupLists(groupDict):
    '''
    Function to decompress a dictionary with amino acid lists distributed in different groups to 
    an overall list to allow checking for duplicates.
    '''
    listOfEntries = []
    for entry in groupDict.keys():
        acidList = groupDict.get(entry)
        for acid in acidList:
            listOfEntries.append(acid)
    return listOfEntries

def markDuplicates(proteinDict):
    '''
    Function to list amino acids which can be found both in groups with positive and negative counts
    after a subtraction. Duplicates will be marked to allow different coloring.
    '''
    positiveGroups = proteinDict.get('positive')
    negativeGroups = proteinDict.get('negative')
    
    positiveList = decompressGroupLists(positiveGroups)
    negativeList = decompressGroupLists(negativeGroups)
        
    for entry in positiveList:
        if entry in negativeList:
            if 'duplicate' in proteinDict.keys():
                currDups = proteinDict.get('duplicate')
                currDups.append(entry)
                proteinDict['duplicate'] = currDups
            else:
                dupList = []
                dupList.append(entry)
                proteinDict['duplicate'] = dupList
    return proteinDict

def splitToBins(aminoAcidList): 
    '''
    Function to distribute the selected amino acids responsible for top counts into four groups
    with each group holding roughly a quarter of the total residues sorted by their contact count.
    '''
    binnedDict = dict()
    if len(aminoAcidList) <= 4: # If there are only very few residues, just assign all to a single group
        for entry in aminoAcidList:
            binnedDict = constructBinnedDicts('group1', binnedDict, entry)
    else:
        residuesPerBin = int(round(len(aminoAcidList)/4))

        counter = 0
        while counter < len(aminoAcidList):
            if counter < residuesPerBin:
                binnedDict = constructBinnedDicts('group1', binnedDict, aminoAcidList[counter])
            elif counter < (2 * residuesPerBin):
                binnedDict = constructBinnedDicts('group2', binnedDict, aminoAcidList[counter])
            elif counter < (3 * residuesPerBin):
                binnedDict = constructBinnedDicts('group3', binnedDict, aminoAcidList[counter])
            else:
                binnedDict = constructBinnedDicts('group4', binnedDict, aminoAcidList[counter])
            counter += 1
    return binnedDict

def constructBinnedDicts(groupKey, binDict, newEntry):
    '''
    Function to prepare or fill the dictionary assigning contact count data from a subtraction
    run into four sub-groups for visualization.
    '''
    if groupKey in binDict.keys():
        currentList = binDict.get(groupKey)
        currentList.append(newEntry)
        binDict[groupKey] = currentList
    else:
        newList = []
        newList.append(newEntry)
        binDict[groupKey] = newList
    return binDict

def mapReceptorAndLigand(topTuples):
    '''
    Separate the receptor and ligand amino acids from the top contact tuples. If data was processed at 
    atomic resolution, only the amino acid containing the atom will be processed further but atomic data
    will be dropped.
    '''
    receptorList = []
    ligandList = []
    
    for contactPair in topTuples:
        if len(contactPair[0]) == 2: # Amino acid resolution of input
            if contactPair[0] not in receptorList:
                receptorList.append(contactPair[0])
        elif len(contactPair[0]) == 3: # atomic resolution of input; drop atom information for hot spot mapping
            receptorTuple = contactPair[0]
            acidTuple = (receptorTuple[0], receptorTuple[1])
            if acidTuple not in receptorList:
                receptorList.append(acidTuple)
        if len(contactPair[1]) == 2:
            if contactPair[1] not in ligandList:
                ligandList.append(contactPair[1])
        elif len(contactPair[1]) == 3:
            ligandTuple = contactPair[1]
            ligAcidTuple = (ligandTuple[0], ligandTuple[1])
            if ligAcidTuple not in ligandList:
                ligandList.append(ligAcidTuple)                          
    return receptorList, ligandList

def isAminoAcidType(aminoAcidCode):
    '''
    Function to return the type for a given 3-letter code of an amino acid, if it is known. Currently, all 
    20 standard amino acids are classified. In addition, N epsilon or N delta protonated histidines (HSE/HSD
    from CHARMM force field) are included.
    '''
    # Dictionary for amino acid type assignment
    aminoAcidTypeDict = {'ASP': 'acidic', 'GLU': 'acidic', 
                         'HIS': 'basic', 'HSE': 'basic', 'HSD': 'basic', 'LYS': 'basic', 'ARG': 'basic',
                         'CYS': 'polar', 'GLY': 'polar', 'ASN': 'polar', 'GLN': 'polar', 'SER': 'polar', 'THR': 'polar',
                         'PHE': 'aromatic', 'TRP': 'aromatic', 'TYR': 'aromatic',
                         'ALA': 'apolar', 'ILE': 'apolar', 'LEU': 'apolar', 'MET': 'apolar', 'PRO': 'apolar', 
                         'VAL': 'apolar'}
    
    if aminoAcidCode in aminoAcidTypeDict.keys():
        aminoAcidType = aminoAcidTypeDict.get(aminoAcidCode)
    else:
        aminoAcidType = None # currently unknown type
    return aminoAcidType

def typeColorScheme(aminoAcid):
    '''
    Color scheme for PyMOL based on amino acid types. Calls a function to determine the amino acid
    type and returns a corresponding color for that type.
    
    Apolar/aliphatic (ALA, ILE, LEU, MET, PRO, VAL) are colored in dark gray;
    Acidic (ASP, GLU) in red; Basic (HIS, LYS, ARG) in blue; Aromatic (PHE, TRP, TYR) in dark purple; 
    Polar (CYS, GLY, ASN, GLN, SER, THR) in green. 
    Residues and amino acids of unknown type are colored in cyan.
    '''
    # PyMOL Colors assigned to the different amino acid types
    typeColorDict = {'acidic': 'red', 'apolar': 'gray30', 'aromatic': 'deeppurple', 'basic': 'blue', 'polar': 'forest'}
    
    aminoAcidType = isAminoAcidType(aminoAcid)
    if aminoAcidType in typeColorDict.keys():
        color = typeColorDict.get(aminoAcidType)
    else:
        color = 'cyan'                
    return color

def subtractColorScheme(posNeg, groupID):
    '''
    Function to assign the colors which are defined by RGB codes in writeColorMap() to groups containing
    residues. Residues were binned into four groups for positive and negative counts, each. 
    Positive counts are colored blue, negative counts are colored red.
    The assigned colors are blue, midblue, midlightblue and lightblue for the positive groups and
    red, midred, midlightred and lightred for negative groups.
    '''
    colorDict = {'positive': ['blue', 'midblue', 'midlightblue','lightblue'], 
                 'negative': ['red',  'midred', 'midlightred', 'lightred']}
    
    if groupID == 'group1':
        return colorDict.get(posNeg)[0]
    elif groupID == 'group2':
        return colorDict.get(posNeg)[1]
    elif groupID == 'group3':
        return colorDict.get(posNeg)[2]
    else:
        return colorDict.get(posNeg)[3]
        
def writeColorMap(fobj_out):
    '''
    Color scheme for subtraction mode: Residues are binned into four groups for positive and negative
    counts, each. The most negative/positive bin is colored red/blue, respectively. Following bins
    subsequently get lighter. The color definitions are done by specifying RGB codes and can be 
    altered in this function, if required.
    
    Current defaults are (rgb codes):
    blue = (0.00, 0.00, 1.00); midblue =  (0.02, 0.50, 1.00); midlightblue = (0.50, 0.70, 1.0 ); 
    lightblue = (0.86, 1.00 ,1.00); red = (1.00, 0.00, 0.00); midred = (1.00, 0.30, 0.30);
    midlightred = (1.00, 0.50, 0.50), lightred = (1.00, 0.80, 0.80)
    '''
    colorList = ['set_color blue= [0.00, 0.00, 1.00]',
                 'set_color midblue=  [0.02, 0.50, 1.00]',
                 'set_color midlightblue= [0.50, 0.70, 1.00]',
                 'set_color lightblue= [0.86, 1.00, 1.00]',
                 'set_color red=   [1.00, 0.00, 0.00]',
                 'set_color midred=  [1.00, 0.30, 0.30]',
                 'set_color midlightred=  [1.00, 0.50, 0.50]',
                 'set_color lightred=  [1.00, 0.80, 0.80]']
    
    for colorDefinition in colorList:
        fobj_out.write(colorDefinition + '\n')
    return
    
def writePymolScript(proteinID, proteinData, subtractionMode, outFile):
    '''
    Function to write the PyMOL script visualizing the determined top contacts on a protein
    surface. 
    
    A script will be written following this general structure:
    
    load [path-to-protein/protein.pqrm], proteinname, 0, pqr
    show surface, proteinname
    color gray80, proteinname
    set transparency, 0.5
    show cartoon, proteinname
    hide lines, proteinname
    hide nonbonded, proteinname
    
    show sticks, resi [number]
    color [some color], resi [number]
    label n. CA and i. [number], "AA_number" # custom text: AA and number
    
    set label_size, 20
    set stick_radius, 0.1
    set cartoon_side_chain_helper, on
    
    '''
    fobj_out = open(outFile, "w")
    #Protein ID is a list holding both the full path to protein and the protein name
    proteinName = proteinID[1]
    fobj_out.write('load ' + proteinID[0] + ', ' + proteinName + ', 0, pqr \n') 
    fobj_out.write('show surface, ' + proteinName + '\n')
    fobj_out.write('color gray80, ' + proteinName + '\n') # color surface close to white
    fobj_out.write('set transparency, 0.5 \n') #optional line to see the labels
    fobj_out.write('show cartoon, ' + proteinName + '\n')
    fobj_out.write('hide lines, ' + proteinName + '\n')
    fobj_out.write('hide nonbonded, ' + proteinName + '\n')
    if not subtractionMode:
        aminoAcidList = proteinData.get('standard')
        for acid in aminoAcidList:
            color = typeColorScheme(acid[1])
            fobj_out.write('show sticks, resi ' + str(acid[0]) + '\n')
            fobj_out.write('color ' + color + ', resi ' + str(acid[0]) + '\n')
            fobj_out.write('label n. CA and i. ' + str(acid[0]) + ', "' + acid[1] + str(acid[0]) + '"\n')
    elif subtractionMode:
        writeColorMap(fobj_out)
        if 'duplicate' in proteinData.keys():
            duplicates = proteinData.get('duplicate')
        else:
            duplicates = []
        positives = proteinData.get('positive')
        for group in sorted(positives.keys(), reverse=True): # start with last group and lightest color 
            # Thus best color group will be kept if amino acid is in more than one group
            currCol = subtractColorScheme('positive', group)
            positiveList = positives.get(group)
            for acid in positiveList:                    
                fobj_out.write('show sticks, resi ' + str(acid[0]) + '\n')
                if acid in duplicates:
                    fobj_out.write('color purple, resi ' + str(acid[0]) + '\n')
                else:
                    fobj_out.write('color ' + currCol + ', resi ' + str(acid[0]) + '\n')
                fobj_out.write('label n. CA and i. ' + str(acid[0]) + ', "' + acid[1] + str(acid[0]) + '"\n')
        negatives = proteinData.get('negative')
        for group in sorted(negatives.keys(), reverse=True):
            currCol = subtractColorScheme('negative', group)
            negativeList = negatives.get(group)
            for acid in negativeList:
                if acid not in duplicates:
                    fobj_out.write('show sticks, resi ' + str(acid[0]) + '\n')
                    fobj_out.write('color ' + currCol + ', resi ' + str(acid[0]) + '\n')
                    fobj_out.write('label n. CA and i. ' + str(acid[0]) + ', "' + acid[1] + str(acid[0]) + '"\n')
    fobj_out.write('set label_size, 20 \n')
    fobj_out.write('set stick_radius, 0.1 \n')
    fobj_out.write('set cartoon_side_chain_helper, on \n')
    fobj_out.close()
    return
