#
# example script for doing a Monte Carlo simulation with MCMap
# 
# J. M. Foerster 
# I. Poehner
# G. M. Ullmann, Bayreuth Nov 2017
#
#=============================================================
# Required external software: APBS, CHARMM
#
# Required input files:
#      receptor crd / psf-File
#      ligand crd / psf-File
#
#===================================================
#
# The master script has several break points, for instance
#
#     csh master.csh 5 
#
# moves to the fifth step
#
# There are the following steps:
#
#   1 -- shift the PQR files for Cc to the center of mass
#   2 -- run APBS
#   3 -- create the exclusion grid
#   4 -- run MCMap
#   5 -- process simulation output with print-coor
#        and generate different representations.
#        Run PyCoALA to generate the contact map plot and map the 
#        interactions to the surface of the proteins via a PyMOL script
#
#===================================================
# Environment variables and paths to the required programs:
# 
setenv MCMAP    ../bin/
setenv PYSC  ../bin/PyCoALA/
## Set PATH to APBS:
setenv APBS  /sw/sci/app/apbs/apbs-1.4-amd64/bin/
setenv DXMT  /sw/sci/app/apbs/apbs-1.4-amd64/share/apbs/tools/bin/
#=============================================================
# Here, define the file names of your receptor and ligand file
# as well as the ionic strength in M.
set crystal = 2pcc
set rec = 2pcc_ccp
set lig = 2pcc_cytc
set ion = 0.01
#
# Prepare PQRM with PSFCRD2PQR and separate ligand / receptor
# Shift the coordinate center of mass to the origin
# With this step, the molecule is centered in the middle of future grids.
#
if ($1 <= 1) then
  $MCMAP/psfcrd2pqr $crystal.psf $crystal.crd PQRM > 2pcc.pqrm
  head -4629 2pcc.pqrm > $rec.pqrm
  grep 'HEM   296' 2pcc.pqrm >> $rec.pqrm
  tail -2594 2pcc.pqrm | grep -v HOH | grep -v 'HEM   296' > $lig.pqrm
  $MCMAP/shift-center $lig.pqrm > $lig-shifted.pqrm
  $MCMAP/shift-center $rec.pqrm > $rec-shifted.pqrm
endif

#
#----------------------------------------------------------------
# Calculate the electrostatic potentials using APBS:
# APBS and its input options are described at www.poissonboltzmann.org

if ($1 <= 2) then

foreach i ($ion)
  echo $i
cat <<EOF >  apbs-$i.in
read
  mol pqr  $rec-shifted.pqrm
end
elec
  mg-manual
  dime 257 257 257
  grid 2.0 2.0 2.0
  gcent 0.0 0.0 0.0
  nlev 4
  mol 1
  lpbe
  bcfl sdh
  srfm mol
  chgm spl0
  ion charge  1 conc $i radius 2.0
  ion charge -1 conc $i radius 2.0
  pdie   4.0
  sdie  80.0
  sdens  10.0
  srad  1.4
  temp  300.0
  calcenergy no
  calcforce  no
  write pot dx pot-$i-2.0
end
elec
  mg-manual
  dime 289 289 289
  grid 1.0 1.0 1.0
  gcent 0.0 0.0 0.0
  nlev 4
  mol 1
  lpbe
  bcfl focus
  srfm mol
  chgm spl0
  ion charge  1 conc $i radius 2.0
  ion charge -1 conc $i radius 2.0
  pdie   4.0
  sdie  80.0
  sdens  10.0
  srad  1.4
  temp  300.0
  calcenergy no
  calcforce  no
  write pot dx pot-$i-1.0
end
quit
EOF
cat apbs-$i.in
$APBS/apbs apbs-$i.in > apbs-$i.out   
#
# convert the potential from kT/e to kcal/mol/e with DXMATH, a subprogramm of APBS
#
cat <<EOF >dxmath.inp
pot-$i-2.0.dx
0.592 *
pot-$i-2.0-kcal.dx =
EOF
$DXMT/dxmath dxmath.inp
cat <<EOF >dxmath.inp
pot-$i-1.0.dx
0.592 *
pot-$i-1.0-kcal.dx =
EOF
$DXMT/dxmath dxmath.inp
end
endif      
#
#----------------------------------------------------------------
# Make the Exclusion grid to prevent collisions of ligand and receptor
#
if ($1 <= 3) then
$MCMAP/make-excl $rec-shifted.pqrm 1.4 0.25 > $rec.excl-grid
endif
#
#-----------------------------------------------------------------
#
# Do the Monte Carlo Simulation with MCMap
#
if ($1 <= 4) then
foreach i ($ion)
cat <<EOF  > mcmap-$i.inp  
number_of_runs   1000                                            
nprint           1                                              
nprintdens       100                                             
nprintdist       1000                                            
istuck           50                                              
rad_in           130.0                                           
rad_out          180.0                                           
temperature      300.0                                           
mcsteps          25000 
dev_trans        3.0                                             
dev_rot          5.0                                             
receptor         $rec-shifted.pqrm                               
ligand           $lig-shifted.pqrm                               
pot_out          pot-$i-2.0-kcal.dx                              
pot_in           pot-$i-1.0-kcal.dx                              
excl             $rec.excl-grid                                  
traj             $rec-$i.traj.gz                                    
structene        $rec-$i.structene.gz                               
structdist       $rec-$i.structdist.gz                           
lowenergy        -16.0                                           
cmdist           70.0                                            
dockdensity      $rec-$i.dx                                      
EOF
  $MCMAP/mcmap mcmap-$i.inp > mcmap-$i.out
end
endif

if ($1 == 5) then
#
# get the coordinates of the ligand in form of indiviual pqr files for every 1000th (n-th) ligand position
#
mkdir Struct
$MCMAP/print-coor -m lig -o $rec-$ion.structdist.gz -r $rec-shifted.pqrm \
                -l $lig-shifted.pqrm -n 1000 
#
# get the coordinates of the receptor in form of indiviual pqr files for every 1000th (n-th) receptor  position
#
$MCMAP/print-coor -m rec -o $rec-$ion.structdist.gz -r $rec-shifted.pqrm \
                -l $lig-shifted.pqrm -n 1000
#
# get the cm-coordinates of the ligand
#
$MCMAP/print-coor -m cm-lig -o $rec-$ion.structdist.gz -r $rec-shifted.pqrm \
                -l $lig-shifted.pqrm > lig-cm.pdb
#
# get the cm-coordinates of the receptor
#
$MCMAP/print-coor -m cm-rec -o $rec-$ion.structdist.gz -r $rec-shifted.pqrm \
                -l $lig-shifted.pqrm  > rec-cm.pdb
#
# get the dx of the ligand
#
$MCMAP/print-coor -m dx-lig -o $rec-$ion.structdist.gz -r $rec-shifted.pqrm \
                -l $lig-shifted.pqrm -d 100.0 -s 1.0 > lig.dx
#
# get the dx of the receptor
#
$MCMAP/print-coor -m dx-rec -o $rec-$ion.structdist.gz -r $rec-shifted.pqrm \
                -l $lig-shifted.pqrm -d 100.0 -s 1.0 > rec.dx

# 
# get the HEME distribution of all hemes being separated 20 angstroem of each other
# The minimal distance calculation can be computationally very demanding;
#
grep HEM $rec-shifted.pqrm > HEM-$rec-shifted.pqrm
grep HEM $lig-shifted.pqrm > HEM-$lig-shifted.pqrm
$MCMAP/print-coor -m min-dist -o $rec-$ion.structdist.gz -r HEM-$rec-shifted.pqrm \
                -l HEM-$lig-shifted.pqrm -d 20.0 -s 1.0 > min-dist.dx

#
# Here, a new orientation file with a subensemble is created which only includes the orientations
# within 5 angstroem of the receptor surface
#
$MCMAP/make-excl $rec-shifted.pqrm  5.0 0.25 > $rec.incl-grid
$MCMAP/print-coor -m ens-surf -o $rec-$ion.structdist.gz -r $rec-shifted.pqrm \
                -l $lig-shifted.pqrm -inc $rec.incl-grid -d 100.0 -s 1.0> structure-included.dx


#
# get the contact map histogram for all contacts within a range of 8 anstroem.
# In order to speed up the simulation, only heavy atoms are considered.
#
grep -v ' 1.008' $rec-shifted.pqrm > $rec-shifted-h.pqrm
grep -v ' 1.008' $lig-shifted.pqrm > $lig-shifted-h.pqrm
$MCMAP/print-coor -m aa-contact -d 8.0 -o $rec-$ion.structdist.gz -r $rec-shifted-h.pqrm \
                -l $lig-shifted-h.pqrm > titrat-histo_$ion.dat

#
# write the config file for PyCoALA and the surface mapping with PyMOL: 
#
cat <<EOF > config.txt
# This line is a comment.
# For further information on any of the chosen parameters and possible default values, please refer to the documentation.

# Input file(s) of the contact count histogram data obtained with print-coor contact or print-coor aa-contact mode. Providing input data is mandatory.
# For subtraction mode, specify two input files comma-separated. Subtraction will be performed file 1 - file 2. Note that for subtraction, providing orientations is mandatory.
Input = titrat-histo_$ion.dat

# Number of total encounter complexes reported in your simulation for normalizing the contact count data.
# Setting this option is mandatory when performing a subtraction of results from two runs.
# For subtraction, specify the two total orientation counts comma separated in the same order as the corresponding input files.
# Orientations can either be provided as MCMap orientation files (structdist.gz) or as numbers.
# If no number of orientations is provided, no normalization will be performed.
#Orientations = $rec-$ion.structdist.gz

# Name of the plotted histogram output file. Plot will be written as PNG.
Plot = 2pcc_contact_plot.png 

# Title for the plot. Title needs to be surrounded by "".
Title = "Contact map for CcP with Cc"

# Label for the x-axis (receptor amino acid/atom), label needs to be surrounded by "".
xlabel = "Amino acid number CcP"

# Label for the y-axis (ligand amino acid/atom), label needs to be surrounded by "".
ylabel = "Amino acid number Cc" 

# Optional switch for plotting count data on a logarithmic scale if using single, non-normalized input.
# Can be True or False.
Logscale = True

# PQRM file of the receptor.
Receptor = $rec-shifted.pqrm

# PQRM file of the ligand.
Ligand = $lig-shifted.pqrm

# Percentage of residues with top counts (or top and lowest counts for subtraction) with respect to the total number of contacts to be visualized with the PyMOL script.
Display = 0.5

# Output file for a text report on the top contacts that will be visualized on the protein surface and their contact counts.
MaxContacts = report_contact_plot.txt

# Output file for the receptor PyMOL script.
ReceptorPml = yccp_script_contact_map.pml

# Output file for the ligand PyMOL script.
LigandPml = ycc_script_contact_map.pml
EOF

#
# Run the contact map processing script, which maps the contacts on the structure by a PyMOL script.
# Use python3 in order to massively speed up the mapping
#
python3 $PYSC/PyCoALA.py config.txt

##
## Load the PyMOL scripts with PyMOL:
##
# pymol yccp_script_contact_map.pml
# pymol ycc_script_contact_map.pml

endif
#*****************************************************************
