#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "memutil.h"
#include "complex.h"

#define NR_END 1
#define FREE_ARG void*

dcomplex Cadd(dcomplex a, dcomplex b)
{
	dcomplex c;
	c.r=a.r+b.r;
	c.i=a.i+b.i;
	return c;
}

dcomplex Csub(dcomplex a, dcomplex b)
{
	dcomplex c;
	c.r=a.r-b.r;
	c.i=a.i-b.i;
	return c;
}


dcomplex Cmul(dcomplex a, dcomplex b)
{
	dcomplex c;
	c.r=a.r*b.r-a.i*b.i;
	c.i=a.i*b.r+a.r*b.i;
	return c;
}

dcomplex Complex(double re, double im)
{
	dcomplex c;
	c.r=re;
	c.i=im;
	return c;
}

dcomplex Conjg(dcomplex z)
{
	dcomplex c;
	c.r=z.r;
	c.i = -z.i;
	return c;
}

dcomplex Cdiv(dcomplex a, dcomplex b)
{
	dcomplex c;
	double r,den;
	if (fabs(b.r) >= fabs(b.i)) {
		r=b.i/b.r;
		den=b.r+r*b.i;
		c.r=(a.r+r*a.i)/den;
		c.i=(a.i-r*a.r)/den;
	} else {
		r=b.r/b.i;
		den=b.i+r*b.r;
		c.r=(a.r*r+a.i)/den;
		c.i=(a.i*r-a.r)/den;
	}
	return c;
}

double Cabs(dcomplex z)
{
	double x,y,ans,temp;
	x=fabs(z.r);
	y=fabs(z.i);
	if (x == 0.0)
		ans=y;
	else if (y == 0.0)
		ans=x;
	else if (x > y) {
		temp=y/x;
		ans=x*sqrt(1.0+temp*temp);
	} else {
		temp=x/y;
		ans=y*sqrt(1.0+temp*temp);
	}
	return ans;
}

dcomplex Csqrt(dcomplex z)
{
	dcomplex c;
	double x,y,w,r;
	if ((z.r == 0.0) && (z.i == 0.0)) {
		c.r=0.0;
		c.i=0.0;
		return c;
	} else {
		x=fabs(z.r);
		y=fabs(z.i);
		if (x >= y) {
			r=y/x;
			w=sqrt(x)*sqrt(0.5*(1.0+sqrt(1.0+r*r)));
		} else {
			r=x/y;
			w=sqrt(y)*sqrt(0.5*(r+sqrt(1.0+r*r)));
		}
		if (z.r >= 0.0) {
			c.r=w;
			c.i=z.i/(2.0*w);
		} else {
			c.i=(z.i >= 0) ? w : -w;
			c.r=z.i/(2.0*c.i);
		}
		return c;
	}
}

dcomplex RCmul(double x, dcomplex a)
{
	dcomplex c;
	c.r=x*a.r;
	c.i=x*a.i;
	return c;
}

/*........................................................*/
dcomplex *dcomplex_vector(long nl, long nh)
/* allocate a dcomplex vector with subscript range v[nl..nh] */
{
        dcomplex *v;

        v=(dcomplex *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(dcomplex)));
        if (!v) memerror("allocation failure in dcomplex_vector()");
        return v-nl+NR_END;
}

/*........................................................*/
void free_dcomplex_vector(dcomplex *v, long nl, long nh)
/* free a vector allocated with cvector() */
{
  if (nl<nh) memerror("failure in free_dcomplex_vector()");
if (v!=NULL)    free((FREE_ARG) (v+nl-NR_END));
}
/*........................................................*/

dcomplex **dcomplex_matrix(long nrl, long nrh, long ncl, long nch)
/* allocate a dcomplex matrix with subscript range m[nrl..nrh][ncl..nch] */
{
        long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
        dcomplex **m;

        /* allocate pointers to rows */
        m=(dcomplex **) malloc((size_t)((nrow+NR_END)*sizeof(dcomplex*)));
        if (!m) memerror("allocation failure 1 in dcomplex_matrix()");
        m += NR_END;
        m -= nrl;

        /* allocate rows and set pointers to them */
        m[nrl]=(dcomplex *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(dcomplex)));
        if (!m[nrl]) memerror("allocation failure 2 in dcomplex_matrix()");
        m[nrl] += NR_END;
        m[nrl] -= ncl;

        for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

        /* return pointer to array of pointers to rows */
        return m;
}

/*........................................................*/
void free_dcomplex_matrix(dcomplex **m, long nrl, long nrh, long ncl, long nch)
/* free a double matrix allocated by dmatrix() */
{
  if ((nrl>nrh)||(ncl>nch)) memerror("Problem in free_dcomplex_matrix");
  if (m!=NULL)    {free((FREE_ARG) (m[nrl]+ncl-NR_END));
                   free((FREE_ARG) (m+nrl-NR_END));
                  }
}
/*..........................................*/
/* normalize a complex vector               */
/*                                          */
void Cnormalize(dcomplex *v,int n)
{
    int i;
    dcomplex sum;

    sum.r=0.0; sum.i=0.0;
    for (i=1; i<=n; i++)
        {sum=Cadd(sum,Cmul(v[i],v[i]));
        }
    sum=Csqrt(sum);
    for (i=1; i<=n; i++)
        {if (Cabs(sum)!=0.0) v[i]=Cdiv(v[i],sum);
        }
}
