/* G. Matthias Ullmann                                          */
/* Bayreuth, April 2010                                         */
/*                                                              */
/* Some routines for reading and writing Gaussian Cube files    */
/*                                                              */
/* ............................................................ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "util.h"
#include "memutil.h"
#include "grid.h"



/* ............................................................ */
/* read the volumenometric data of a Gaussian Cube file         */
/*                                                              */
/* ATTENTION: This procedure may read some Cube files not       */
/*            correctly. In order to transfer it into a dGrid   */
/*            structure, special charaterizations need to be    */
/*            justified.                                        */
/*                                                              */
/* The length unit in this file is assumed to be Bohr           */
/*                                                              */
void ReadCube(Cube *dx, FILE *fp)
{
     char line[LINELENGTH];
     int i, h, nc, ic, natom;
     int ix, iy, iz;
     double hv[7];

     /* skip the first two lines */
     READ_LINE(line, fp);
     READ_LINE(line, fp);

     /* now read the file */
     READ_LINE(line, fp);
     h=sscanf(line, "%d %lf %lf %lf",
              &natom, &(*dx).ori[1], &(*dx).ori[2], &(*dx).ori[3]);
     if (h!=4) ERROR("error while reading cube-file 1");

     READ_LINE(line, fp);
     if (h!=4) ERROR("error while reading cube-file 2");
     h=sscanf(line, "%d %lf %lf %lf", &(*dx).npt[1],
            &(*dx).delta[1][1], &(*dx).delta[1][2], &(*dx).delta[1][3]);
             
     READ_LINE(line, fp);
     h=sscanf(line, "%d %lf %lf %lf",&(*dx).npt[2], 
            &(*dx).delta[2][1], &(*dx).delta[2][2], &(*dx).delta[2][3]);
     if (h!=4) ERROR("error while reading cube-file 3");

     READ_LINE(line, fp);
     h=sscanf(line, "%d %lf %lf %lf",&(*dx).npt[3], 
            &(*dx).delta[3][1], &(*dx).delta[3][2], &(*dx).delta[3][3]);
     if (h!=4) ERROR("error while reading cube-file 4");

     /* now skip the atom information */
     for (i=1; i<=natom; i++)
         READ_LINE(line, fp);

     /* now read the potential data */
     (*dx).data = d3tensor(1,(*dx).npt[1],1,(*dx).npt[2],1,(*dx).npt[3]);

     ic = 0;
     ix=1; iy=1; iz=1;
     while (fgets(line,LINELENGTH,fp)!=NULL)
           {nc=sscanf(line, "%lf %lf %lf %lf %lf %lf",
                      &hv[1],&hv[2],&hv[3],&hv[4],&hv[5],&hv[6]);
            for (i=1; i<=nc; i++)
                {(*dx).data[ix][iy][iz+(i-1)]=hv[i];
                }
            if (nc <=0) break;
            ic = ic + nc;
            iz = iz + nc;
            if (iz > (*dx).npt[3])
                {iz = 1; iy++;
                }
            if (iy > (*dx).npt[2])
                {iz = 1; iy = 1; ix++;
                }
            if (ix > (*dx).npt[1]+1)
               {fprintf(stderr,"%s", line);
                WARN("Error while reading cube-file");
               }
           }
     if (ic!=(*dx).npt[1]*(*dx).npt[2]*(*dx).npt[3])
         ERROR("ic!=npt[1]*npt[2]*npt[3] while reading cube-file");
}
/* ............................................................ */
/* calculates the difference between two cube files             */
/* it returns 0 if successful                                   */
/*           -1 if a problem occured                            */
/*                                                              */
int DiffCube(Cube *ddx, Cube *dx1, Cube *dx2)
{

#define EPS 0.0000000001

    int ix, iy, iz;

    /* check if the file are comparable */
    if ((fabs((*dx1).ori[1]     -(*dx2).ori[1])>EPS)||
        (fabs((*dx1).ori[2]     -(*dx2).ori[2])>EPS)||
        (fabs((*dx1).ori[3]     -(*dx2).ori[3])>EPS)||
        (fabs((*dx1).delta[1][1]-(*dx2).delta[1][1])>EPS)||
        (fabs((*dx1).delta[1][2]-(*dx2).delta[1][2])>EPS)||
        (fabs((*dx1).delta[1][3]-(*dx2).delta[1][3])>EPS)||
        (fabs((*dx1).delta[2][1]-(*dx2).delta[2][1])>EPS)||
        (fabs((*dx1).delta[2][2]-(*dx2).delta[2][2])>EPS)||
        (fabs((*dx1).delta[2][3]-(*dx2).delta[2][3])>EPS)||
        (fabs((*dx1).delta[3][1]-(*dx2).delta[3][1])>EPS)||
        (fabs((*dx1).delta[3][2]-(*dx2).delta[3][2])>EPS)||
        (fabs((*dx1).delta[3][3]-(*dx2).delta[3][3])>EPS)||
        ((*dx1).npt[1]!=(*dx2).npt[1])||
        ((*dx1).npt[2]!=(*dx2).npt[2])||
        ((*dx1).npt[3]!=(*dx2).npt[3]))
           {WARN("Cube-files do not have the same dimensions or are shifted");
            return 1;
           }

    /* assign the dimensions */
    (*ddx).ori[1]      = (*dx2).ori[1];
    (*ddx).ori[2]      = (*dx2).ori[2];
    (*ddx).ori[3]      = (*dx2).ori[3];
    (*ddx).delta[1][1] = (*dx2).delta[1][1];
    (*ddx).delta[1][2] = (*dx2).delta[1][2];
    (*ddx).delta[1][3] = (*dx2).delta[1][3];
    (*ddx).delta[2][1] = (*dx2).delta[2][1];
    (*ddx).delta[2][2] = (*dx2).delta[2][2];
    (*ddx).delta[2][3] = (*dx2).delta[2][3];
    (*ddx).delta[3][1] = (*dx2).delta[3][1];
    (*ddx).delta[3][2] = (*dx2).delta[3][2];
    (*ddx).delta[3][3] = (*dx2).delta[3][3];
    (*ddx).npt[1]      = (*dx2).npt[1];
    (*ddx).npt[2]      = (*dx2).npt[2];
    (*ddx).npt[3]      = (*dx2).npt[3];

    /* calculate the difference */
    (*ddx).data = d3tensor(1,(*ddx).npt[1],1,(*ddx).npt[2],1,(*ddx).npt[3]);

    for (ix=1; ix<=(*ddx).npt[1]; ix++) 
        {for (iy=1; iy<=(*ddx).npt[2]; iy++) 
             {for (iz=1; iz<=(*ddx).npt[3]; iz++) 
                 {(*ddx).data[ix][iy][iz] = (*dx2).data[ix][iy][iz]
                                           -(*dx1).data[ix][iy][iz];
                  }
             }
        }
    return 0;
}
/* ............................................................ */
/* write a CUBE to a file                                       */
/* I use a dummy atom just to have one                          */
/*                                                              */
void WriteCube(FILE *fp, Cube *dx, char const *info)
{
    int ix, iy, iz;

    fprintf(fp,"GMU Generated Cube File\n");
    fprintf(fp,"%s\n", info);
    fprintf(fp,"%5d %12.6f %12.6f %12.6f\n",1,(*dx).ori[1],(*dx).ori[2],(*dx).ori[3]);
    fprintf(fp,"%5d %12.6f %12.6f %12.6f\n", (*dx).npt[1],
            (*dx).delta[1][1], (*dx).delta[1][2], (*dx).delta[1][3]);
    fprintf(fp,"%5d %12.6f %12.6f %12.6f\n",(*dx).npt[2], 
            (*dx).delta[2][1], (*dx).delta[2][2], (*dx).delta[2][3]);
    fprintf(fp,"%5d %12.6f %12.6f %12.6f\n",(*dx).npt[3], 
            (*dx).delta[3][1], (*dx).delta[3][2], (*dx).delta[3][3]);
    fprintf(fp,"%5d %12.6f %12.6f %12.6f %12.6f\n",1, 1.0,
            0.0, 0.0, 0.0);

    for (ix=1; ix<=(*dx).npt[1]; ix++) 
        {for (iy=1; iy<=(*dx).npt[2]; iy++) 
             {for (iz=1; iz<=(*dx).npt[3]; iz++) 
                 {fprintf(fp," %13.5E",(*dx).data[ix][iy][iz]);
                  if ((iz-1) % 6 == 5) fprintf(fp,"\n"); 
                 }
              fprintf(fp,"\n");
             }
        }
}

