/* Matthias Ullmann, Berlin 19.07 1995			*/
/* nach D. Hoffmann					*/

#include <stdio.h>
#include <math.h>
#include "my-kabsch.h"
#include "mathematik.h"
#include "memutil.h"
#include "util.h"
#include "properties.h"

/*
"Kabsch" calculates a "best rotation" fit of coordinates set
x to the reference set y. The weights of the n atoms are 
taken from array w.
The centroids of the two sets are assumed to be located in
the origin.
The routine modifies the data set x: on successful termination
x contains the fitted set.
Literature: W. Kabsch, Acta Cryst (1976). A_32_, 922
            W. Kabsch, Acta Cryst (1978). A_34_, 827-828

Kabsch2 does the same, but takes a matrix RtR(or U) as an argument.
U can then be used in the calling routine.
*/
double Kabsch(int N, double **y, double **x, double *w)
{
  int n, mu1index, mu2index, mu3index;
  double **R, **RtR, **a, **b, **U, *mu, *e,
  sqrtmu1, sqrtmu2, sqrtmu3, factor, E0;

  R=dmatrix(1, 3, 1, 3);
  RtR=dmatrix(1, 3, 1, 3);
  a=dmatrix(1, 3, 1, 3);
  b=dmatrix(1, 3, 1, 3);
  mu=dvector(1, 3);
  e=dvector(1, 3);
  U=RtR;

  R[1][1]=R[1][2]=R[1][3]=0.0;
  R[2][1]=R[2][2]=R[2][3]=0.0;
  R[3][1]=R[3][2]=R[3][3]=0.0;

  E0=0;
  for (n=N; n; n--) {
    R[1][1] += w[n]*y[n][1]*x[n][1];
    R[1][2] += w[n]*y[n][1]*x[n][2];
    R[1][3] += w[n]*y[n][1]*x[n][3];

    R[2][1] += w[n]*y[n][2]*x[n][1];
    R[2][2] += w[n]*y[n][2]*x[n][2];
    R[2][3] += w[n]*y[n][2]*x[n][3];

    R[3][1] += w[n]*y[n][3]*x[n][1];
    R[3][2] += w[n]*y[n][3]*x[n][2];
    R[3][3] += w[n]*y[n][3]*x[n][3];

    E0 += w[n]*(
		x[n][1]*x[n][1] + x[n][2]*x[n][2] + x[n][3]*x[n][3]
		+y[n][1]*y[n][1] + y[n][2]*y[n][2] + y[n][3]*y[n][3]
		);
  }
  E0 *= 0.5;
  /*
     puts("R={");
     printf("{%.8lf, %.8lf, %.8lf},\n", R[1][1], R[1][2], R[1][3]);
     printf("{%.8lf, %.8lf, %.8lf},\n", R[2][1], R[2][2], R[2][3]);
     printf("{%.8lf, %.8lf, %.8lf}\n", R[3][1], R[3][2], R[3][3]);
     puts("}");
     */
  RtR[1][1]=R[1][1]*R[1][1]+R[2][1]*R[2][1]+R[3][1]*R[3][1];
  RtR[1][2]=R[1][1]*R[1][2]+R[2][1]*R[2][2]+R[3][1]*R[3][2];
  RtR[1][3]=R[1][1]*R[1][3]+R[2][1]*R[2][3]+R[3][1]*R[3][3];

  RtR[2][1]=R[1][2]*R[1][1]+R[2][2]*R[2][1]+R[3][2]*R[3][1];
  RtR[2][2]=R[1][2]*R[1][2]+R[2][2]*R[2][2]+R[3][2]*R[3][2];
  RtR[2][3]=R[1][2]*R[1][3]+R[2][2]*R[2][3]+R[3][2]*R[3][3];
	
  RtR[3][1]=R[1][3]*R[1][1]+R[2][3]*R[2][1]+R[3][3]*R[3][1];
  RtR[3][2]=R[1][3]*R[1][2]+R[2][3]*R[2][2]+R[3][3]*R[3][2];
  RtR[3][3]=R[1][3]*R[1][3]+R[2][3]*R[2][3]+R[3][3]*R[3][3];

  tred2(RtR, 3, mu, e);
  tqli(mu, e, 3, RtR);
	
  if (mu[1]>mu[2]) {
    if (mu[2]>mu[3]) {
      mu1index=1;
      mu2index=2;
      mu3index=3;
    }
    else
      if (mu[1]>mu[3]) {
	mu1index=1;
	mu2index=3;
	mu3index=2;
      }
      else {
	mu1index=3;
	mu2index=1;
	mu3index=2;
      }
  }
  else {
    if (mu[1]>mu[3]) {
      mu1index=2;
      mu2index=1;
      mu3index=3;
    }
    else
      if (mu[2]>mu[3]) {
	mu1index=2;
	mu2index=3;
	mu3index=1;
      }
      else {
	mu1index=3;
	mu2index=2;
	mu3index=1;
      }
  }
  /*
     printf("mu1=%lf\n mu2=%lf\n mu3=%lf\n", mu[mu1index], mu[mu2index], mu[mu3index]);
     */
  a[1][1]=RtR[1][mu1index];
  a[1][2]=RtR[2][mu1index];
  a[1][3]=RtR[3][mu1index];
		
  a[2][1]=RtR[1][mu2index];
  a[2][2]=RtR[2][mu2index];
  a[2][3]=RtR[3][mu2index];

  a[3][1]=a[1][2]*a[2][3]-a[1][3]*a[2][2];
  a[3][2]=a[1][3]*a[2][1]-a[1][1]*a[2][3];
  a[3][3]=a[1][1]*a[2][2]-a[1][2]*a[2][1];
  /*
     printf("a1={%lf,%lf,%lf}\n",a[1][1],a[1][2],a[1][3]);
     printf("a2={%lf,%lf,%lf}\n",a[2][1],a[2][2],a[2][3]);
     printf("a3={%lf,%lf,%lf}\n",a[3][1],a[3][2],a[3][3]);
     */
  sqrtmu1=sqrt(mu[mu1index]);
  factor=1.0/sqrtmu1;
  b[1][1]=(R[1][1]*a[1][1]+R[1][2]*a[1][2]+R[1][3]*a[1][3])*factor;
  b[1][2]=(R[2][1]*a[1][1]+R[2][2]*a[1][2]+R[2][3]*a[1][3])*factor;
  b[1][3]=(R[3][1]*a[1][1]+R[3][2]*a[1][2]+R[3][3]*a[1][3])*factor;
		
  sqrtmu2=sqrt(mu[mu2index]);
  factor=1.0/sqrtmu2;
  b[2][1]=(R[1][1]*a[2][1]+R[1][2]*a[2][2]+R[1][3]*a[2][3])*factor;
  b[2][2]=(R[2][1]*a[2][1]+R[2][2]*a[2][2]+R[2][3]*a[2][3])*factor;
  b[2][3]=(R[3][1]*a[2][1]+R[3][2]*a[2][2]+R[3][3]*a[2][3])*factor;
		
  sqrtmu3=sqrt(fabs(mu[mu3index]));
  b[3][1]=b[1][2]*b[2][3]-b[1][3]*b[2][2];
  b[3][2]=b[1][3]*b[2][1]-b[1][1]*b[2][3];
  b[3][3]=b[1][1]*b[2][2]-b[1][2]*b[2][1];

  /*
     factor=
     b[3][1]*(R[1][1]*a[3][1]+R[1][2]*a[3][2]+R[1][3]*a[3][3])
     +b[3][2]*(R[2][1]*a[3][1]+R[2][2]*a[3][2]+R[2][3]*a[3][3])
     +b[3][3]*(R[3][1]*a[3][1]+R[3][2]*a[3][2]+R[3][3]*a[3][3]);
     
     if (factor < 0.0) {
     b[3][1] *= -1.0;
     b[3][2] *= -1.0;
     b[3][3] *= -1.0;
     }
     */
  /*	
     printf("b1={%lf,%lf,%lf}\n",b[1][1],b[1][2],b[1][3]);
     printf("b2={%lf,%lf,%lf}\n",b[2][1],b[2][2],b[2][3]);
     printf("b3={%lf,%lf,%lf}\n",b[3][1],b[3][2],b[3][3]);
     */
  U[1][1]=b[1][1]*a[1][1]+b[2][1]*a[2][1]+b[3][1]*a[3][1];
  U[1][2]=b[1][1]*a[1][2]+b[2][1]*a[2][2]+b[3][1]*a[3][2];
  U[1][3]=b[1][1]*a[1][3]+b[2][1]*a[2][3]+b[3][1]*a[3][3];

  U[2][1]=b[1][2]*a[1][1]+b[2][2]*a[2][1]+b[3][2]*a[3][1];
  U[2][2]=b[1][2]*a[1][2]+b[2][2]*a[2][2]+b[3][2]*a[3][2];
  U[2][3]=b[1][2]*a[1][3]+b[2][2]*a[2][3]+b[3][2]*a[3][3];

  U[3][1]=b[1][3]*a[1][1]+b[2][3]*a[2][1]+b[3][3]*a[3][1];
  U[3][2]=b[1][3]*a[1][2]+b[2][3]*a[2][2]+b[3][3]*a[3][2];
  U[3][3]=b[1][3]*a[1][3]+b[2][3]*a[2][3]+b[3][3]*a[3][3];
  /*
     puts("U={");
     printf("{%.8lf, %.8lf, %.8lf},\n", U[1][1], U[1][2], U[1][3]);
     printf("{%.8lf, %.8lf, %.8lf},\n", U[2][1], U[2][2], U[2][3]);
     printf("{%.8lf, %.8lf, %.8lf}\n", U[3][1], U[3][2], U[3][3]);
     puts("}");
     printf("        detU=%lf\n", U[1][1]*U[2][2]*U[3][3]
     +U[1][2]*U[2][3]*U[3][1]
     +U[1][3]*U[2][1]*U[3][2]
     -U[3][1]*U[2][2]*U[1][3]
     -U[3][2]*U[2][3]*U[1][1]
     -U[3][3]*U[2][1]*U[1][2]);
     */

  for (n=N; n; n--) {
    e[1]=U[1][1]*x[n][1]+U[1][2]*x[n][2]+U[1][3]*x[n][3];
    e[2]=U[2][1]*x[n][1]+U[2][2]*x[n][2]+U[2][3]*x[n][3];
    e[3]=U[3][1]*x[n][1]+U[3][2]*x[n][2]+U[3][3]*x[n][3];
    x[n][1]=e[1];
    x[n][2]=e[2];
    x[n][3]=e[3];
  }
  return E0-sqrtmu1-sqrtmu2-sqrtmu3;
}






/*..............................................................*/
/* U is a pointer to a matrix[1-3][1-3]*/

double my_kabsch(int N, double **y, double **x, double *w, double **U)
{
  int n, mu1index, mu2index, mu3index;
  double sqrtmu1, sqrtmu2, sqrtmu3, factor, E0;
  double R[4][4], a[4][4], b[4][4], mu[4], e[4];

/*
  double **R, **a, **b, *mu, *e;
  R=dmatrix(1, 3, 1, 3);
  a=dmatrix(1, 3, 1, 3);
  b=dmatrix(1, 3, 1, 3);
  mu=dvector(1, 3);
  e=dvector(1, 3);
*/

  R[1][1]=R[1][2]=R[1][3]=0.0;
  R[2][1]=R[2][2]=R[2][3]=0.0;
  R[3][1]=R[3][2]=R[3][3]=0.0;

  E0=0;
  for (n=N; n; n--) {
    R[1][1] += w[n]*y[n][1]*x[n][1];
    R[1][2] += w[n]*y[n][1]*x[n][2];
    R[1][3] += w[n]*y[n][1]*x[n][3];

    R[2][1] += w[n]*y[n][2]*x[n][1];
    R[2][2] += w[n]*y[n][2]*x[n][2];
    R[2][3] += w[n]*y[n][2]*x[n][3];

    R[3][1] += w[n]*y[n][3]*x[n][1];
    R[3][2] += w[n]*y[n][3]*x[n][2];
    R[3][3] += w[n]*y[n][3]*x[n][3];

    E0 += w[n]*(
		x[n][1]*x[n][1] + x[n][2]*x[n][2] + x[n][3]*x[n][3]
		+y[n][1]*y[n][1] + y[n][2]*y[n][2] + y[n][3]*y[n][3]
		);
  }
  E0 *= 0.5;
  /*
     puts("R={");
     printf("{%.8lf, %.8lf, %.8lf},\n", R[1][1], R[1][2], R[1][3]);
     printf("{%.8lf, %.8lf, %.8lf},\n", R[2][1], R[2][2], R[2][3]);
     printf("{%.8lf, %.8lf, %.8lf}\n", R[3][1], R[3][2], R[3][3]);
     puts("}");
     */
  U[1][1]=R[1][1]*R[1][1]+R[2][1]*R[2][1]+R[3][1]*R[3][1];
  U[1][2]=R[1][1]*R[1][2]+R[2][1]*R[2][2]+R[3][1]*R[3][2];
  U[1][3]=R[1][1]*R[1][3]+R[2][1]*R[2][3]+R[3][1]*R[3][3];

  U[2][1]=R[1][2]*R[1][1]+R[2][2]*R[2][1]+R[3][2]*R[3][1];
  U[2][2]=R[1][2]*R[1][2]+R[2][2]*R[2][2]+R[3][2]*R[3][2];
  U[2][3]=R[1][2]*R[1][3]+R[2][2]*R[2][3]+R[3][2]*R[3][3];
	
  U[3][1]=R[1][3]*R[1][1]+R[2][3]*R[2][1]+R[3][3]*R[3][1];
  U[3][2]=R[1][3]*R[1][2]+R[2][3]*R[2][2]+R[3][3]*R[3][2];
  U[3][3]=R[1][3]*R[1][3]+R[2][3]*R[2][3]+R[3][3]*R[3][3];

  tred2(U, 3, mu, e);
  tqli(mu, e, 3, U);
	
  if (mu[1]>mu[2]) {
    if (mu[2]>mu[3]) {
      mu1index=1;
      mu2index=2;
      mu3index=3;
    }
    else
      if (mu[1]>mu[3]) {
	mu1index=1;
	mu2index=3;
	mu3index=2;
      }
      else {
	mu1index=3;
	mu2index=1;
	mu3index=2;
      }
  }
  else {
    if (mu[1]>mu[3]) {
      mu1index=2;
      mu2index=1;
      mu3index=3;
    }
    else
      if (mu[2]>mu[3]) {
	mu1index=2;
	mu2index=3;
	mu3index=1;
      }
      else {
	mu1index=3;
	mu2index=2;
	mu3index=1;
      }
  }
  /*
     printf("mu1=%lf\n mu2=%lf\n mu3=%lf\n", mu[mu1index], mu[mu2index], mu[mu3index]);
     */
  a[1][1]=U[1][mu1index];
  a[1][2]=U[2][mu1index];
  a[1][3]=U[3][mu1index];
		
  a[2][1]=U[1][mu2index];
  a[2][2]=U[2][mu2index];
  a[2][3]=U[3][mu2index];

  a[3][1]=a[1][2]*a[2][3]-a[1][3]*a[2][2];
  a[3][2]=a[1][3]*a[2][1]-a[1][1]*a[2][3];
  a[3][3]=a[1][1]*a[2][2]-a[1][2]*a[2][1];
  /*
     printf("a1={%lf,%lf,%lf}\n",a[1][1],a[1][2],a[1][3]);
     printf("a2={%lf,%lf,%lf}\n",a[2][1],a[2][2],a[2][3]);
     printf("a3={%lf,%lf,%lf}\n",a[3][1],a[3][2],a[3][3]);
     */
  sqrtmu1=sqrt(mu[mu1index]);
  factor=1.0/sqrtmu1;
  b[1][1]=(R[1][1]*a[1][1]+R[1][2]*a[1][2]+R[1][3]*a[1][3])*factor;
  b[1][2]=(R[2][1]*a[1][1]+R[2][2]*a[1][2]+R[2][3]*a[1][3])*factor;
  b[1][3]=(R[3][1]*a[1][1]+R[3][2]*a[1][2]+R[3][3]*a[1][3])*factor;
		
  sqrtmu2=sqrt(mu[mu2index]);
  factor=1.0/sqrtmu2;
  b[2][1]=(R[1][1]*a[2][1]+R[1][2]*a[2][2]+R[1][3]*a[2][3])*factor;
  b[2][2]=(R[2][1]*a[2][1]+R[2][2]*a[2][2]+R[2][3]*a[2][3])*factor;
  b[2][3]=(R[3][1]*a[2][1]+R[3][2]*a[2][2]+R[3][3]*a[2][3])*factor;
		
  sqrtmu3=sqrt(fabs(mu[mu3index]));
  b[3][1]=b[1][2]*b[2][3]-b[1][3]*b[2][2];
  b[3][2]=b[1][3]*b[2][1]-b[1][1]*b[2][3];
  b[3][3]=b[1][1]*b[2][2]-b[1][2]*b[2][1];

  /*
     factor=
     b[3][1]*(R[1][1]*a[3][1]+R[1][2]*a[3][2]+R[1][3]*a[3][3])
     +b[3][2]*(R[2][1]*a[3][1]+R[2][2]*a[3][2]+R[2][3]*a[3][3])
     +b[3][3]*(R[3][1]*a[3][1]+R[3][2]*a[3][2]+R[3][3]*a[3][3]);
     
     if (factor < 0.0) {
     b[3][1] *= -1.0;
     b[3][2] *= -1.0;
     b[3][3] *= -1.0;
     }
     */
  /*	
     printf("b1={%lf,%lf,%lf}\n",b[1][1],b[1][2],b[1][3]);
     printf("b2={%lf,%lf,%lf}\n",b[2][1],b[2][2],b[2][3]);
     printf("b3={%lf,%lf,%lf}\n",b[3][1],b[3][2],b[3][3]);
     */
  U[1][1]=b[1][1]*a[1][1]+b[2][1]*a[2][1]+b[3][1]*a[3][1];
  U[1][2]=b[1][1]*a[1][2]+b[2][1]*a[2][2]+b[3][1]*a[3][2];
  U[1][3]=b[1][1]*a[1][3]+b[2][1]*a[2][3]+b[3][1]*a[3][3];

  U[2][1]=b[1][2]*a[1][1]+b[2][2]*a[2][1]+b[3][2]*a[3][1];
  U[2][2]=b[1][2]*a[1][2]+b[2][2]*a[2][2]+b[3][2]*a[3][2];
  U[2][3]=b[1][2]*a[1][3]+b[2][2]*a[2][3]+b[3][2]*a[3][3];

  U[3][1]=b[1][3]*a[1][1]+b[2][3]*a[2][1]+b[3][3]*a[3][1];
  U[3][2]=b[1][3]*a[1][2]+b[2][3]*a[2][2]+b[3][3]*a[3][2];
  U[3][3]=b[1][3]*a[1][3]+b[2][3]*a[2][3]+b[3][3]*a[3][3];
  /* 
     puts("U={");
     printf("{%.8lf, %.8lf, %.8lf},\n", U[1][1], U[1][2], U[1][3]);
     printf("{%.8lf, %.8lf, %.8lf},\n", U[2][1], U[2][2], U[2][3]);
     printf("{%.8lf, %.8lf, %.8lf}\n", U[3][1], U[3][2], U[3][3]);
     puts("}");
     printf("        detU=%lf\n", U[1][1]*U[2][2]*U[3][3]
     +U[1][2]*U[2][3]*U[3][1]
     +U[1][3]*U[2][1]*U[3][2]
     -U[3][1]*U[2][2]*U[1][3]
     -U[3][2]*U[2][3]*U[1][1]
     -U[3][3]*U[2][1]*U[1][2]);
     */

/*  Auscommentiert - M. Ullmann Hier findet die Rotation statt! 
  for (n=N; n; n--) {
    e[1]=U[1][1]*x[n][1]+U[1][2]*x[n][2]+U[1][3]*x[n][3];
    e[2]=U[2][1]*x[n][1]+U[2][2]*x[n][2]+U[2][3]*x[n][3];
    e[3]=U[3][1]*x[n][1]+U[3][2]*x[n][2]+U[3][3]*x[n][3];
    x[n][1]=e[1];
    x[n][2]=e[2];
    x[n][3]=e[3];
  }*/

  return E0-sqrtmu1-sqrtmu2-sqrtmu3;
}


/*........minmax......................................................*/
double my_kabsch2(int N, double **y, double **x, double **U, double *trans)
{
  int n, mu1index, mu2index, mu3index;
  double **R, **a, **b, *mu, *e,
  sqrtmu1, sqrtmu2, sqrtmu3, factor, E0;

  R=dmatrix(1, 3, 1, 3);
  a=dmatrix(1, 3, 1, 3);
  b=dmatrix(1, 3, 1, 3);
  mu=dvector(1, 3);
  e=dvector(1, 3);

  R[1][1]=R[1][2]=R[1][3]=0.0;
  R[2][1]=R[2][2]=R[2][3]=0.0;
  R[3][1]=R[3][2]=R[3][3]=0.0;

  E0=0;
  for (n=N; n; n--) {
    R[1][1] += y[n][1]*x[n][1];
    R[1][2] += y[n][1]*x[n][2];
    R[1][3] += y[n][1]*x[n][3];

    R[2][1] += y[n][2]*x[n][1];
    R[2][2] += y[n][2]*x[n][2];
    R[2][3] += y[n][2]*x[n][3];

    R[3][1] += y[n][3]*x[n][1];
    R[3][2] += y[n][3]*x[n][2];
    R[3][3] += y[n][3]*x[n][3];

    E0 += x[n][1]*x[n][1] + x[n][2]*x[n][2] + x[n][3]*x[n][3]
	  +y[n][1]*y[n][1] + y[n][2]*y[n][2] + y[n][3]*y[n][3];
  }
  E0 *= 0.5;
  /*
     puts("R={");
     printf("{%.8lf, %.8lf, %.8lf},\n", R[1][1], R[1][2], R[1][3]);
     printf("{%.8lf, %.8lf, %.8lf},\n", R[2][1], R[2][2], R[2][3]);
     printf("{%.8lf, %.8lf, %.8lf}\n", R[3][1], R[3][2], R[3][3]);
     puts("}");
     */
  U[1][1]=R[1][1]*R[1][1]+R[2][1]*R[2][1]+R[3][1]*R[3][1];
  U[1][2]=R[1][1]*R[1][2]+R[2][1]*R[2][2]+R[3][1]*R[3][2];
  U[1][3]=R[1][1]*R[1][3]+R[2][1]*R[2][3]+R[3][1]*R[3][3];

  U[2][1]=R[1][2]*R[1][1]+R[2][2]*R[2][1]+R[3][2]*R[3][1];
  U[2][2]=R[1][2]*R[1][2]+R[2][2]*R[2][2]+R[3][2]*R[3][2];
  U[2][3]=R[1][2]*R[1][3]+R[2][2]*R[2][3]+R[3][2]*R[3][3];
	
  U[3][1]=R[1][3]*R[1][1]+R[2][3]*R[2][1]+R[3][3]*R[3][1];
  U[3][2]=R[1][3]*R[1][2]+R[2][3]*R[2][2]+R[3][3]*R[3][2];
  U[3][3]=R[1][3]*R[1][3]+R[2][3]*R[2][3]+R[3][3]*R[3][3];

  tred2(U, 3, mu, e);
  tqli(mu, e, 3, U);
	
  if (mu[1]>mu[2]) {
    if (mu[2]>mu[3]) {
      mu1index=1;
      mu2index=2;
      mu3index=3;
    }
    else
      if (mu[1]>mu[3]) {
	mu1index=1;
	mu2index=3;
	mu3index=2;
      }
      else {
	mu1index=3;
	mu2index=1;
	mu3index=2;
      }
  }
  else {
    if (mu[1]>mu[3]) {
      mu1index=2;
      mu2index=1;
      mu3index=3;
    }
    else
      if (mu[2]>mu[3]) {
	mu1index=2;
	mu2index=3;
	mu3index=1;
      }
      else {
	mu1index=3;
	mu2index=2;
	mu3index=1;
      }
  }
  /*
     printf("mu1=%lf\n mu2=%lf\n mu3=%lf\n", mu[mu1index], mu[mu2index], mu[mu3index]);
     */
  a[1][1]=U[1][mu1index];
  a[1][2]=U[2][mu1index];
  a[1][3]=U[3][mu1index];
		
  a[2][1]=U[1][mu2index];
  a[2][2]=U[2][mu2index];
  a[2][3]=U[3][mu2index];

  a[3][1]=a[1][2]*a[2][3]-a[1][3]*a[2][2];
  a[3][2]=a[1][3]*a[2][1]-a[1][1]*a[2][3];
  a[3][3]=a[1][1]*a[2][2]-a[1][2]*a[2][1];
  /*
     printf("a1={%lf,%lf,%lf}\n",a[1][1],a[1][2],a[1][3]);
     printf("a2={%lf,%lf,%lf}\n",a[2][1],a[2][2],a[2][3]);
     printf("a3={%lf,%lf,%lf}\n",a[3][1],a[3][2],a[3][3]);
     */
  sqrtmu1=sqrt(mu[mu1index]);
  factor=1.0/sqrtmu1;
  b[1][1]=(R[1][1]*a[1][1]+R[1][2]*a[1][2]+R[1][3]*a[1][3])*factor;
  b[1][2]=(R[2][1]*a[1][1]+R[2][2]*a[1][2]+R[2][3]*a[1][3])*factor;
  b[1][3]=(R[3][1]*a[1][1]+R[3][2]*a[1][2]+R[3][3]*a[1][3])*factor;
		
  sqrtmu2=sqrt(mu[mu2index]);
  factor=1.0/sqrtmu2;
  b[2][1]=(R[1][1]*a[2][1]+R[1][2]*a[2][2]+R[1][3]*a[2][3])*factor;
  b[2][2]=(R[2][1]*a[2][1]+R[2][2]*a[2][2]+R[2][3]*a[2][3])*factor;
  b[2][3]=(R[3][1]*a[2][1]+R[3][2]*a[2][2]+R[3][3]*a[2][3])*factor;
		
  sqrtmu3=sqrt(fabs(mu[mu3index]));
  b[3][1]=b[1][2]*b[2][3]-b[1][3]*b[2][2];
  b[3][2]=b[1][3]*b[2][1]-b[1][1]*b[2][3];
  b[3][3]=b[1][1]*b[2][2]-b[1][2]*b[2][1];

  /*
     factor=
     b[3][1]*(R[1][1]*a[3][1]+R[1][2]*a[3][2]+R[1][3]*a[3][3])
     +b[3][2]*(R[2][1]*a[3][1]+R[2][2]*a[3][2]+R[2][3]*a[3][3])
     +b[3][3]*(R[3][1]*a[3][1]+R[3][2]*a[3][2]+R[3][3]*a[3][3]);
     
     if (factor < 0.0) {
     b[3][1] *= -1.0;
     b[3][2] *= -1.0;
     b[3][3] *= -1.0;
     }
     */
  /*	
     printf("b1={%lf,%lf,%lf}\n",b[1][1],b[1][2],b[1][3]);
     printf("b2={%lf,%lf,%lf}\n",b[2][1],b[2][2],b[2][3]);
     printf("b3={%lf,%lf,%lf}\n",b[3][1],b[3][2],b[3][3]);
     */
  U[1][1]=b[1][1]*a[1][1]+b[2][1]*a[2][1]+b[3][1]*a[3][1];
  U[1][2]=b[1][1]*a[1][2]+b[2][1]*a[2][2]+b[3][1]*a[3][2];
  U[1][3]=b[1][1]*a[1][3]+b[2][1]*a[2][3]+b[3][1]*a[3][3];

  U[2][1]=b[1][2]*a[1][1]+b[2][2]*a[2][1]+b[3][2]*a[3][1];
  U[2][2]=b[1][2]*a[1][2]+b[2][2]*a[2][2]+b[3][2]*a[3][2];
  U[2][3]=b[1][2]*a[1][3]+b[2][2]*a[2][3]+b[3][2]*a[3][3];

  U[3][1]=b[1][3]*a[1][1]+b[2][3]*a[2][1]+b[3][3]*a[3][1];
  U[3][2]=b[1][3]*a[1][2]+b[2][3]*a[2][2]+b[3][3]*a[3][2];
  U[3][3]=b[1][3]*a[1][3]+b[2][3]*a[2][3]+b[3][3]*a[3][3];
  /* */
     puts("U={");
     printf("{%.8f, %.8f, %.8f},\n", U[1][1], U[1][2], U[1][3]);
     printf("{%.8f, %.8f, %.8f},\n", U[2][1], U[2][2], U[2][3]);
     printf("{%.8f, %.8f, %.8f}\n", U[3][1], U[3][2], U[3][3]);
     puts("}");
     printf("        detU=%f\n", U[1][1]*U[2][2]*U[3][3]
     +U[1][2]*U[2][3]*U[3][1]
     +U[1][3]*U[2][1]*U[3][2]
     -U[3][1]*U[2][2]*U[1][3]
     -U[3][2]*U[2][3]*U[1][1]
     -U[3][3]*U[2][1]*U[1][2]);
   /*  */

  for (n=N; n; n--) {
    e[1]=U[1][1]*x[n][1]+U[1][2]*x[n][2]+U[1][3]*x[n][3];
    e[2]=U[2][1]*x[n][1]+U[2][2]*x[n][2]+U[2][3]*x[n][3];
    e[3]=U[3][1]*x[n][1]+U[3][2]*x[n][2]+U[3][3]*x[n][3];
    x[n][1]=e[1];
    x[n][2]=e[2];
    x[n][3]=e[3];
  }
  return E0-sqrtmu1-sqrtmu2-sqrtmu3;
}

/*..............................................................*/
/* Procedure creates a rotation matrix and a translation vector */
/* in a	way to get the best fit of xyz1 with xyz2. The trans-	*/
/* lation vector and the rotation matrix is given back to the   */
/* calling subroutine.						*/
/*								*/
/* xyz2 is taken as reference!!!!				*/
/*								*/
/* in the calling subroutine, the following commands must be	*/
/* used in order to get a Kabsching:				*/
/*								*/
/*  kabsch(rot_matrix, center, trans, protein1.xyz, 		*/
/*         protein2.xyz, protein1.number_of_atoms, weight);	*/
/*  rotate(protein1.xyz, protein1.number_of_atoms, center, 	*/
/*	   rot_matrix);						*/
/*  translation(protein1.xyz, protein1.number_of_atoms,trans);  */
/*								*/

double kabsch(double **rot_matrix, double *center1, double *trans,
	      double **xyz1, double **xyz2, int atoms, double *weight)
{
double center2[4], trans1[4], trans2[4];
double rms_value;

center(xyz1, atoms, center1);
center(xyz2, atoms, center2);

trans1[1]= -center1[1]; trans1[2]= -center1[2]; trans1[3]= -center1[3];
trans2[1]= -center2[1]; trans2[2]= -center2[2]; trans2[3]= -center2[3];

translation(xyz1, atoms, trans1);
translation(xyz2, atoms, trans2);

trans[1]= trans1[1] - trans2[1];
trans[2]= trans1[2] - trans2[2];
trans[3]= trans1[3] - trans2[3];

rms_value=my_kabsch(atoms, xyz2, xyz1, weight, rot_matrix);


translation(xyz1, atoms, center1);
translation(xyz2, atoms, center2);

return rms_value;
}  
/*..............................................................*/

void tqli(double d[], double e[], int n, double **z)
{
	int m,l,iter,i,k;
	double s,r,p,g,f,dd,c,b;

	for (i=2;i<=n;i++) e[i-1]=e[i];
	e[n]=0.0;
	for (l=1;l<=n;l++) {
		iter=0;
		do {
			for (m=l;m<=n-1;m++) {
				dd=fabs(d[m])+fabs(d[m+1]);
				if ((double)(fabs(e[m])+dd) == dd) break;
			}
			if (m != l) {
				if (iter++ == 30) memerror("Too many iterations in tqli");
				g=(d[l+1]-d[l])/(2.0*e[l]);
				r=pythag(g,1.0);
				g=d[m]-d[l]+e[l]/(g+SIGN(r,g));
				s=c=1.0;
				p=0.0;
				for (i=m-1;i>=l;i--) {
					f=s*e[i];
					b=c*e[i];
					e[i+1]=(r=pythag(f,g));
					if (r == 0.0) {
						d[i+1] -= p;
						e[m]=0.0;
						break;
					}
					s=f/r;
					c=g/r;
					g=d[i+1]-p;
					r=(d[i]-g)*s+2.0*c*b;
					d[i+1]=g+(p=s*r);
					g=c*r-b;
					for (k=1;k<=n;k++) {
						f=z[k][i+1];
						z[k][i+1]=s*z[k][i]+c*f;
						z[k][i]=c*z[k][i]-s*f;
					}
				}
				if (r == 0.0 && i >= l) continue;
				d[l] -= p;
				e[l]=g;
				e[m]=0.0;
			}
		} while (m != l);
	}
}
/*..............................................................*/
void tred2(double **a, int n, double d[], double e[])
{
	int l,k,j,i;
	double scale,hh,h,g,f;

	for (i=n;i>=2;i--) {
		l=i-1;
		h=scale=0.0;
		if (l > 1) {
			for (k=1;k<=l;k++)
				scale += fabs(a[i][k]);
			if (scale == 0.0)
				e[i]=a[i][l];
			else {
				for (k=1;k<=l;k++) {
					a[i][k] /= scale;
					h += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g=(f >= 0.0 ? -sqrt(h) : sqrt(h));
				e[i]=scale*g;
				h -= f*g;
				a[i][l]=f-g;
				f=0.0;
				for (j=1;j<=l;j++) {
					a[j][i]=a[i][j]/h;
					g=0.0;
					for (k=1;k<=j;k++)
						g += a[j][k]*a[i][k];
					for (k=j+1;k<=l;k++)
						g += a[k][j]*a[i][k];
					e[j]=g/h;
					f += e[j]*a[i][j];
				}
				hh=f/(h+h);
				for (j=1;j<=l;j++) {
					f=a[i][j];
					e[j]=g=e[j]-hh*f;
					for (k=1;k<=j;k++)
						a[j][k] -= (f*e[k]+g*a[i][k]);
				}
			}
		} else
			e[i]=a[i][l];
		d[i]=h;
	}
	d[1]=0.0;
	e[1]=0.0;
	/* Contents of this loop can be omitted if eigendvectors not
			wanted except for statement d[i]=a[i][i]; */
	for (i=1;i<=n;i++) {
		l=i-1;
		if (d[i]) {
			for (j=1;j<=l;j++) {
				g=0.0;
				for (k=1;k<=l;k++)
					g += a[i][k]*a[k][j];
				for (k=1;k<=l;k++)
					a[k][j] -= g*a[k][i];
			}
		}
		d[i]=a[i][i];
		a[i][i]=1.0;
		for (j=1;j<=l;j++) a[j][i]=a[i][j]=0.0;
	}
}
/*..............................................................*/
double pythag(double a, double b)
{
	double absa,absb;
	absa=fabs(a);
	absb=fabs(b);
	if (absa > absb) return absa*sqrt(1.0+Sqr(absb/absa));
	else return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+Sqr(absa/absb)));
}




/****************** E X A M P L E *****************************
/ Program to superimpose two structure to each other			/ 
/ Matthias Ullmann, Berlin 08.08.1995					/
/ The program fittest two pdb-formated files to each other using the	/
/ algorithm form Kabsch.						/
/ Literature: W. Kabsch, Acta Cryst (1976). A_32_, 922			/
/             W. Kabsch, Acta Cryst (1978). A_34_, 827-828		/

/!!!!!!!!!!!!!!!!!!!! A T T E N T I O N !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!/
/ The mass assignment is not correct					/

/ to compile:
output-Aendern - HIER

cc superimp.c my-kabsch.o ~/C/LIB/util.o ~/C/LIB/nrutil.o ~/C/LIB/mathematik.o ~/C/LIB/pdb_atom.o ~/C/LIB/muf_atom.o -lm -o ~/BIN/superimp

/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include "my-kabsch.h"
#include "util_c.h"
#include "nrutil.h"
#include "konstant.h"
#include "muf_atom.h"
#include "pdb_atom.h"
#include "mathematik.h"

double CenterOfMass(double **coord, double *mass, int number, double *masscenter);
void TranslateCoordinates(int N, double **Coors, double *vector) ;

main(int argc, char *argv[])
{
FILE *fpin, *fpref, *fpfit;
char buffer[MAXSTRING], help[MAXSTRING];
char ref_file[MAXSTRING], fit_file[MAXSTRING];
int i,j;
double RefCenter[4], FitCenter[4], e[4], trans[4];

int no_atoms_to_fit;		/ number of atoms to fit 		/
double *weight, **ref, **fit;	/ ref, fit and weight-vector for the fit /
double *mass_ref, *mass_fit;	/ mass vector /
int *atoms_no_ref;		/ atoms numbers of the ref.-struct. 	/
int *atoms_no_fit;		/ atoms numbers of the fit.-struct. 	/
PDB_atom pdb_record;		/ help-record to read the pdb-file  	/
MUF_atom muf_ref[MAXATOMS*MAXAMINOACIDS];	
				/ MUF-Struture for the ref structure 	/
MUF_atom muf_fit[MAXATOMS*MAXAMINOACIDS];	
				/ MUF-Struture for the fit structure 	/
int NoRef, NoFit; 		/ Number of atoms in the structures	/
double **KabschMatrix;

/ ---------------------------------------------------------- /
/ Code /

if (argc!=2)
    {usage("superimp <input-file>", argc, 2);
    }

/ ---------------------------------------------------------- /
/ reading the input file and opening of the coordinate files /
/ and allocation of memory for the int-vectors		      /
fpin=file_open(argv[1],"r");
while (fgets(buffer, MAXSTRING, fpin) != NULL)
  {if (strncmp(buffer,"!",1))
	{sscanf(buffer,"%d",&no_atoms_to_fit);
	 fgets(buffer, MAXSTRING, fpin);
	 sscanf(buffer,"%s",fit_file);
	 fpfit=file_open(fit_file,"r");
	 fgets(buffer, MAXSTRING, fpin);
	 sscanf(buffer,"%s",ref_file);
	 fpref=file_open(ref_file,"r");
	 break;
	}
  }
fit=dmatrix(1,no_atoms_to_fit,1,3);
ref=dmatrix(1,no_atoms_to_fit,1,3);
KabschMatrix=dmatrix(1,3,1,3);
weight=dvector(1,no_atoms_to_fit);
mass_ref=dvector(1,no_atoms_to_fit);
mass_fit=dvector(1,no_atoms_to_fit);
atoms_no_ref=ivector(1,no_atoms_to_fit);
atoms_no_fit=ivector(1,no_atoms_to_fit);
i=0;
while (fgets(buffer, MAXSTRING, fpin) != NULL)
  {if (strncmp(buffer,"!",1))
	{i++;
	 sscanf(buffer,"%d %d %lf",&atoms_no_fit[i],&atoms_no_ref[i],&weight[i]);
	}
  }
/ ---------------------------------------------------------- /
/ Reading the coordinate Files				      /
/ 1. referenz structure				      /

i=0;
while (fgets(buffer, MAXSTRING, fpref) != NULL)
  {if ((!strncmp(buffer,"ATOM",4)) || (!strncmp(buffer,"HETATM",6)))
      {i++;
       read_neuPDBatom(buffer,&pdb_record);
       pdb2muf(&muf_ref[i],&pdb_record);
       muf_ref[i].mass=1.0;
       /fprint_MUFatom(stdout,&muf_ref[i]);/
      }
  }
NoRef=i;

/ 2. structure	to fit					      /
i=0;
while (fgets(buffer, MAXSTRING, fpfit) != NULL)
  {if ((!strncmp(buffer,"ATOM",4)) || (!strncmp(buffer,"HETATM",6)))
      {i++;
       read_neuPDBatom(buffer,&pdb_record);
       pdb2muf(&muf_fit[i],&pdb_record);
       muf_fit[i].mass=1.0;
      }
  }
NoFit=i;

fclose(fpref);
fclose(fpfit);

/ ---------------------------------------------------------- /
/ searching for the atoms for the fit			      /
/ 1. referenz structure				      /
fprintf(stderr,"Referenz-Atoms: %s\n",ref_file);
for(i=1; i<=no_atoms_to_fit; i++)
   {for(j=1; j<=NoRef; j++)
	{if (atoms_no_ref[i]==muf_ref[j].atomno)
	    {/HIER fprint_MUFatom(stderr,&muf_ref[j]);/
	     ref[i][1]=muf_ref[j].xyz[1];
	     ref[i][2]=muf_ref[j].xyz[2];
	     ref[i][3]=muf_ref[j].xyz[3];
	     mass_ref[i]=muf_ref[j].mass;
	    }
	}
   }
/ 2. structure	to fit					      /
fprintf(stderr,"\n\nAtoms to fit: %s\n",fit_file);
for(i=1; i<=no_atoms_to_fit; i++)
   {for(j=1; j<=NoFit; j++)
	{if (atoms_no_fit[i]==muf_fit[j].atomno)
	    {fit[i][1]=muf_fit[j].xyz[1];
	     fit[i][2]=muf_fit[j].xyz[2];
	     fit[i][3]=muf_fit[j].xyz[3];
	     mass_fit[i]=muf_fit[j].mass;
	    }
	}
   }


/ ------------------------ K A B S C H --------------------------------------- /
/ Calculation of the Center (of mass)	and transformations   			/
/ 1. referenz structure				      			/

/ The Center of mass become transfer to the coordinate origin /
CenterOfMass(ref, mass_ref,no_atoms_to_fit,RefCenter);
CenterOfMass(fit, mass_fit,no_atoms_to_fit,FitCenter);
trans[1]=-RefCenter[1];
trans[2]=-RefCenter[2];
trans[3]=-RefCenter[3];
TranslateCoordinates(no_atoms_to_fit, ref, trans);

trans[1]=-FitCenter[1];
trans[2]=-FitCenter[2];
trans[3]=-FitCenter[3];
TranslateCoordinates(no_atoms_to_fit, fit, trans);

my_kabsch(no_atoms_to_fit, ref, fit, weight, KabschMatrix);

trans[1]=-FitCenter[1];
trans[2]=-FitCenter[2];
trans[3]=-FitCenter[3];
translate(muf_fit, NoFit, trans);
for (i=1; i<=NoFit; i++) 
   {e[1]=KabschMatrix[1][1]*muf_fit[i].xyz[1]+KabschMatrix[1][2]*muf_fit[i].xyz[2]+KabschMatrix[1][3]*muf_fit[i].xyz[3];
    e[2]=KabschMatrix[2][1]*muf_fit[i].xyz[1]+KabschMatrix[2][2]*muf_fit[i].xyz[2]+KabschMatrix[2][3]*muf_fit[i].xyz[3];
    e[3]=KabschMatrix[3][1]*muf_fit[i].xyz[1]+KabschMatrix[3][2]*muf_fit[i].xyz[2]+KabschMatrix[3][3]*muf_fit[i].xyz[3];
    muf_fit[i].xyz[1]=e[1];
    muf_fit[i].xyz[2]=e[2];
    muf_fit[i].xyz[3]=e[3];
   }

trans[1]=RefCenter[1];
trans[2]=RefCenter[2];
trans[3]=RefCenter[3];
translate(muf_fit, NoFit, trans);

for (i=1; i<=NoFit; i++) 
   {muf2pdb(&pdb_record, &muf_fit[i]);
    fprint_neuPDBatom(stdout,&pdb_record);
   }

/------------------------ A U S G A B E ------------------------------/

fprintf(stderr,"\n\n\nRotationsmatrix:\n");
fprintf(stderr,"{%14.8lf, %14.8lf, %14.8lf},\n", KabschMatrix[1][1], KabschMatrix[1][2], KabschMatrix[1][3]);
fprintf(stderr,"{%14.8lf, %14.8lf, %14.8lf},\n", KabschMatrix[2][1], KabschMatrix[2][2], KabschMatrix[2][3]);
fprintf(stderr,"{%14.8lf, %14.8lf, %14.8lf}\n", KabschMatrix[3][1], KabschMatrix[3][2], KabschMatrix[3][3]);
fprintf(stderr,"\ndetU=%lf\n", KabschMatrix[1][1]*KabschMatrix[2][2]*KabschMatrix[3][3]
     +KabschMatrix[1][2]*KabschMatrix[2][3]*KabschMatrix[3][1]
     +KabschMatrix[1][3]*KabschMatrix[2][1]*KabschMatrix[3][2]
     -KabschMatrix[3][1]*KabschMatrix[2][2]*KabschMatrix[1][3]
     -KabschMatrix[3][2]*KabschMatrix[2][3]*KabschMatrix[1][1]
     -KabschMatrix[3][3]*KabschMatrix[2][1]*KabschMatrix[1][2]);

fprintf(stderr,"\nCenter of Mass of the Referenz-Atoms\n");
fprintf(stderr,"{%14.8lf, %14.8lf, %14.8lf},\n", RefCenter[1], RefCenter[2], RefCenter[3]);

fprintf(stderr,"\nCenter of Mass of the Atoms to fit\n");
fprintf(stderr,"{%14.8lf, %14.8lf, %14.8lf},\n", FitCenter[1], FitCenter[2], FitCenter[3]);

fprintf(stderr,"\n\nTranslations-Vektor\n");
fprintf(stderr,"{%14.8lf, %14.8lf, %14.8lf},\n", trans[1], trans[2], trans[3]);

free_dmatrix(ref, 1,no_atoms_to_fit,1,3);
free_dmatrix(fit, 1,no_atoms_to_fit,1,3);
free_dvector(mass_ref, 1,no_atoms_to_fit);
free_dvector(mass_fit, 1,no_atoms_to_fit);
free_dvector(weight, 1,no_atoms_to_fit);
free_ivector(atoms_no_ref, 1,no_atoms_to_fit);
free_ivector(atoms_no_fit, 1,no_atoms_to_fit);

fprintf(stderr,"\n\nDone\n");
}


/............................................................../
/ procedure calculates the center of mass.			/
/ all arrays have NR-convention				/
/ it returns the total mass.					/
double CenterOfMass(double **coord, double *mass, int number, double *masscenter)
{
double total_mass;
int i;
masscenter[1]=0.0;
masscenter[2]=0.0;
masscenter[3]=0.0;

total_mass=0.0;
for (i=1; i<=number; i++)
	{masscenter[1]=masscenter[1]+(mass[i]*coord[i][1]);
	 masscenter[2]=masscenter[2]+(mass[i]*coord[i][2]);
	 masscenter[3]=masscenter[3]+(mass[i]*coord[i][3]);
	 total_mass=total_mass+mass[i];
	}	
masscenter[1]=masscenter[1]/total_mass;
masscenter[2]=masscenter[2]/total_mass;
masscenter[3]=masscenter[3]/total_mass;


return total_mass;
}

/............................................................../
/ von Daniel							/
void TranslateCoordinates(int number, double **Coors, double *vector) 
{
int i;

for (i=1; i<=number; i++)
   {Coors[i][1] = Coors[i][1] + vector[1]; 
    Coors[i][2] = Coors[i][2] + vector[2]; 
    Coors[i][3] = Coors[i][3] + vector[3]; 
  }
  return;
}
        
****************** E X A M P L E *****************************/
