/* Matthias Ullmann 					*/
/* Berlin, 04.10.1995					*/
/* 							*/
/* Cell base neighbor list for Monte Carlo Simulations	*/
/* 							*/
/* Parameter suggestion:				*/
/* for cut on 10.0 and cutoff 12.0 a grid spacing of 	*/
/* 4.0 has been found to be ideal			*/
/* .....................................................*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "nb-list.h"
#include "util.h"
#include "memutil.h"
#include "mol.h"
#include "grid.h"

#define NR_END 1
/* .....................................................*/
/* prints infos about the grid                          */
void iListGrid_info(FILE *fpout, iListGrid *grid, char *info)
{
fprintf(fpout,SEPARATE); 
 fprintf(fpout,"%s\n\n", info); 
fprintf(fpout,"MIN: "); PRINT_iVEC(fpout,(*grid).imin);
fprintf(fpout,"MAX: "); PRINT_iVEC(fpout,(*grid).imax);
fprintf(fpout,"Spacing  : %8.3f\n", (*grid).spacing);
fprintf(fpout,"Inverse Spacing  : %8.3f\n", (*grid).inv_spacing);
fprintf(fpout,SEPARATE); 
}

/* .....................................................*/
/* Allocating a vector of iList with 			*/
/* subscript range v[nl..nh] 				*/
iList *iList_vector(long nl, long nh)
{
	iList *v;

	v=(iList *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(iList)));
	if (!v) ERROR("allocation failure in iList_vector()");
	return v-nl+NR_END;
}
/* .....................................................*/
/* Deallocating a vector of iList with 			*/
/* subscript range v[nl..nh] 				*/
void free_iList_vector(iList *v, long nl, long nh)
{
  if (nl>nh) memerror("Problem in free_iList_vector()");
if (v!=NULL)    free((FREE_ARG) (v+nl-NR_END));
}
/* .....................................................*/
/* Allocating a 3D tensor of iList			*/
/* with subscript range v[nl..nh][ncl..nch][ndl..ndh]	*/
iList ***iList_3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
{
long  i,j,nrow=nrh-nrl+1,ncol=nch-ncl+1,ndep=ndh-ndl+1;
iList  ***t;

	/* allocate pointers to pointers to rows */
	t=(iList ***) malloc((size_t)((nrow+NR_END)*sizeof(iList**)));
	if (!t) ERROR("allocation failure 1 in iList_3tensor()");
	t += NR_END;
	t -= nrl;

	/* allocate pointers to rows and set pointers to them */
	t[nrl]=(iList **) malloc((size_t)((nrow*ncol+NR_END)*sizeof(iList*)));
	if (!t[nrl]) ERROR("allocation failure 2 in iList_3tensor()");
	t[nrl] += NR_END;
	t[nrl] -= ncl;

	/* allocate rows and set pointers to them */
	
	t[nrl][ncl]=(iList *) malloc((size_t)((nrow*ncol*ndep+NR_END)*sizeof(iList)));
	if (!t[nrl][ncl]) ERROR("allocation failure 3 in iList_3tensor()");
	t[nrl][ncl] += NR_END;
	t[nrl][ncl] -= ndl;

	for(j=ncl+1;j<=nch;j++) t[nrl][j]=t[nrl][j-1]+ndep;
	for(i=nrl+1;i<=nrh;i++) {
		t[i]=t[i-1]+ncol;
		t[i][ncl]=t[i-1][ncl]+ncol*ndep;
		for(j=ncl+1;j<=nch;j++) t[i][j]=t[i][j-1]+ndep;
	}

	/* return pointer to array of pointers to rows */
	return t;
}

/* .....................................................*/
/* Deallocating a 3D tensor of iList			*/
/* with subscript range v[nl..nh][ncl..nch][ndl..ndh]	*/
void free_iList_3tensor(iList ***t, long nrl, long nrh, long ncl, 
                        long nch, long ndl, long ndh)
{
  if ((nrl>nrh)||(ncl>nch)||(ndl>ndh)) 
    {fprintf(stderr,"%ld-%ld; %ld-%ld; %ld-%ld\n",nrl,nrh,ncl,nch,ndl,ndh);
     memerror("Problem in free_iList_3tensor()");
    }
  if (t!=NULL)    
        {free((FREE_ARG) (t[nrl][ncl]+ndl-NR_END));
         free((FREE_ARG) (t[nrl]+ncl-NR_END));
         free((FREE_ARG) (t+nrl-NR_END));
        }
}
/* .....................................................*/
/* Initialzes a vector of iList of the 	length 'length'	*/
/* Initialization: *next =NULL				*/
/*		   *prev =NULL				*/
/*		   number=counting			*/
/*							*/
void init_iList_vector(iList *v, int length)
{
iList *help;
int i;

help=v;
for (i=1; i<=length; i++)
    {help++; /* first increse - NR-Convention Vectors!! */
     (*help).next  =NULL;
     (*help).prev  =help;
     (*help).number=i;
    }
}

/* .....................................................*/
/*  initialzes iListGrid Tensor component	        */
/* of the range 					*/
/* [imin[1]..imax[1]][imin[2]..imax[2]][imin[3]..imax[3]]*/
/* Initialization: *next =NULL				*/
/*							*/
/* The end and start point of the grid (i.e. imax[] and	*/
/* imin[]) as well as the spacing parameter must be 	*/
/* assigned before calling the function.		*/
/*							*/
void init_iListGrid(iListGrid *grid)
{
int i,j,k;
iList *list;
/* Testings */
if (((*grid).inv_spacing==0.0)||((*grid).spacing==0.0))
   {ERROR("Grid spacing is zero in  init_iListGrid()");
   }
if (((*grid).imax==NULL)||((*grid).imax==NULL))
   {ERROR("imax or imin not allocated in init_iListGrid()");
   }

/* Initialization */
for (i=(*grid).imin[1]; i<=(*grid).imax[1]; i++)
    for (j=(*grid).imin[2]; j<=(*grid).imax[2]; j++)
	for (k=(*grid).imin[3]; k<=(*grid).imax[3]; k++)
	    {list=&(*grid).root[i][j][k];
	     (*list).next  =NULL;
	     (*list).prev  =NULL;
	     (*list).number=0;
	    }
}

/* .................................................... */
/* Adds a atom to the list				*/
/*							*/
/* INV_SPACING should be assigned before calling this	*/
/* procedure						*/
/*							*/

void add_to_list(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3;
iList *help, *v;

/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {WARN("The number is zero in add_to_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3);
     ERROR("Point not in box in add_to_list().");
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     help=(*v).next;
     if (help!=NULL)
        {(*help).prev=list;
	}
     (*v).next=list;
     (*list).next=help;
     (*list).prev=v;
    }
}

/* .................................................... */
/* Removes a entry from the list and redirects 		*/
/* the pointers.					*/
/*							*/
void remove_from_list(iList *list)
{
iList *help;

help=(*list).prev;
(*help).next=(*list).next;

if ((*list).next!=NULL)
   {help=(*list).next;
    (*help).prev=(*list).prev;
   }
/* CHANGE - kann weggelassen werden */
(*list).prev=NULL;
(*list).next=NULL;
}
/* .................................................... */
/* Moves an atom from the old grid cell to a new	*/
/*							*/
/* an alternative but slightly slower version (two 	*/
/* function calls) would be:				*/
/*							*/
/*  remove_from_list(list);				*/
/*  add_to_list(grid, list, x);				*/
/*							*/
/* This should be used if some problem are assumed in	*/
/* nb-list updating					*/
/*							*/

void move_in_list(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3;
iList *help, *v;

/* first remove the atom from the list */
help=(*list).prev;
(*help).next=(*list).next;
if ((*list).next!=NULL)
   {help=(*list).next;
    (*help).prev=(*list).prev;
   }

/* now reput it in the list  */
/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {WARN("The number is zero in move_in_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {WARN("Point not in box in add_to_list().");
     fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3); 
     (*list).prev=NULL;
     (*list).next=NULL;
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     help=(*v).next;
     if (help!=NULL)
        {(*help).prev=list;
	}
     (*v).next=list;
     (*list).next=help;
     (*list).prev=v;
    }
}
/* .................................................... */
void move_in_list_test(iListGrid *grid, iList *list, double *x)
{
remove_from_list(list);
add_to_list(grid, list, x);
}
/* .................................................... */
/* Fills the list with the given atom coordinates	*/
/*							*/
void fill_iListGrid(iListGrid *grid, iList *list, double **xyz, int atoms)
{
int i;

for (i=1; i<=atoms; i++)
    {add_to_list(grid, &list[i], xyz[i]);
    }
}
/********************************************************/
/* From here on, these are procedures for single linked */
/* neighbor lists					*/
/* .................................................... */
/* Add a atom to a single-linked list			*/
/*							*/	   
void add_to_list2(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3;
iList *help, *v;

/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {WARN("The number is zero in add_to_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3);
     ERROR("Point not in box in add_to_list().");
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     help=(*v).next;
     (*v).next=list;
     (*list).next=help;
    }
}

/* .................................................... */
/* fills neighbor list with a single-linked list.	*/
/*							*/
void fill_iListGrid2(iListGrid *grid, iList *list, double **x,  int natoms)
{
int i,j,k;
iList *help;

/* Initialization */
for (i=(*grid).imin[1]; i<=(*grid).imax[1]; i++)
    for (j=(*grid).imin[2]; j<=(*grid).imax[2]; j++)
	for (k=(*grid).imin[3]; k<=(*grid).imax[3]; k++)
	    {help=&(*grid).root[i][j][k];
	     (*help).next  =NULL;
	    }

for (i=1; i<=natoms;i++)
    {add_to_list2(grid, &list[i], x[i]);
    }
}

/* .................................................... */
/* removes an atom from a single-linked list		*/
/*							*/
void remove_from_list2(iListGrid *grid, iList *list, double *x)
{
int t1, t2, t3,i;
iList *prev, *v;

/* map the atoms on the grid */
t1=(int)rint(x[1]*(*grid).inv_spacing);
t2=(int)rint(x[2]*(*grid).inv_spacing);
t3=(int)rint(x[3]*(*grid).inv_spacing);


if ((*list).number==0) 
   {WARN("The number is zero in add_to_list().");
   }

/* Testing if the atom is within the grid */
if (  (t1<(*grid).imin[1]) || (t1>(*grid).imax[1])
    ||(t2<(*grid).imin[2]) || (t2>(*grid).imax[2])
    ||(t3<(*grid).imin[3]) || (t3>(*grid).imax[3]))
    {fprintf(stderr,"%8.3f%8.3f%8.3f - %4d%4d%4d\n",
		     x[1], x[2], x[3], t1, t2, t3);
     ERROR("Point not in box in new_remove_from_list().");
    }
else{/* redirect the pointer */
     v=&(*grid).root[t1][t2][t3];
     if ((*v).next==NULL) 
	{ERROR("This should never happen in remove_from_list2().");
	}
     prev=v;
     v=(*v).next;
     i=(*v).number;
     while(i!=(*list).number)
	  {prev=v;
           v=(*v).next;
           i=(*v).number;
	   if (v==NULL) 
	      {ERROR("This should never happen in remove_from_list2().");
	      }
	  }
     (*prev).next=(*v).next;
    }
}

/*........................................................*/
/* This function assigns an pt to a grid cell             */
/*                                                        */
/* returns 0 if the point is within the grid otherwise 1  */
/*                                                        */
int pt_to_cell(int *ipt, double *pt, iListGrid *grd)
{

    /* map the pt on the grid */
    ipt[1]=(int)rint(pt[1]*(*grd).inv_spacing);
    ipt[2]=(int)rint(pt[2]*(*grd).inv_spacing);
    ipt[3]=(int)rint(pt[3]*(*grd).inv_spacing);

    /* fprintf(stderr,"%d  %d %d\n", ipt[1], ipt[2], ipt[3]); */

    if (  (ipt[1]<(*grd).imin[1]) || (ipt[1]>(*grd).imax[1])
        ||(ipt[2]<(*grd).imin[2]) || (ipt[2]>(*grd).imax[2])
        ||(ipt[3]<(*grd).imin[3]) || (ipt[3]>(*grd).imax[3]))
        {return 1;
        }
    else {return 0;}
}

/*........................................................*/
/* make a neighbor list for a protein                     */
/*                                                        */
void ProteinToGrid(iListGrid *grid, Mol *p, double maxrad)
{
  double min[4], max[4];
  iList *list;

  min_max_coord(min, max, (*p).xyz, (*p).natoms);
  max[1] = max[1] + maxrad;
  max[2] = max[2] + maxrad;
  max[3] = max[3] + maxrad;
  min[1] = min[1] - maxrad;
  min[2] = min[2] - maxrad;
  min[3] = min[3] - maxrad;

  /* Assigning the values for the grid */
  (*grid).inv_spacing=1.0/(*grid).spacing;
  (*grid).imax=ivector(1,3);
  (*grid).imin=ivector(1,3);
  (*grid).imax[1]=(int)ceil(max[1]*(*grid).inv_spacing);
  (*grid).imax[2]=(int)ceil(max[2]*(*grid).inv_spacing);
  (*grid).imax[3]=(int)ceil(max[3]*(*grid).inv_spacing);
		  
  (*grid).imin[1]=(int)floor(min[1]*(*grid).inv_spacing);
  (*grid).imin[2]=(int)floor(min[2]*(*grid).inv_spacing);
  (*grid).imin[3]=(int)floor(min[3]*(*grid).inv_spacing);

  /* make the grid spacing to odd integer, if they are not odd integer */
  odding_down((*grid).imin,3);
  odding_up((*grid).imax,3);

  list=iList_vector(1,(*p).natoms);
  init_iList_vector(list,(*p).natoms);
  (*grid).root=iList_3tensor((*grid).imin[1], (*grid).imax[1],
                             (*grid).imin[2], (*grid).imax[2],
                             (*grid).imin[3], (*grid).imax[3]);

  init_iListGrid(grid);
  fill_iListGrid(grid, list, (*p).xyz, (*p).natoms);
}

