/* G. Matthias Ullmann                                  */
/*                                                      */
/* .....................................................*/

#include <stdio.h>
#include "mathematik.h"
#include <math.h> 
#include "memutil.h"  

/* ......................................... ...............*/
/* Linear equation solution by Gauss-Jordan elimination,    */
/* equation (2.1.1) above. a[1..n][1..n] is the input 	    */
/* matrix. b[1..n] is input containing the right-hand side  */
/* vector. On output, a is replaced by its matrix inverse,  */
/* and b is replaced by the corresponding set of 	    */
/* solution vectors.                                        */

#define SWAP(a,b) {temp=(a);(a)=(b);(b)=temp;}              

void gaussj(double **a, int n, double *b)                                                           
{
  int *indxc,*indxr,*ipiv;
  int i,icol=0,irow=0,j,k,l,ll;
  double big,dum,pivinv,temp;

  indxc=ivector(1,n);
  indxr=ivector(1,n);
  ipiv=ivector(1,n);
  for (j=1;j<=n;j++) ipiv[j]=0;
  for (i=1;i<=n;i++) 
      {big=0.0;
       for (j=1;j<=n;j++)
	   if (ipiv[j] != 1)
                for (k=1;k<=n;k++) 
	            {if (ipiv[k] == 0) 
			{if (fabs(a[j][k]) >= big) 
			    {big=fabs(a[j][k]);
                             irow=j;
                             icol=k;
                            }
                         } 
                      else if (ipiv[k] > 1) memerror("gaussj: Singular Matrix-1");
                     }
             ++(ipiv[icol]);
             if (irow != icol)
	        {for (l=1;l<=n;l++) SWAP(a[irow][l],a[icol][l])
		 SWAP(b[irow],b[icol])
                }
                indxr[i]=irow;
                indxc[i]=icol;
                if (a[icol][icol] == 0.0) memerror("gaussj: Singular Matrix-2");
                pivinv=1.0/a[icol][icol];
                a[icol][icol]=1.0;
                for (l=1;l<=n;l++) a[icol][l] *= pivinv;
                b[icol] *= pivinv;
                for (ll=1;ll<=n;ll++)
                     if (ll != icol) 
		        {dum=a[ll][icol];
                         a[ll][icol]=0.0;
                         for (l=1;l<=n;l++) a[ll][l] -= a[icol][l]*dum;
                         b[ll] -= b[icol]*dum;
                        }
      }
  for (l=n;l>=1;l--) 
      {if (indxr[l] != indxc[l])
          for (k=1;k<=n;k++) SWAP(a[k][indxr[l]],a[k][indxc[l]]);
      }
  free_ivector(ipiv,1,n);
  free_ivector(indxr,1,n);
  free_ivector(indxc,1,n);
}
                                                                
#undef SWAP


/* ......................................... ...............*/
/* Solves the set of n linear equations AX = B. Here        */
/* a[1..n][1..n] is input, not as the matrix A but 	    */
/* rather as its LU decomposition, determined by the 	    */
/* routine ludcmp. indx[1..n] is input as the permutation   */
/* vector returned by ludcmp. b[1..n] is input as the 	    */
/* right-hand side vector B, and returns with the solution  */
/* vector X. a, n, and indx are not modi^Led by this 	    */
/* routine and can be left in place for successive calls    */
/* with diㄦent right-hand sides b. This routine takes into */
/* account the possibility that b will begin with many zero */
/* elements, so it is equationscient for use in matrix 	    */
/* inversion.                                               */
      
void lubksb(double **a, int n, int *indx, double *b)
{
        int i,ii=0,ip,j;
        double sum;

        for (i=1;i<=n;i++)
	    {ip=indx[i];
             sum=b[ip];
             b[ip]=b[i];
             if (ii)
                for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
             else if (sum) 
                  {ii=i;
                  }
             b[i]=sum;
            }
        for (i=n;i>=1;i--)
	    {sum=b[i];
             for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
             b[i]=sum/a[i][i];
            }
}

/* .....................................................*/
/* Given a matrix a[1..n][1..n], this routine replaces  */
/* it by the LU decomposition of a rowwise permutation  */
/* of itself. a and n are input. a is output, arranged  */
/* as in equation (2.3.14) above; indx[1..n] is an      */
/* output vector that records the row permutation       */
/* eected by the partial pivoting; d is output as 1     */
/* depending on whether the number of row interchanges  */
/* was even or odd, respectively. This routine is used  */
/* in combination with lubksb to solve linear equations */
/* or invert a matrix.                                  */	

#define TINY 1.0e-20;

void ludcmp(double **a, int n, int *indx, double *d)
{
        int i,imax=0,j,k;
        double big,dum,sum,temp;
        double *vv;

        vv=dvector(1,n);
        *d=1.0;
        for (i=1;i<=n;i++)
	    {big=0.0;
             for (j=1;j<=n;j++)
                 if ((temp=fabs(a[i][j])) > big) big=temp;
             if (big == 0.0) memerror("Singular matrix in routine ludcmp");
             vv[i]=1.0/big;
            }
        for (j=1;j<=n;j++)
	    {for (i=1;i<j;i++)
 	         {sum=a[i][j];
                  for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
                  a[i][j]=sum;
                 }
             big=0.0;
             for (i=j;i<=n;i++)
	         {sum=a[i][j];
                  for (k=1;k<j;k++)
                      sum -= a[i][k]*a[k][j];
                  a[i][j]=sum;
                  if ( (dum=vv[i]*fabs(sum)) >= big)
		     {big=dum;
                      imax=i;
                     }
                 }
             if (j != imax)
	        {for (k=1;k<=n;k++)
		     {dum=a[imax][k];
                      a[imax][k]=a[j][k];
                      a[j][k]=dum;
                     }
                 *d = -(*d);
                 vv[imax]=vv[j];
                }
             indx[j]=imax;
             if (a[j][j] == 0.0) a[j][j]=TINY;
             if (j != n)
	        {dum=1.0/(a[j][j]);
                 for (i=j+1;i<=n;i++) a[i][j] *= dum;
                }
	    }
        free_dvector(vv,1,n);
}
#undef TINY





