/* Matthias Ullmann, Berlin 10.10.1994					*/
/* Enthaelt einige Prozeduren zum generieren von Zufallszahlen.		*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "random.h"
#include "constants.h"


/*..............................................................................*/
/* It returns a random vector with components taken from a Gauss distribution	*/
/* with zero mean and a given deviation.					*/
void random_vector(double *vector, double deviation, long *seed)
{
vector[1]=deviation*RandomGaussDeviation(seed);
vector[2]=deviation*RandomGaussDeviation(seed);
vector[3]=deviation*RandomGaussDeviation(seed);
}


/*..............................................................................*/
/* Function returns normally distributed deviate with zero mean and unit 	*/
/* variance, using Ran as source of uniform deviates.				*/
/*                      							*/
/* The number idum should be a negative integer, to make the first if true.	*/

double RandomGaussDeviation(long *idum)
{
        double Ran(long *idum);
        static int iSet=0;
        static double gSet;
        double fac,rsq,v1,v2;

        if  (iSet == 0)
	    {do {v1=2.0*Ran(idum)-1.0;
                 v2=2.0*Ran(idum)-1.0;
                 rsq=v1*v1+v2*v2;
                } while (rsq >= 1.0 || rsq == 0.0);
             fac=sqrt(-2.0*log(rsq)/rsq);
             gSet=v1*fac;
             iSet=1;
             return v2*fac;
            }
	else {iSet=0;
              return gSet;
             }
}
/****************************************************************/
/* Metropolis algorithm						*/
/* ............................................................ */
/* metropolis returns '1' if the step is accepted		*/
/* delta_energy= energy_new - energy_old			*/
/*								*/
/* In order to avoid an underflow error in exp(), a constant was*/
/* defined in energy.h. No error message is issued.		*/
/*                                                              */
/* the energy argument should be in kT                          */

int Metropolis(double delta_energy, long *seed)
{
double help;
/* fprintf(stdout,"Delta Energy: %lf\n",delta_energy);*/
/*fprintf(stdout,"Delta Energy: %e\n",-delta_energy/(temp*R_GAS));*/


if (delta_energy < 0.0) 
   {return ACCEPTED; /* 1 */
   }
else {help=-delta_energy;
      if (help<TOO_SMALL)
	 {return REJECTED;
	 }
      else {if ((Ran(seed)) < exp(help))
	       {return ACCEPTED; /* 1 */;
	       }
	    else {return REJECTED;
		 }
	   }
     }
}
/* ............................................................ */
/* metropolis returns '1' if the step is accepted		*/
/* delta_energy= energy_new - energy_old			*/
/*								*/
/* In order to avoid an underflow error in exp(), a constant was*/
/* defined in energy.h. No error message is issued.		*/

int metropolis(double delta_energy, double temp, long *seed)
{
double help;
/* fprintf(stdout,"Delta Energy: %lf\n",delta_energy);*/
/*fprintf(stdout,"Delta Energy: %e\n",-delta_energy/(temp*R_GAS));*/


if (delta_energy < 0.0) 
   {return ACCEPTED; /* 1 */
   }
else {help=-delta_energy/(temp*R_GAS);
      if (help<TOO_SMALL)
	 {return REJECTED;
	 }
      else {if ((temp!=0.0) && ((Ran(seed)) < exp(help)))
	       {return ACCEPTED; /* 1 */;
	       }
	    else {return REJECTED;
		 }
	   }
     }
}
/* ............................................................ */
/* metropolis returns '1' if the step is accepted		*/
/* delta_energy= energy_new - energy_old			*/
/*								*/
/* In order to avoid an underflow error in exp(), a constant was*/
/* defined in energy.h. No error message is issued.		*/
/*								*/
/* uses beta instead of the temparature.			*/
/*								*/
int metropolis2(double delta_energy, double beta, long *seed)
{
double help;


if (delta_energy < 0.0) 
   {return ACCEPTED; /* 1 */
   }
else {help=-delta_energy/beta;
      if (help<TOO_SMALL)
	 {return REJECTED;
	 }
      else {if ((beta!=0.0) && ((Ran(seed)) < exp(help)))
	       {return ACCEPTED; /* 1 */;
	       }
	    else {return REJECTED;
		 }
	   }
     }
}

/********************************************************************************/
/* Random number generation  				                        */
/********************************************************************************/


/*..............................................................................*/
/* Long period (>2*10^18) random number genertator of L'Ecuyer with BAYES-	*/
/* DURHAM shuffle and added safeguards. Returns a uniform deviate between 0.0	*/
/* and 1.0 (exclusive the endpoint values). Call with idum a negative integer	*/
/* to initialize; thereafter, do not alter idum between  successive deviates 	*/
/* in a sequence. RNMX should approximate the largest floating value that is 	*/
/* less than 1.0.								*/
/*										*/
/* The number idum should be a negative integer, to make the first if true.	*/
/*										*/


#define NTAB 32
double RandomBayesDurham(long *idum)
{
        long j;
        long k;
        static long idum2=123456789;
        static long iy=0;
        static long iv[NTAB];
        double temp;

        long IM1=2147483563;
        long IM2=2147483399;
	long IMM1=(IM1-1);
        double AM=(1.0/IM1);
        long IA1=40014;
        long IA2=40692;
        long IQ1=53668;
        long IQ2=52774;
        long IR1=12211;
        long IR2=3791;
        double NDIV=(1.0+IMM1/NTAB);
        double EPS=1.2e-7;
        double RNMX=(1.0-EPS);

        if (*idum <= 0)
	   {if (-(*idum) < 1) *idum=1;
            else *idum = -(*idum);
            idum2=(*idum);
            for (j=NTAB+7; j>=0; j--)
	        {k=(*idum)/IQ1;
                 *idum=IA1*(*idum-k*IQ1)-k*IR1;
                 if (*idum < 0) *idum += IM1;
                 if (j < NTAB)  iv[j] = *idum;
                }
                iy=iv[0];
        }
        k=(*idum)/IQ1;
        *idum=IA1*(*idum-k*IQ1)-k*IR1;
        if (*idum < 0) *idum += IM1;
        k=idum2/IQ2;
        idum2=IA2*(idum2-k*IQ2)-k*IR2;
        if (idum2 < 0) idum2 += IM2;
        j=iy/NDIV;
        iy=iv[j]-idum2;
        iv[j] = *idum;
        if (iy < 1) iy += IMM1;
        if ((temp=AM*(double)iy) > RNMX)
	     return RNMX;
        else return temp;
}
#undef NTAB

/*..............................................................................*/
/* 
Diese Routine liefert Zufallszahlen auf dem Intervall [0;Pi] deren 
Haeufigkeit proportional zu sin(x) ist. 
*/

double RandomSinus(long *idum)
{
  double new2, rho;
  int NotAccepted;

  NotAccepted=1;
  while (NotAccepted) {
    rho=sin(M_PI*(new2=Ran(idum)));
    if (Ran(idum)>1.0-rho)
      NotAccepted=0;
  }

  return (M_PI*new2);
}
