/* Matthias Ullmann, Berlin 15.01.1995					*/
/* 									*/
/* routines for surface and volume calculation 				*/
/* 									*/
/* method by F. Eisenhaber:						*/
/* F.Eisenhaber, P.Lijnzaad, P.Argos, M.Scharf				*/
/*    "The Double Cubic Lattice Method: Efficient Approaches to		*/
/*    Numerical Integration of Surface Area and Volume and to Dot	*/
/*    Surface Contouring of Molecular Assemblies"			*/
/*    Journal of Computational Chemistry (1995) v.16, N3, pp.273-284	*/
/*									*/
/* .................................................................... */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "mathematik.h"
#include "constants.h"
#include "memutil.h"
#include "util.h"
#include "grid.h"
#include "surface.h"
#include "graph.h"

/************************************************************************/
/* function returns the typ of tesselation		     		*/
/*									*/
int tessellation_type(int densit) 
{
      int i1=1, i2=1;
        
      while (10*i1*i1+2 < densit) i1++;
      while (30*i2*i2+2 < densit) i2++;
      if (10*i1*i1-2 < 30*i2*i2-2) 
           {return UNSP_ICO_ARC;}   /* i.e. class I  tessellation */
      else {return UNSP_ICO_DOD;}   /* i.e. class II tessellation */
}


/* .................................................................... */


/************************************************************************/
/* .................................................................... */
/* Mapping the unit sphere to a grid using a linked list. 		*/
/* This grid looks a little bit different than the grid used for the 	*/
/* grid atom.Tthe reason is, that it doesn't change during a calculation*/
/* and so it need less memory and is hopefully slightly faster 		*/
/* (especially on maschines with fast integer operations)		*/
/*									*/
void unitsphere_to_grid(Surf *surf)
{
  int ***help;
  int i,j,k,n,m, len;
  
  /* determine the grid dimensions - 2.0=radius of a unit sphere */
  if ((ceil((double)(*surf).cubes/2.0))==(floor((double)(*surf).cubes/2.0)))
     {ERROR("Cubes should be an odd integer. unitsphere_to_grid()");
     }
  len=(int)((double)((*surf).cubes-1)/2.0);
  (*surf).spacing=2.0/(double)((*surf).cubes-1);
  (*surf).inv_spacing=1.0/(*surf).spacing;
  
  /* first look for the maximal length of a dot_no-vector */
  help=i3tensor(-len, len, -len, len, -len, len);
  
  for(i=-len;i<=len;i++)	/* Initializing */
     for(j=-len;j<=len;j++)
         for(k=-len;k<=len;k++)
             {help[i][j][k]=0;
             }
  for(n=1; n<=(*surf).n_dot; n++) /* counting */
     {i=(int)rint((*surf).xus[n][1]*(*surf).inv_spacing);
      j=(int)rint((*surf).xus[n][2]*(*surf).inv_spacing);
      k=(int)rint((*surf).xus[n][3]*(*surf).inv_spacing);
      help[i][j][k]++;
     }
  (*surf).nmax=0;	/* search the highest value */
  for(i=-len;i<=len;i++)
     for(j=-len;j<=len;j++)
         for(k=-len;k<=len;k++)
            {if ((*surf).nmax<help[i][j][k]) (*surf).nmax=help[i][j][k];
            }
  /* allocate memory for dot_no */
  (*surf).dot_no=i4tensor(-len, len, -len, len, -len, len,0,(*surf).nmax);
  (*surf).imax=ivector(1,3);
  (*surf).imin=ivector(1,3);
  
  (*surf).imin[1]=-len;	(*surf).imax[1]=len;
  (*surf).imin[2]=-len;	(*surf).imax[2]=len;
  (*surf).imin[3]=-len;	(*surf).imax[3]=len;
  
  /* initialize the whole grid with zero */
  for(i=-len;i<=len;i++)
     for(j=-len;j<=len;j++)
         for(k=-len;k<=len;k++)
            for(m=0;m<=(*surf).nmax;m++)
               (*surf).dot_no[i][j][k][m]=0;
  
  /* assign the points to dot_no */
  for(n=1; n<=(*surf).n_dot; n++) /* counting */
     {i=(int)rint((*surf).xus[n][1]*(*surf).inv_spacing);
      j=(int)rint((*surf).xus[n][2]*(*surf).inv_spacing);
      k=(int)rint((*surf).xus[n][3]*(*surf).inv_spacing);
      (*surf).dot_no[i][j][k][0]++;
      (*surf).dot_no[i][j][k][(*surf).dot_no[i][j][k][0]]=n;
     }
  /* check consitency of the numbers and free the help tensor */
  n=0;
  for(i=-len;i<=len;i++)
     for(j=-len;j<=len;j++)
         for(k=-len;k<=len;k++)
            {n=n+help[i][j][k];
             if ((*surf).dot_no[i][j][k][0]!=help[i][j][k])
                ERROR("Couning error in unitsphere_to_grid().");
            }
  
  if (n!=(*surf).n_dot)
     ERROR("Number of dots not equal to assigned in  unitsphere_to_grid().");
  
  free_i3tensor(help, -len, len, -len, len, -len, len);
}
/* .................................................................... */
/* procedure calculates the solvent accesible area of a given atom	*/
/* 									*/
/* when this function is called, the atom should not be in the list.	*/
/* otherwise the solvent accessible area is allways be zero		*/
/*									*/
/* Variables								*/
/* il,jl,kl	- left hand values for the second grid			*/
/* ir,jr,kr	- right hand values for the second grid			*/
/* 									*/
double calculate_area(int no, char **flag, double **xyz, double *radius,
		      Surf *surf, iListGrid *grid)
{
int i,j,k,ih,jh,kh,ix,iy,iz;	   /* loop in the first grid  */
int il,jl,kl,ir,jr,kr, ia,ja,ka,c; /* loop in the second grid */
int acc, ne; /* ne==neighbor , acc == number of accessible point */
iList *help;
double d12[4], d2, r12, r22, r1, r2, right,len,fact;

/* set some initial values */
acc=(*surf).n_dot;
r1=radius[no];  r12=r1*r1;

/* map the atom on the grid */
ix=(int)rint(xyz[no][1]*(*grid).inv_spacing);
iy=(int)rint(xyz[no][2]*(*grid).inv_spacing);
iz=(int)rint(xyz[no][3]*(*grid).inv_spacing);


/* loop over all neighbors and look which surface dots are buried */
for (i=-(*grid).look; i<=(*grid).look; i++)
  {ih=ix+i;
   if(!((ih>(*grid).imax[1])||(ih<(*grid).imin[1]))) 
     {for (j=-(*grid).look; j<=+(*grid).look; j++)
        {jh=iy+j;
	 if (!((jh>(*grid).imax[2])||(jh<(*grid).imin[2]))) 
	   {for (k=-(*grid).look; k<=+(*grid).look; k++)
	       {kh=iz+k;
	        if (!((kh>(*grid).imax[3])||(kh<(*grid).imin[3])))
	          {help=&(*grid).root[ih][jh][kh];
/* ======= code folded from here */ 
while((*help).next!=NULL) /* The in no entry in the grid!! */
  {help=(*help).next; ne=(*help).number;
   /* check via triangle inequility if the spheres interfere */
   SubVector(xyz[ne],xyz[no],d12);    d2=SqrLength(d12);
   r2=radius[ne]; r22=r2*r2;
   if (d2<=((r1+r2)*(r1+r2)))
      {/* calculate the right hand side value for the inequility */
       right=(d2+r12-r22)/(2*r1);
       /* determine in which cells in the grid second grid must be looked */
       fact=1.0/((*surf).spacing*r1);
       len=((double)((*surf).cubes-1)/2.0);

       il=(int)floor(MAXIMUM(((d12[1]-r2)*fact),-len)); 
       jl=(int)floor(MAXIMUM(((d12[2]-r2)*fact),-len)); 
       kl=(int)floor(MAXIMUM(((d12[3]-r2)*fact),-len));
 
       ir=(int)ceil(MINIMUM(((d12[1]+r2)*fact),len)); 
       jr=(int)ceil(MINIMUM(((d12[2]+r2)*fact),len)); 
       kr=(int)ceil(MINIMUM(((d12[3]+r2)*fact),len)); 

       /* loop over the grid and the surface dots */
       for(ia=il; ia<=ir; ia++) 
	  {for(ja=jl; ja<=jr; ja++) 
	      {for(ka=kl; ka<=kr; ka++) 
		  {/*fprintf(stderr,"%d  %d  %d\n",ia,ja,ka);*/
		   if (0!=(*surf).dot_no[ia][ja][ka][0])
		    {for(c=1; c<=(*surf).dot_no[ia][ja][ka][0]; c++)
		      {if (flag[no][(*surf).dot_no[ia][ja][ka][c]]==OUTSIDE)
		          {if (ScalarProduct(d12, (*surf).xus[(*surf).dot_no[ia][ja][ka][c]])>right)
			      {flag[no][(*surf).dot_no[ia][ja][ka][c]]=INSIDE;
			       acc--;
			      }
			  }
		      }
		    }
		  }
	      }
          }
      }
  }
/* ======= unfolding the code */
	          }
	     	}
	   }
        }
     }
  }

/* do the area calculation */
return Area(acc,(*surf).n_dot,r1);
}

/************************************************************************/
/* The actual surface calculation					*/
/*									*/
/* .................................................................... */
/* calls the surface calculation procedures				*/
/* it return the surface area						*/
/* 									*/
double calculate_surface(double *area, char **flag, double **xyz, 
			 double *radius, int natom, Surf *surf, 
			 iListGrid *grid, iList *list)
{
int i;
double surface=0.0;
/* loop over all atoms */
for (i=1; i<=natom; i++)
    {/* remove the atom from the grid */
     remove_from_list(&list[i]);
     /* do the calculation of the solvent accessible area of this atom */
     area[i]=calculate_area(i, flag, xyz, radius, surf, grid);
     surface=surface+area[i];
     /* read the atom to the list */
     add_to_list(grid, &list[i], xyz[i]);
    }
return surface;
}

/* .................................................................... */
/* calculates the dotsurface of a protein				*/
/*									*/

void calculate_dotsurface(DotSurf *dotsurf, char **flag, Surf *surf,
			  double **xyz, double *radius, int natoms,
			  char **idstring)
{
int i,j,k;

k=0;
for (i=1; i<=natoms; i++)
    {for(j=1; j<=(*surf).n_dot;j++)
	{ if (flag[i][j]==OUTSIDE)
	    {k++;
             if (k>(*dotsurf).ndots) 
                {ERROR("Too many dots in calculate_dotsurface()");}
	     (*dotsurf).dots[k][1]=(radius[i]*(*surf).xus[j][1])+xyz[i][1];
	     (*dotsurf).dots[k][2]=(radius[i]*(*surf).xus[j][2])+xyz[i][2];
	     (*dotsurf).dots[k][3]=(radius[i]*(*surf).xus[j][3])+xyz[i][3];
             strcpy((*dotsurf).idstring[k], idstring[i]);
	     if ((*surf).n_dot==0)
                {ERROR("(*surf).n_dot==0");}
             (*dotsurf).area[k]=Area(1,(*surf).n_dot,radius[i]);
	    }
	 else {if (flag[i][j]!=INSIDE)
		  {WARN("Unknown flag in calculate_dotsurface().");
		   fprintf(stderr,"%d  %d  %c %d\n",i,j,flag[i][j],(int)flag[i][j]);;
		  }
	      }
	}
    }
}
/* .................................................................... */
/* This routine calculates the surface does all memory allocation etc   */
/*                                                                      */

double surface(double *atomsurf, double **xyz, double *radius, 
               int natoms, int densit, double prob, DotSurf *dotsurf,
               int surfflag, char **idstring)
{
  int i,j;
  double max[4], min[4], off_set, srf, *rad;
  double len; 
  int itest;
  Surf surf;
  iListGrid atom_list;
  iList *atoms;
  char **flag;  /* [1..natom][1..n_dot] INSIDE - OUTSIDE */


  rad=dvector(1,natoms);
  for (i=1; i<=natoms; i++)
        rad[i]=radius[i]+prob;

  /* determinte which tessellation type should be used */
  itest=tessellation_type(densit);
  /* calculate the tesselation of a unit sphere */
  switch (itest)
    {case UNSP_ICO_ARC:    DEBUGGING(fprintf(stdout,"\t Use tessellation type II\n");)
                           ico_dot_arc(densit, &surf);
	                   DEBUGGING(fprintf(stdout,"\t I use %d dots\n",surf.n_dot);)
                           break;
    case UNSP_ICO_DOD:     DEBUGGING(fprintf(stdout,"\t Use tessellation type I\n");)
                           ico_dot_dod(densit, &surf);
                           DEBUGGING(fprintf(stdout,"\t I use %d dots\n",surf.n_dot);)
                           break;
      default:             ERROR("\t Don't know this tessellation type");
      }

   /* ============================================================= */
   /* renormalize in order to be sure that it is a unit sphere       */
   for (j=1; j<=surf.n_dot; j++)
      {len = sqrt(SqrLength(surf.xus[j]));
       surf.xus[j][1]=surf.xus[j][1]/len;
       surf.xus[j][2]=surf.xus[j][2]/len;
       surf.xus[j][3]=surf.xus[j][3]/len;
      }
   /* ============================================================= */
   /* sort the dots in a Grid linked cubes=cubes in one dimension   */
   surf.cubes=5;
   unitsphere_to_grid(&surf);

   /* ============================================================ */
  /* allocate and initilize the flags with OUTSIDE */
  /* look for the needed grid dimension */
  atom_list.imax=ivector(1,3);
  atom_list.imin=ivector(1,3);

  off_set=10.0;
  atom_list.spacing=2.0*max_element(rad, natoms);
  atom_list.inv_spacing=1.0/atom_list.spacing;
  min_max_coord(min, max, xyz, natoms);
  atom_list.inv_spacing=1.0/atom_list.spacing;
  atom_list.imax[1]=(int)ceil((max[1]+off_set) *atom_list.inv_spacing);
  atom_list.imax[2]=(int)ceil((max[2]+off_set) *atom_list.inv_spacing);
  atom_list.imax[3]=(int)ceil((max[3]+off_set) *atom_list.inv_spacing);
  atom_list.imin[1]=(int)floor((min[1]-off_set)*atom_list.inv_spacing);
  atom_list.imin[2]=(int)floor((min[2]-off_set)*atom_list.inv_spacing);
  atom_list.imin[3]=(int)floor((min[3]-off_set)*atom_list.inv_spacing);
  atom_list.look=1;
  /* make the grid spacing to odd integer, if they are not odd integer */
  odding_down(atom_list.imin,3);
  odding_up(atom_list.imax,3);
  /* allocation */
  flag=charmatrix(1,natoms,1,surf.n_dot);
  atoms=iList_vector(1,natoms);
  atom_list.root=iList_3tensor(atom_list.imin[1], atom_list.imax[1],
                               atom_list.imin[2], atom_list.imax[2],
                               atom_list.imin[3], atom_list.imax[3]);

  for (j=1; j<=natoms; j++)
         {init_charvector(flag[j],surf.n_dot,OUTSIDE);
          atomsurf[j]=0.0;
         }
  init_iList_vector(atoms,natoms);
  init_iListGrid(&atom_list);
  fill_iListGrid(&atom_list, atoms, xyz, natoms);
  srf=calculate_surface(atomsurf, flag, xyz,rad, natoms,
			&surf,&atom_list, atoms);

   if (surfflag==1)
      {/* calculate the dot surface */
       /* count the surface dots and allocate memory for the surface dots */
       (*dotsurf).ndots=0;
       for (i=1; i<=natoms; i++)
           for(j=1; j<=surf.n_dot;j++)
              if (flag[i][j]==OUTSIDE) (*dotsurf).ndots++;
       (*dotsurf).dots=dmatrix(1,(*dotsurf).ndots,1,3);
       (*dotsurf).idstring = charmatrix(1,(*dotsurf).ndots, 0, 100);
       (*dotsurf).area=dvector(1,(*dotsurf).ndots);
       /* do the calculation */
       calculate_dotsurface(dotsurf,flag,&surf,xyz, 
			    rad,natoms, idstring);
     }

  /* release memory */
  free_i4tensor(surf.dot_no, surf.imin[1], surf.imax[1],
                             surf.imin[2], surf.imax[2],
                             surf.imin[3], surf.imax[3],
                             0,surf.nmax);
  free_dmatrix(surf.xus,1,surf.n_dot, 1,3);
  free_ivector(surf.imin, 1,3);
  free_ivector(surf.imax, 1,3);
  free_dvector(rad,1,natoms);
  free_charmatrix(flag,1,natoms,1,surf.n_dot);
  free_iList_vector(atoms,1,natoms);
  free_iList_3tensor(atom_list.root,atom_list.imin[1], atom_list.imax[1],
                               atom_list.imin[2], atom_list.imax[2],
                               atom_list.imin[3], atom_list.imax[3]);
  free_ivector(atom_list.imax,1,3);
  free_ivector(atom_list.imin,1,3);

  return srf;
}
/* .................................................................... */
/* This routine calculates the surface does all memory allocation etc   */
/*                                                                      */
/* itest -- tesselation type                                            */

double DotSurface(double *atomsurf, double **xyz, double *radius, 
                  int natoms, int densit, double prob, DotSurf *dotsurf,
		  int surfflag, int itest, long *seed, FILE *log, char **idstring)
{
   int i,j;
   double max[4], min[4], off_set, srf, *rad;
   Surf surf;
   iListGrid atom_list;
   iList *atoms;
   char **flag;  /* [1..natom][1..n_dot] INSIDE - OUTSIDE */


   rad=dvector(1,natoms);
   for (i=1; i<=natoms; i++)
        rad[i]=radius[i]+prob;

    /* ============================================================= */
   /* determinte which tessellation type should be used */
   if (itest==0) itest=tessellation_type(densit);
   /* calculate the tesselation of a unit sphere */
   switch (itest)
      {case UNSP_ICO_ARC:  fprintf(log,"\t Use tessellation type II\n"),
                           ico_dot_arc(densit, &surf);
                           break;
       case UNSP_ICO_DOD:  fprintf(log,"\t Use tessellation type I\n");
                           ico_dot_dod(densit, &surf);
                           break;
       case UNSP_RANDOM:   fprintf(log,"\t Use tessellation type III - Random\n");
	                   random_tess(densit, &surf, seed);
                           break;
       default:            ERROR("\t Don't know this tessellation type");
      }
    fprintf(log,"\t Use %d dots\n",surf.n_dot);

   /* ============================================================== */
   /* print result to logfile                                        */
    /*
   fprintf(log,"Points on the unit sphere: %d\n",surf.n_dot);
   for (j=1; j<=surf.n_dot; j++)
       {fprint_DUMMYatom(log, j, "DUMM", "DUM",surf.xus[j], 0.0, 0.0);
       }
    */
   /* ============================================================= */
   /* sort the dots in a Grid linked cubes=cubes in one dimension   */
   surf.cubes=5;
   unitsphere_to_grid(&surf);


  /* ============================================================ */
  /* allocate and initilize the flags with OUTSIDE */
  /* look for the needed grid dimension */
  atom_list.imax=ivector(1,3);
  atom_list.imin=ivector(1,3);

  off_set=10.0;
  atom_list.spacing=2.0*max_element(rad, natoms);
  atom_list.inv_spacing=1.0/atom_list.spacing;
  min_max_coord(min, max, xyz, natoms);
  atom_list.inv_spacing=1.0/atom_list.spacing;
  atom_list.imax[1]=(int)ceil((max[1]+off_set) *atom_list.inv_spacing);
  atom_list.imax[2]=(int)ceil((max[2]+off_set) *atom_list.inv_spacing);
  atom_list.imax[3]=(int)ceil((max[3]+off_set) *atom_list.inv_spacing);
  atom_list.imin[1]=(int)floor((min[1]-off_set)*atom_list.inv_spacing);
  atom_list.imin[2]=(int)floor((min[2]-off_set)*atom_list.inv_spacing);
  atom_list.imin[3]=(int)floor((min[3]-off_set)*atom_list.inv_spacing);
  atom_list.look=1;
  /* make the grid spacing to odd integer, if they are not odd integer */
  odding_down(atom_list.imin,3);
  odding_up(atom_list.imax,3);

  /* allocation */
  flag=charmatrix(1,natoms,1,surf.n_dot);
  atoms=iList_vector(1,natoms);
  atom_list.root=iList_3tensor(atom_list.imin[1], atom_list.imax[1],
                               atom_list.imin[2], atom_list.imax[2],
                               atom_list.imin[3], atom_list.imax[3]);

  for (j=1; j<=natoms; j++)
         {init_charvector(flag[j],surf.n_dot,OUTSIDE);
          atomsurf[j]=0.0;
         }
  init_iList_vector(atoms,natoms);
  init_iListGrid(&atom_list);
  fill_iListGrid(&atom_list, atoms, xyz, natoms);
  srf=calculate_surface(atomsurf, flag, xyz,rad, natoms,
                       &surf,&atom_list, atoms);

  if (surfflag==1)
      {/* calculate the dot surface */
       /* count the surface dots and allocate memory for the surface dots */
       (*dotsurf).ndots=0;
       for (i=1; i<=natoms; i++)
           for(j=1; j<=surf.n_dot;j++)
              if (flag[i][j]==OUTSIDE) (*dotsurf).ndots++;
           
       (*dotsurf).dots     = dmatrix(1,(*dotsurf).ndots,1,3);
       (*dotsurf).area     = dvector(1,(*dotsurf).ndots);
       (*dotsurf).idstring = charmatrix(1,(*dotsurf).ndots, 0, 100);
       /* do the calculation */
       calculate_dotsurface(dotsurf,flag,&surf,xyz, 
			    rad,natoms, idstring);
      }

  /* release memory */
  free_i4tensor(surf.dot_no, surf.imin[1], surf.imax[1],
                             surf.imin[2], surf.imax[2],
                             surf.imin[3], surf.imax[3],
                             0,surf.nmax);
  free_dmatrix(surf.xus,1,surf.n_dot, 1,3);
  free_ivector(surf.imin, 1,3);
  free_ivector(surf.imax, 1,3);
  free_dvector(rad,1,natoms);
  free_charmatrix(flag,1,natoms,1,surf.n_dot);
  free_iList_vector(atoms,1,natoms);
  free_iList_3tensor(atom_list.root,atom_list.imin[1], atom_list.imax[1],
                               atom_list.imin[2], atom_list.imax[2],
                               atom_list.imin[3], atom_list.imax[3]);
  free_ivector(atom_list.imax,1,3);
  free_ivector(atom_list.imin,1,3);

  return srf;
}

/* ======================================================= */
/* read surface points from a file                         */
/*                                                         */

void ReadSrfPts(DotSurf *dotsurf, char *fname)
{
  FILE *fpin;
  int n, h;
  char line[LINELENGTH];

  fpin =   file_open(fname,"r");

  /* count lines */
  n=0;
  while (fgets(line,LINELENGTH,fpin)!=NULL)  
        {if (0==strncmp("SRFDOT",line,6)) n++;
        }

  /* allocate */
  (*dotsurf).ndots = n;
  (*dotsurf).dots     = dmatrix(1,(*dotsurf).ndots,1,3);
  (*dotsurf).area     = dvector(1,(*dotsurf).ndots);
  (*dotsurf).idstring = charmatrix(1,(*dotsurf).ndots, 0, 100);
  /* read info */
  rewind(fpin);
  n=0;
  while (fgets(line,LINELENGTH,fpin)!=NULL)  
        {if (0==strncmp("SRFDOT",line,6)) 
            {n++;
             h = sscanf(line, "%*s %lf %lf %lf %lf %s",
                        &(*dotsurf).dots[n][1],
                        &(*dotsurf).dots[n][2],
                        &(*dotsurf).dots[n][3],
                        &(*dotsurf).area[n],
                        (*dotsurf).idstring[n]);
             if (h!=5) muerror("Problem while reading SRFDOT");
            }

        }

  fclose(fpin);
}

/* ======================================================= */
/* join adjencent surface points in a graph (adjecency     */
/* matrix). if the distance is <=distlimit, two points     */
/* are connected                                           */
/*                                                         */

void ConnectPointsMatrix(int **graph, DotSurf *dotsurf, double distlimit)
{
  int i,j;
  double dist2, d2;

  d2 = distlimit*distlimit;

  for (i=1; i<=(*dotsurf).ndots; i++)
      {graph[i][i]=1;
       for (j=1; j<i; j++)
           {dist2 = SqrDistance((*dotsurf).dots[i], (*dotsurf).dots[j]);
            if (dist2<=d2)  graph[i][j]=graph[j][i]=1; /* connection    */
            else            graph[i][j]=graph[j][i]=0; /* no connection */
           }
      }
}
/* ======================================================= */
/* join adjencent surface points in a graph (list).        */
/* if the distance is <=distlimit, two points are          */
/* connected                                          */
/*                                                         */
void ConnectPointsList(Node *adj, DotSurf *dotsurf, double distlimit)
{
  int i,j;
  double dist2, d2;
  Node *t;

  d2 = distlimit*distlimit;

  for (i=1; i<=(*dotsurf).ndots; i++)
      {for (j=1; j<i; j++)
	   {dist2 = SqrDistance((*dotsurf).dots[i], (*dotsurf).dots[j]);
	    if (dist2<=d2)  /* connection    */
	       {AllocNode(t);
		(*t).v    = i;
		(*t).next = adj[j].next;
		adj[j].next = t;

		AllocNode(t);
		(*t).v = j;
		(*t).next = adj[i].next;
		adj[i].next = t;
	       }
	   }
      }
}
