/* Matthias Ullmann, Berlin 15.01.1995					*/
/* 									*/
/* routines for surface and volume calculation 				*/
/* 									*/
/* method by F. Eisenhaber:						*/
/* F.Eisenhaber, P.Lijnzaad, P.Argos, M.Scharf				*/
/*    "The Double Cubic Lattice Method: Efficient Approaches to		*/
/*    Numerical Integration of Surface Area and Volume and to Dot	*/
/*    Surface Contouring of Molecular Assemblies"			*/
/*    Journal of Computational Chemistry (1995) v.16, N3, pp.273-284	*/
/*									*/
/* .................................................................... */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "memutil.h"
#include "util.h"
#include "surface.h"
#include "mathematik.h"
#include "random.h"

/************************************************************************/
/* creation of a regular dot distribution - tesselation			*/
/* 									*/
/* The code is mainly taken from Frank Eisenhaber - NSC			*/
/* It has been adopted to my array alloaction starting by 1 and ending 	*/
/* at n rather than starting at 0 and ending at n-1.			*/
/* The resulting code is a little bit complex and hard to understand.	*/
/* DON'T TOUCH IT, if you have the possibility to avoid it.		*/
/* Since these function are called only once in a run, it is not a CPU- */
/* time critical part.							*/
/* .................................................................... */
/* this procedure catches errors in asin.				*/
/* 									*/
#define ASIN safe_asin
double safe_asin(double f) 
{
if ( (fabs(f) < 1.00) ) return( asin(f) );
if ( (fabs(f) - 1.00)  <= DP_TOL ) 
    fprintf(stderr,"ERROR: ASIN : invalid argument %f", f);
return(M_PI_2);
}
/* .................................................................... */
/* 									*/
void divarc(double *x1, double *x2, int div1, int div2, double *xr) 
{
double xd, yd, zd, dd, d1, d2, s, x, y, z;
double phi, sphi, cphi;

xd = x1[2]*x2[3]-x2[2]*x1[3];
yd = x1[3]*x2[1]-x2[3]*x1[1];
zd = x1[1]*x2[2]-x2[1]*x1[2];
dd = sqrt(xd*xd + yd*yd + zd*zd);

if (dd < DP_TOL) fprintf(stderr,"ERROR: divarc: rotation axis of length %f\n", dd);

d1 = x1[1]*x1[1] + x1[2]*x1[2] + x1[3]*x1[3];
if (d1 < 0.5) fprintf(stderr,"ERROR: divarc: vector 1 of sq.length %f\n", d1);
d2 = x2[1]*x2[1] + x2[2]*x2[2] + x2[3]*x2[3];
if (d2 < 0.5) fprintf(stderr,"ERROR: divarc: vector 2 of sq.length %f\n", d2);

phi = ASIN(dd/sqrt(d1*d2));
phi = phi*((double)div1)/((double)div2);
sphi = sin(phi); 
cphi = cos(phi);
s  = (x1[1]*xd + x1[2]*yd + x1[3]*zd)/dd;

x = xd*s*(1.-cphi)/dd + x1[1] * cphi + (yd*x1[3]-x1[2]*zd)*sphi/dd;
y = yd*s*(1.-cphi)/dd + x1[2] * cphi + (zd*x1[1]-x1[3]*xd)*sphi/dd;
z = zd*s*(1.-cphi)/dd + x1[3] * cphi + (xd*x1[2]-x1[1]*yd)*sphi/dd;
dd = sqrt(x*x + y*y + z*z);

xr[1] = x/dd; xr[2] = y/dd; xr[3] = z/dd;
}

/* .................................................................... */
/* routine for dot distributions on the surface of the unit sphere      */
/*									*/

void icosaeder_vertices(double **xus) 
{
      double rg, rh;
      
      rh = sqrt(1.-2.*cos(TORAD(72.)))/(1.-cos(TORAD(72.)));
      rg = cos(TORAD(72.))/(1.-cos(TORAD(72.)));
      /* icosaeder vertices */
      xus[ 1][1] = 0.;                  xus[ 1][2] = 0.;                  xus[ 1][3] =  1.;
      xus[ 2][1] = rh*cos(TORAD(72.));  xus[ 2][2] = rh*sin(TORAD(72.));  xus[ 2][3] =  rg;
      xus[ 3][1] = rh*cos(TORAD(144.)); xus[ 3][2] = rh*sin(TORAD(144.)); xus[ 3][3] =  rg;
      xus[ 4][1] = rh*cos(TORAD(216.)); xus[ 4][2] = rh*sin(TORAD(216.)); xus[ 4][3] =  rg;
      xus[ 5][1] = rh*cos(TORAD(288.)); xus[ 5][2] = rh*sin(TORAD(288.)); xus[ 5][3] =  rg;
      xus[ 6][1] = rh;                  xus[ 6][2] = 0;                   xus[ 6][3] =  rg;
      xus[ 7][1] = rh*cos(TORAD(36.));  xus[ 7][2] = rh*sin(TORAD(36.));  xus[ 7][3] = -rg;
      xus[ 8][1] = rh*cos(TORAD(108.)); xus[ 8][2] = rh*sin(TORAD(108.)); xus[ 8][3] = -rg;
      xus[ 9][1] = -rh;                 xus[ 9][2] = 0;                   xus[ 9][3] = -rg;
      xus[10][1] = rh*cos(TORAD(252.)); xus[10][2] = rh*sin(TORAD(252.)); xus[10][3] = -rg;
      xus[11][1] = rh*cos(TORAD(324.)); xus[11][2] = rh*sin(TORAD(324.)); xus[11][3] = -rg;
      xus[12][1] = 0.;                  xus[12][2] = 0.;                  xus[12][3] = -1.;
}
/* .................................................................... */
/* construct the class I tessellation					*/
/*									*/
/* densit...required dots per unit sphere 		     		*/
/* dot distribution on a unit sphere based on an icosaeder   		*/
/* great circle average refining of icosahedral face         		*/
/*									*/
void ico_dot_arc(int densit, Surf *surf) 
{
      double x1[4], x2[4], x3[4], xij[4], xji[4], xki[4], xjk[4], xik[4], xkj[4];
      double a, rh;
      double x,y,z,d ;
      int i, j, k, tess, tn, tl, tl2; 
      
      /* Code */
      /* calculate the number of dots per unit sphere */
      a = sqrt((((double) densit)-2.)/10.);
      tess = (int) ceil(a);
      (*surf).n_dot = 10*tess*tess+2;
      if ((*surf).n_dot < densit) 
      	{fprintf(stderr,"WARNING: ico_dot_arc: error in formula for tessellation level (%d->%d, %d)\n", 
      		 tess, (*surf).n_dot, densit);
      	}
      
      /* allocation of dot-matrix */
      (*surf).xus =dmatrix(1,(*surf).n_dot, 1,3);
      
      /* calculate the first 12 dots */
      icosaeder_vertices((*surf).xus);
      
      if (tess > 1) 
         {tn = 12;
          rh = sqrt(1.-2.*cos(TORAD(72.)))/(1.-cos(TORAD(72.)));
          a = rh*rh*2.*(1.-cos(TORAD(72.)));
          /* calculate tessellation of icosaeder edges */
          for (i=1; i<=11; i++) 
      	{for (j=i+1; j<=12; j++) 
      	    {x = (*surf).xus[i][1]-(*surf).xus[j][1];
      	     y = (*surf).xus[i][2]-(*surf).xus[j][2]; 
      	     z = (*surf).xus[i][3]-(*surf).xus[j][3];
                   d = x*x+y*y+z*z;
                   if (fabs(a-d) > DP_TOL) continue;
                   for (tl=1; tl<tess; tl++) 
      		 {if (tn >= (*surf).n_dot) 
                           {fprintf(stderr,"ERROR: ico_dot: tn exceeds dimension of xus\n");}
                        tn++;
                        divarc((*surf).xus[i], (*surf).xus[j], tl, tess, (*surf).xus[tn]);
                       }
                  }
              }
          /* calculate tessellation of icosaeder faces */
          for (i=1; i<=10; i++) 
      	{for (j=i+1; j<=11; j++) 
      	    {x = (*surf).xus[i][1]-(*surf).xus[j][1];
                   y = (*surf).xus[i][2]-(*surf).xus[j][2]; 
      	     z = (*surf).xus[i][3]-(*surf).xus[j][3];
                   d = x*x+y*y+z*z;
                   if (fabs(a-d) > DP_TOL) continue;
      
      	     for (k=j+1; k<=12; k++) 
      		{x = (*surf).xus[i][1]-(*surf).xus[k][1];
                	 y = (*surf).xus[i][2]-(*surf).xus[k][2]; 
      		 z = (*surf).xus[i][3]-(*surf).xus[k][3];
                	 d = x*x+y*y+z*z;
                	 if (fabs(a-d) > DP_TOL) continue;
                	 x = (*surf).xus[j][1]-(*surf).xus[k][1];
                	 y = (*surf).xus[j][2]-(*surf).xus[k][2]; 
      		 z = (*surf).xus[j][3]-(*surf).xus[k][3];
                	 d = x*x+y*y+z*z;
                	 if (fabs(a-d) > DP_TOL) continue;
      
                	 for (tl=1; tl<tess-1; tl++) 
      		     {divarc((*surf).xus[j], (*surf).xus[i], tl, tess, xji);
                  	      divarc((*surf).xus[k], (*surf).xus[i], tl, tess, xki);
      		      for (tl2=1; tl2<tess-tl; tl2++) 
      			  {divarc((*surf).xus[i], (*surf).xus[j], tl2, tess, xij);
                    		   divarc((*surf).xus[k], (*surf).xus[j], tl2, tess, xkj);
                    		   divarc((*surf).xus[i], (*surf).xus[k], tess-tl-tl2, tess, xik);
                    		   divarc((*surf).xus[j], (*surf).xus[k], tess-tl-tl2, tess, xjk);
                    		   if (tn >= (*surf).n_dot) 
				     fprintf(stderr,"ERROR: tn exceeds dimension of xus");
                    		   divarc(xki, xji, tl2, tess-tl,  x1);
                    		   divarc(xkj, xij, tl,  tess-tl2, x2);
                    		   divarc(xjk, xik, tl,  tl+tl2,   x3);
                    		   x = x1[1]+x2[1]+x3[1]; 
      			   y = x1[2]+x2[2]+x3[2]; 
      			   z = x1[3]+x2[3]+x3[3];
                    		   d = sqrt(x*x+y*y+z*z);
                    		   tn++;
                    		   (*surf).xus[tn][1] = x/d;
                    		   (*surf).xus[tn][2] = y/d;
                    		   (*surf).xus[tn][3] = z/d;
                    		  }	/* cycle tl2 */
                  	     }		/* cycle tl  */
                	 }		/* cycle k   */
                   }			/* cycle j   */
            	}			/* cycle i   */
          if ((*surf).n_dot != tn) 
	    fprintf(stderr,"ERROR: ico_dot: n_dot(%d) and tn(%d) differ", (*surf).n_dot, tn);
         }		
      
      		/* end of if (tess > 1) */
}
/* .................................................................... */
/* tessellation class II						*/
/* 									*/
/* densit...required dots per unit sphere 		     		*/
/* dot distribution on a unit sphere based on an icosaeder   		*/
/* great circle average refining of icosahedral face         		*/
/* 									*/
void ico_dot_dod(int densit, Surf *surf) 
{
int i, j, k, tl, tl2, tn, tess, j1, j2;
double rh;
double a, d, x, y, z, ai_d, adod;
double xki[4], xji[4], x1[4], xkj[4], xij[4], x2[4], xjk[4], xik[4], x3[4];

/* Code */
/* determine the number of dots */
a = sqrt((((double) densit)-2.)/30.);
tess = MAXIMUM((int) ceil(a), 1);
(*surf).n_dot = 30*tess*tess+2;
if ((*surf).n_dot < densit) 
	{fprintf(stderr,"ERROR: ico_dot_dot: error in formula for tessellation level (%d->%d, %d)\n", tess, (*surf).n_dot, densit);
	}

/* allocation of a matrix */
(*surf).xus =dmatrix(1,(*surf).n_dot, 1,3);

/* calculate the first 12 dots */
icosaeder_vertices((*surf).xus);

tn=13; 

/* square of the edge of an icosaeder */
rh = sqrt(1.-2.*cos(TORAD(72.)))/(1.-cos(TORAD(72.)));
a = rh*rh*2.*(1.-cos(TORAD(72.)));
/* dodecaeder vertices */
for(i=1; i<=10; i++) 
   {for(j=i+1; j<=11; j++) 
       {x = (*surf).xus[i][1]-(*surf).xus[j][1];
        y = (*surf).xus[i][2]-(*surf).xus[j][2]; 
	z = (*surf).xus[i][3]-(*surf).xus[j][3];
        d = x*x+y*y+z*z;
        if (fabs(a-d) > DP_TOL) continue; 
        for (k=j+1; k<=12; k++) 
	    {x = (*surf).xus[i][1]-(*surf).xus[k][1];
             y = (*surf).xus[i][2]-(*surf).xus[k][2]; 
	     z = (*surf).xus[i][3]-(*surf).xus[k][3];
             d = x*x+y*y+z*z;
             if (fabs(a-d) > DP_TOL) continue;
             x = (*surf).xus[j][1]-(*surf).xus[k][1];
             y = (*surf).xus[j][2]-(*surf).xus[k][2]; 
	     z = (*surf).xus[j][3]-(*surf).xus[k][3];
             d = x*x+y*y+z*z;
             if (fabs(a-d) > DP_TOL) continue;
             x = (*surf).xus[i][1]+(*surf).xus[j][1]+(*surf).xus[k][1];
             y = (*surf).xus[i][2]+(*surf).xus[j][2]+(*surf).xus[k][2];
             z = (*surf).xus[i][3]+(*surf).xus[j][3]+(*surf).xus[k][3];
             d = sqrt(x*x+y*y+z*z);
             (*surf).xus[tn][1]=x/d; 
	     (*surf).xus[tn][2]=y/d; 
	     (*surf).xus[tn][3]=z/d;
             tn++;
            }
        }
   }
if (tess > 1) 
   {tn = 33; 
    /* square of the edge of an dodecaeder */
    adod = 4.*(cos(TORAD(108.))-cos(TORAD(120.)))/(1.-cos(TORAD(120.)));
    /* square of the distance of two adjacent vertices of ico- and dodecaeder */
    ai_d = 2.*(1.-sqrt(1.-a/3.));
    /* calculate tessellation of mixed edges */
    for(i=1; i<=31; i++) 
       {j1 = 13; j2 = 33; a = ai_d;
        if (i>=13) { j1=i+1; a = adod; }
        for(j=j1; j<=j2; j++) 
	   {x = (*surf).xus[i][1]-(*surf).xus[j][1];
            y = (*surf).xus[i][2]-(*surf).xus[j][2]; 
	    z = (*surf).xus[i][3]-(*surf).xus[j][3];
            d = x*x+y*y+z*z;
            if (fabs(a-d) > DP_TOL) continue;
            for(tl=1; tl<tess; tl++) 
	       {if(tn > (*surf).n_dot) 
		  {ERROR("ico_dot: tn exceeds dimension of xus");}
                divarc((*surf).xus[i], (*surf).xus[j],tl, tess, (*surf).xus[tn]);
                tn++;
               }
           }
       }

    /* calculate tessellation of pentakisdodecahedron faces */
    for(i=1; i<=12; i++) 
       {for(j=13; j<=31; j++) 
	   {x = (*surf).xus[i][1]-(*surf).xus[j][1];
            y = (*surf).xus[i][2]-(*surf).xus[j][2]; 
	    z = (*surf).xus[i][3]-(*surf).xus[j][3];
            d = x*x+y*y+z*z;
            if (fabs(ai_d-d) > DP_TOL) continue;
            for (k=j+1; k<=32; k++) 
		{x = (*surf).xus[i][1]-(*surf).xus[k][1];
                 y = (*surf).xus[i][2]-(*surf).xus[k][2]; 
		 z = (*surf).xus[i][3]-(*surf).xus[k][3];
                 d = x*x+y*y+z*z;
                 if (fabs(ai_d-d) > DP_TOL) continue;
                 x = (*surf).xus[j][1]-(*surf).xus[k][1];
                 y = (*surf).xus[j][2]-(*surf).xus[k][2]; 
		 z = (*surf).xus[j][3]-(*surf).xus[k][3];
                 d = x*x+y*y+z*z;
                 if (fabs(adod-d) > DP_TOL) continue;
                 for (tl=1; tl<tess-1; tl++) 
		     {divarc((*surf).xus[j], (*surf).xus[i], tl, tess, xji);
                      divarc((*surf).xus[k], (*surf).xus[i], tl, tess, xki);
                      for (tl2=1; tl2<tess-tl; tl2++) 
			  {divarc((*surf).xus[i], (*surf).xus[j], tl2, tess, xij);
                           divarc((*surf).xus[k], (*surf).xus[j], tl2, tess, xkj);
              		   divarc((*surf).xus[i], (*surf).xus[k], tess-tl-tl2, tess, xik);
                	   divarc((*surf).xus[j], (*surf).xus[k], tess-tl-tl2, tess, xjk);
              		   if (tn > (*surf).n_dot) 
			      {ERROR("ico_dot: tn exceeds dimension of xus");
                	      }
              		   divarc(xki, xji, tl2, tess-tl, x1);
              		   divarc(xkj, xij, tl, tess-tl2, x2);
              		   divarc(xjk, xik, tl, tl+tl2,   x3);
              		   x = x1[1]+x2[1]+x3[1]; 
			   y = x1[2]+x2[2]+x3[2]; 
			   z = x1[3]+x2[3]+x3[3];
              		   d = sqrt(x*x+y*y+z*z);
              		   (*surf).xus[tn][1] = x/d;
              		   (*surf).xus[tn][2] = y/d;
              		   (*surf).xus[tn][3] = z/d;
              		   tn++;
              		  }		/* cycle tl2 */
            	     }		/* cycle tl */
                }		/* cycle k */
           }		/* cycle j */
      }			/* cycle i */
   if((*surf).n_dot != (tn-1)) 
     {fprintf(stderr,"WARNING: ico_dot: n_dot(%d) and tn(%d) differ\n", (*surf).n_dot, tn);
     }
  }		/* end of if (tess > 1) */
}		/* end of routine ico_dot_dod */


/* .................................................................... */
/* place points randomly on a unit sphere                               */
/*                                                                      */

void random_tess(int densit, Surf *surf, long *seed) 
{
  double r=1.0; /* radius of the sphere */
  double phi, z, theta;
  int i;
  
  (*surf).n_dot = densit;
  (*surf).xus =dmatrix(1,(*surf).n_dot, 1,3);

  /* now generate random points */
  for (i=1; i<=(*surf).n_dot; i++)
      {z      = r*(2*Ran(seed)-1.0);
       phi    = 2*M_PI*Ran(seed); 
       theta  = asin(z/r);
       (*surf).xus[i][1] = r*cos(theta)*cos(phi);
       (*surf).xus[i][2] = r*cos(theta)*sin(phi);
       (*surf).xus[i][3] = z;
      }


}
