/* G. Matthias Ullmann 						*/
/* Berlin, 04.10.1995						*/
/* Large changes December 2016					*/
/*								*/
/* Program performs a MC mapping procedure			*/
/*								*/
/* ............................................................ */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <zlib.h>
#include "mol.h"
#include "memutil.h"
#include "util.h"
#include "grid.h"
#include "random.h"
#include "properties.h"
#include "mathematik.h"
#include "dock-help.h"
#include "exclude.h"
#include "surface.h"

/********************************************************/


/* ............................................................ */
/* Output of the run parameters					*/
/*								*/
void fprint_para(FILE *fpout, Run_para *para)
{
    
    fprintf(fpout,SEPARATE);
    fprintf(fpout,"Run parameters of MCMap\n");
    fprintf(fpout,"\t Number of runs:                     %d\n", (*para).nruns);
    fprintf(fpout,"\t Max. Number of Unsuccessful Steps:  %d\n", (*para).max_unsuccess);
    fprintf(fpout,"\t Seed:                               %ld\n", (*para).seed);			 
    fprintf(fpout,"\t Init. separation dist.:             %8.3f\n", (*para).rad_in);
    fprintf(fpout,"\t Max. separation dist.:              %8.3f\n",(*para).rad_out);
    fprintf(fpout,"\t separation parameter:               %8.3f\n",(*para).separation);
    fprintf(fpout,"\t %8s \t %7s\t %9s\t %9s\n", "Temp.","Steps","Rot-Dev","Trans-Dev");
    fprintf(fpout,"\t %8.1f \t %7ld \t %9.1f \t %9.1f\n", (*para).temp,(*para).steps,
	                                                  TOGRD((*para).dev_rot),(*para).dev_trans);
    fprintf(fpout,SEPARATE);
}
/* .....................................................*/
/* Setup - Read in the input file			*/
/*                                                      */
/*                                                      */
void setup(char *file, Run_para *para, 
	   Mol *protein1, Mol *protein2)
{
    time_t tim;
    FILE *fpin;
    char line[LINELENGTH], s1[LINELENGTH], s2[LINELENGTH];
    double hlp;
    
    fpin=file_open(file,"r");

    (*para).nruns      =  1;
    (*para).seed       = -time(&tim);/* Seed for the random number generator         */
    (*para).output     = stdout;     /* Output file                                  */
    (*para).dockdens   = NULL;
    (*para).pot_out    = NULL;       /* outer potential file                         */
    (*para).pot_in     = NULL;       /* inner potential file                         */
    (*para).excl       = NULL;       /* exclusion grid file                          */
    (*para).traj       = NULL;	     /* trajectory file			             */
    (*para).structene  = NULL;	     /* low energy structures file	      	     */
    (*para).structdist = NULL;	     /* low distance structures file		     */
    (*para).rad_in     = 0.0;        /* initial cm separation distance               */
    (*para).rad_out    = 0.0;        /* maxial cm separation distance                */
                                     /* maximum distance for deviation calculation   */
    (*para).separation = 0.0;        /* distance at which the testing                */
                                     /* of the exclusion grid is started             */
                                     /* = minimum distance for deviation calculation */
    (*para).istuck        = 100;
    (*para).max_unsuccess = 1000000;
    (*para).low_energy    = -DBL_MAX;
    (*para).low_struct    = NULL;
    (*para).nprint        = -1; /* for writing trajectories */
    (*para).nprintdist    = -1; /* for writing dist-files  */
    (*para).nprintdens    = -1; /* for writing dens-files  */
    (*para).istuck        = -1; /* for writing restart after stucking  */

    
    (*para).steps           = 10000;
    (*para).temp            = 300.0;
    (*para).dev_rot         = TORAD(5.0);
    (*para).dev_trans       = 3.0;
    (*para).start           = 1; /* where to start the simulation: 0 -- use the input coordinates 
				                                   1 -- start on the inner sphere
				 */

    /* read file */
    while (fgets(line,LINELENGTH,fpin)!=NULL)
        {if (line[0]=='!' || line[0]=='#') /*comment line */
            s1[0]='!'; /* just to do something */
         else if (0>=sscanf(line, "%s %s",s1, s2))
	    {muwarn("Empty line in ReadSetup()");
	    }
         else if (2!=sscanf(line, "%s %s",s1, s2))
	    {fprintf(stderr,SEPARATE);
             fprintf(stderr,"%s",line);
             muerror("Problem in ReadSetup() - parameter has no value");
	    }
         else if (0==strcmp(s1,"start"))
            (*para).start=atoi(s2);
         else if (0==strcmp(s1,"number_of_runs"))
            (*para).nruns=atoi(s2);
         else if (0==strcmp(s1,"nprint"))
            (*para).nprint=atoi(s2);
         else if (0==strcmp(s1,"nprintdist"))
            (*para).nprintdist=atoi(s2);
         else if (0==strcmp(s1,"nprintdens"))
            (*para).nprintdens=atoi(s2);
         else if (0==strcmp(s1,"istuck"))
            (*para).istuck=atoi(s2);
         else if (0==strcmp(s1,"rad_in"))
            (*para).rad_in=atof(s2);
         else if (0==strcmp(s1,"rad_out"))
            (*para).rad_out=atof(s2);
         else if (0==strcmp(s1,"temperature"))
            (*para).temp=atof(s2);
         else if (0==strcmp(s1,"mcsteps"))
            (*para).steps=atoi(s2);
         else if (0==strcmp(s1,"dev_rot"))
            (*para).dev_rot=TORAD(atof(s2));
         else if (0==strcmp(s1,"dev_trans"))
            (*para).dev_trans=atof(s2);
         else if (0==strcmp(s1,"excl"))
            (*para).excl=file_open(s2,"r");
         else if (0==strcmp(s1,"pot_in"))
             {(*para).pot_in=file_open(s2,"r");}
         else if (0==strcmp(s1,"pot_out"))
             {(*para).pot_out=file_open(s2,"r");}
         else if (0==strcmp(s1,"cmdist"))
            (*para).cmdist=atof(s2);
         else if (0==strcmp(s1,"lowenergy"))
            (*para).low_energy=atof(s2);
         else if (0==strcmp(s1,"structene"))
            (*para).structene = gzopen(s2,"w");
         else if (0==strcmp(s1,"structdist"))
            (*para).structdist= gzopen(s2,"w");
         else if (0==strcmp(s1,"traj"))
            (*para).traj= gzopen(s2,"w");
         else if (0==strcmp(s1,"dockdensity"))
            (*para).dockdens=file_open(s2,"w");
         else if (0==strcmp(s1,"separation"))
	    (*para).separation=atof(s2);

         else if ((0==strcmp(s1,"rezeptor"))|| (0==strcmp(s1,"receptor")))
             {fprintf(stdout,SEPARATE);
	      fprintf(stdout,"Receptor: %s\n", s2);
	      GetProtein((*para).output, s2, protein1);
              fprintf(stdout,SEPARATE);
             }
         else if (0==strcmp(s1,"ligand"))
             {fprintf(stdout,SEPARATE);
	      fprintf(stdout,"Ligand: %s\n", s2);
	      GetProtein((*para).output, s2, protein2);
              fprintf(stdout,SEPARATE);
             }
	 else {fprintf(stdout,SEPARATE);
	       fprintf(stderr,"%s",line);
               muerror("Do not understand line in setup file");
	      }
    
        }

    /* check if parameter are ok and overwrite them if required */
    hlp = (*protein1).max_cm_dist+(*protein2).max_cm_dist+4.0;
    if (hlp>(*para).separation)
      {muwarn("Change separation parameter to a more appropriate value");
       fprintf(stderr,"\t separation parameter was    %8.3f\n",(*para).separation);
       (*para).separation=hlp;
       fprintf(stderr,"\t separation parameter is now %8.3f\n",(*para).separation);
      }

    /* --- Parameter output ---- */
    fprint_para((*para).output,para);


}
/* .....................................................*/
/* do some tests                                        */
/*                                                      */
void SomeTests(Run_para *para, Mol *p1, Mol *p2, dGrid *phi_in,
	       dGrid *phi_out)
{
    double d1, d2;
  
    /* test if small grid fits */
    d1 = (*para).separation+(*p2).max_cm_dist;
    d2 = (*phi_in).spacing*0.5*((*phi_in).imax[1]-(*phi_in).imin[1]);
    if (d1>d2)
      {fprintf(stderr,SEPARATE);
       fprintf(stderr,"Problem: d1: %f    d2: %f\n",d1,d2);
       fprintf(stderr,"The inner potential should range at least from %f %f\n\n\n",
	       -d1,d1);
       muerror("inner potential grid too small");
      }

    /* test if outer grid fits */
    d1 = (*para).rad_out+(*p2).max_cm_dist;
    d2 = (*phi_out).spacing*0.5*((*phi_out).imax[1]-(*phi_out).imin[1]);
    if (d1>d2)
      {fprintf(stderr,SEPARATE);
       fprintf(stderr,"d1: %f    d2: %f\n",d1,d2);
       fprintf(stderr,"The out potential should range at least from %f %f\n\n\n",
	       -d1,-d1);
       muerror("outer potential grid too small");
      }

    /* test if the outer potential is cubic */
    if (((*phi_out).imax[1]!=-(*phi_out).imin[1]) ||
        ((*phi_out).imax[2]!=-(*phi_out).imin[2]) ||
        ((*phi_out).imax[3]!=-(*phi_out).imin[3]) ||
        ((*phi_out).imax[1]!= (*phi_out).imax[2]) ||
        ((*phi_out).imax[1]!= (*phi_out).imax[3]))
      muerror("outer potential grid not cubic");

    /* test if inner potential is cubic */
    if (((*phi_in).imax[1]!=-(*phi_in).imin[1]) ||
        ((*phi_in).imax[2]!=-(*phi_in).imin[2]) ||
        ((*phi_in).imax[3]!=-(*phi_in).imin[3]) ||
        ((*phi_in).imax[1]!= (*phi_in).imax[2]) ||
        ((*phi_in).imax[1]!= (*phi_in).imax[3]))
      muerror("inner potential grid not cubic");

    /* test if radii are in right order */
    if (((*para).rad_in-(*para).rad_out)>=0.0)
        {fprintf(stderr,"(*para).rad_in  (*para).rad_out: %f %f \n",
                 (*para).rad_in, (*para).rad_out);
         muerror("Radii in wrong order ---  b>=c");
        }
}

/* .....................................................*/
void InitDens(iGrid *dens, double radius)
{
    int i,j,k;

    (*dens).spacing = 1.0; //CHANGE

    (*dens).inv_spacing=1.0/(*dens).spacing;
    (*dens).imax[1]=(int)ceil((radius+(*dens).spacing)*(*dens).inv_spacing);
    (*dens).imax[2]=(int)ceil((radius+(*dens).spacing)*(*dens).inv_spacing);
    (*dens).imax[3]=(int)ceil((radius+(*dens).spacing)*(*dens).inv_spacing);
            
    (*dens).imin[1]=(int)floor(-(radius+(*dens).spacing)*(*dens).inv_spacing);
    (*dens).imin[2]=(int)floor(-(radius+(*dens).spacing)*(*dens).inv_spacing);
    (*dens).imin[3]=(int)floor(-(radius+(*dens).spacing)*(*dens).inv_spacing);
 
    /* make the grid spacing to odd integer,
       if they are not odd integer */
    odding_down((*dens).imin,3);
    odding_up((*dens).imax,3);

    (*dens).grid=i3tensor((*dens).imin[1], (*dens).imax[1],
                          (*dens).imin[2], (*dens).imax[2],
                          (*dens).imin[3], (*dens).imax[3]);

    /* initializing */
    for (i=(*dens).imin[1];i<=(*dens).imax[1]; i++)
        for (j=(*dens).imin[2];j<=(*dens).imax[2]; j++)
            for (k=(*dens).imin[3];k<=(*dens).imax[3]; k++)
                (*dens).grid[i][j][k]=0;


  
}
/**********************************************************/
int main(int argc, char *argv[])
{
    /* Varialbles */
    Mol  protein1, protein2;
    Run_para para;
    cGrid excl;
    OpenDX dx_in, dx_out, dx;
    dGrid phi_in, phi_out;
    iGrid dens;
    int n;
    char info[LINELENGTH];


    /* Usage */
    usage("usage: mcmap [input file] \n\n", argc, 2);
    version(stdout, "MCMap", "1.0", "Johannes M. Foerster, Ina Poehner, G. Matthias Ullmann");

    /* Code */
    /* read in parameters, coordinates and precalculated stuff*/
    setup(argv[1], &para, &protein1, &protein2);

    /* read the exclusion grid */
    fread_excl_binary(para.excl,&excl);	    
    fclose(para.excl);	   
    cGrid_info(para.output, &excl, "Exclusion Grid");

    /* read the potential grids */
    ReadDX(&dx_out,para.pot_out);
    DX2Grid(&phi_out, &dx_out);
    fclose(para.pot_out);
    dGrid_info(stdout, &phi_out, "Outer Electrostatic Potential Grid");

    ReadDX(&dx_in,para.pot_in);
    DX2Grid(&phi_in, &dx_in);
    fclose(para.pot_in);
    dGrid_info(stdout, &phi_in, "Inner Electrostatic Potential Grid");

    /* test if small grid fits */
    SomeTests(&para, &protein1, &protein2, &phi_in, &phi_out);
    if ((para.separation+protein2.max_cm_dist)>(phi_in.spacing*0.5*(phi_in.imax[1]-phi_in.imin[1])))
      muerror("inner grid too small");

    /* preparing the density grid */
    InitDens(&dens, para.cmdist);
    
    /* MC run */
    for (n=1; n<=para.nruns; n++)
	{fprintf(para.output, RUN,n);
	 mc_run(&phi_out, &phi_in, &excl, &protein2, &para, &dens);
	}

    /* generate the OpendDx file */
    iGrd2DX(&dx, &dens);
    sprintf(info,"Density Plot: lig");
    WriteDX(para.dockdens, &dx, info);

    if (para.dockdens!=NULL)   fclose(para.dockdens);
    if (para.structdist!=NULL) gzclose(para.structdist);
    if (para.traj!=NULL)       gzclose(para.traj);
    if (para.structene!=NULL)  gzclose(para.structene);
    
    return 0;
}

