/* Matthias Ullmann 					*/
/* Bayreuth, April 2010                                 */
/*							*/
/* Reads in a file with many orientations, generates    */
/* the structures, and calcultes the NMR parameters     */
/* .....................................................*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mathematik.h"
#include "mol.h"
#include "my-kabsch.h"
#include "properties.h"
#include "util.h"
#include "memutil.h"
#include "zlib.h"
#include "grid.h"
#include "exclude.h"
#include "dock-help.h"

typedef struct {int mode;
                gzFile orient;
                Mol rec;   
                Mol lig;
                int nprint;
                double dist;
                double space;
                cGrid incl;
                int blab;
               }CmdPara;

/*======================================================*/
#define USAGE \
"usage: print-coord -m [mode] -o [orient] -r [rec.pqrm] -l [lig.pqrm] \n\
                    -n [nprint] -d [dist] -s [spacing] -inc [inclusion.grid]\n\n\
orientation file should be a gzipped-file\n\n\
mode: lig        - ligand coordinates are printed to the dir Struct\n\
      rec        - receptor coordinates are printed to the dir Struct\n\
      cm-lig     - ligand center of mass is printed to stdout\n\
      cm-rec     - receptor center of mass is printed to stdout\n\
      dx-lig     - density of the center of mass of the \n\
                   ligand is printed to an OpendDX-File (stdout)\n\
      dx-rec     - density of the center of mass of the \n\
                   receptor is printed to an OpendDX-File (stdout)\n\
      min-dist   - calculate the minimal distance between the receptor and\n\
	           the ligand, a file containing the minimal distance and \n\
                   the energy is written (mindist-ene.out) and a opendx-file \n\
                   with the minimal approach density is written. \n\
      contact    - writing a contact map histogram of residues, \n\
                   the distance for the contact is given through the variable d\n\
      aa-contact - writing a contact map histogram of amino acid residues, \n\
                   the distance for the contact is given through the variable d\n\
      ens-surf   - calculates the ecnounters, which are in a certain distance \n\
                   to the surface of the receptor  modified encounters are printed \n\
		   to an OpendDX-File (stdout) and a Orientation file (mod_structures.gz)\n\
"
/*======================================================*/

/* .....................................................*/
void InterpretCommand(CmdPara *p, int argc, char *argv[])
{
    int i, k;
    Mol *h;
    FILE *fp;

    (*p).orient = NULL;
    (*p).mode   = 0;
    (*p).nprint = 1;
    (*p).dist   = -100.0;
    (*p).space  = 1.0;
    (*p).blab   = 1;
    h= &(*p).lig; (*h).natoms = 0;
    h= &(*p).rec; (*h).natoms = 0;

    for (i=1; i<argc; i=i+2)
        {if (0==strcmp(argv[i],"-m"))
            {if (argc<i+1) muerror("missing argument for -m");
	     if      (0==strcmp(argv[i+1],"lig"))      (*p).mode = 1;
             else if (0==strcmp(argv[i+1],"rec"))      (*p).mode = 2;
             else if (0==strcmp(argv[i+1],"cm-lig"))   (*p).mode = 3;
             else if (0==strcmp(argv[i+1],"cm-rec"))   (*p).mode = 4;
             else if (0==strcmp(argv[i+1],"dx-lig"))   (*p).mode = 5;
             else if (0==strcmp(argv[i+1],"dx-rec"))   (*p).mode = 6;
             else if (0==strcmp(argv[i+1],"min-dist")) (*p).mode = 7;
             else if (0==strcmp(argv[i+1],"contact"))  (*p).mode = 8;
             else if (0==strcmp(argv[i+1],"ens-surf")) (*p).mode = 9;
             else if (0==strcmp(argv[i+1],"aa-contact")) (*p).mode = 10;
             else {fprintf(stderr,"xx %s xx\n",argv[i+1]);
                   muerror("non-recognized mode option");
                  }
            }
         else if (0==strcmp(argv[i],"-o"))
            (*p).orient=gzopen(argv[i+1],"r");            
         else if (0==strcmp(argv[i],"-r"))
            {GetProtein(stderr, argv[i+1], &(*p).rec);
	      //read_PQRM(argv[i+1], &(*p).rec);
             h= &(*p).rec;
             (*h).total_mass = center_of_mass((*h).xyz, (*h).mass,
                                   (*h).natoms, (*h).cm);
             save_initial_coordinates(&(*p).rec);
             for (k=1; k<=(*h).natoms; k++) 
                 (*h).bval[k] = 1.0;
            }
         else if (0==strcmp(argv[i],"-l"))
	    {GetProtein(stderr, argv[i+1], &(*p).lig);
	      //read_PQRM(argv[i+1], &(*p).lig);
             h= &(*p).lig;
             (*h).total_mass = center_of_mass((*h).xyz, (*h).mass,
                                   (*h).natoms, (*h).cm);
             save_initial_coordinates(&(*p).lig);
             for (k=1; k<=(*h).natoms; k++) 
                 (*h).bval[k] = 1.0;
            }
         else if (0==strcmp(argv[i],"-n"))
            (*p).nprint=atoi(argv[i+1]);
         else if (0==strcmp(argv[i],"-d"))
            (*p).dist=atof(argv[i+1]);
         else if (0==strcmp(argv[i],"-s"))
            (*p).space=atof(argv[i+1]);
         /* Reading inclusion grid */
         else if (0==strcmp(argv[i],"-inc"))
	    {fp = file_open(argv[i+1],"r");
	     fread_excl_binary(fp,&(*p).incl);
	     fclose(fp);
	     cGrid_info(stderr, &(*p).incl,"Inclusion Grid");
            }
         else {fprintf(stderr,"unrecognized option %d %d  :  %s \n",
                       argc, i, argv[i]);
               muerror("error while reading command line");
              }
        }

    /* check for correctness */
    h= &(*p).rec; 
    if ((((*p).mode==2) || ((*p).mode==4) || ((*p).mode==6) || ((*p).mode==9))
         && (0==(*h).natoms))
       {muerror ("for this mode, a receptor file needs to be given");
       }
    if (((*p).mode==5) || ((*p).mode==6) || ((*p).mode==7) || ((*p).mode==9))
       {if ((*p).dist <=0)
           {fprintf(stderr,"dist<0: %e",(*p).dist);
            fprintf(stderr,"maybe dist was not set");
            fprintf(stderr,"for this mode, the \"-d\" option needs to be set.");
            muerror("Problem while reading ch");
           }
       }
    if ((*p).mode==0) muerror("no mode given");
    h= &(*p).lig; 
    if (0==(*h).natoms)
       {muerror ("no ligand given");
       }
    if ((*p).orient==NULL)
        {muerror ("no oriention file given");
        }
        
}

/* .....................................................*/
void SetupGrid(iGrid *d, CmdPara *p)
{
    int i,j,k;

    (*d).spacing = (*p).space;

    if ((*d).spacing<=0.0)
        muerror("Grid spacing <= zero. excl_to_grid()");
    (*d).inv_spacing=1.0/(*d).spacing;

    (*d).imax[1]=ceil(  ((*p).dist+(*d).spacing)*(*d).inv_spacing);
    (*d).imax[2]=ceil(  ((*p).dist+(*d).spacing)*(*d).inv_spacing);
    (*d).imax[3]=ceil(  ((*p).dist+(*d).spacing)*(*d).inv_spacing);
                          
    (*d).imin[1]=floor(-((*p).dist+(*d).spacing)*(*d).inv_spacing);
    (*d).imin[2]=floor(-((*p).dist+(*d).spacing)*(*d).inv_spacing);
    (*d).imin[3]=floor(-((*p).dist+(*d).spacing)*(*d).inv_spacing);

    /* make the number of grid points to an odd integer,
       if they are not odd integer */
    odding_down((*d).imin,3);
    odding_up((*d).imax,3);

    (*d).grid=i3tensor((*d).imin[1], (*d).imax[1],
                       (*d).imin[2], (*d).imax[2],
                       (*d).imin[3], (*d).imax[3]);

    /* initializing */
    for (i=(*d).imin[1];i<=(*d).imax[1]; i++)
        for (j=(*d).imin[2];j<=(*d).imax[2]; j++)
            for (k=(*d).imin[3];k<=(*d).imax[3]; k++)
                (*d).grid[i][j][k]=0;


}

/*..............................................................*/
/* procedure translates a vector                                */

void TranslateVector(double *xyz, double *trans)
{

    xyz[1]=trans[1]+xyz[1];
    xyz[2]=trans[2]+xyz[2];
    xyz[3]=trans[3]+xyz[3];
}

/*..............................................................*/
/* procedure rotates a vector with respect to the given         */
/* center of rotation.                                          */
/*      1. Translation of the vector to the center of           */
/*         rotation.                                            */
/*      2. Rotation                                             */
/*      3. Back Translation                                     */
/* The rotation matix is given to the program.                  */

void RotateVector(double *xyz,
              double *center_of_rotation, double **rot)
{
    double old1[4], old2[4];

    /* Translation of the molecule to the center of rotation. */
    old1[1]=xyz[1]-center_of_rotation[1];
    old1[2]=xyz[2]-center_of_rotation[2];
    old1[3]=xyz[3]-center_of_rotation[3];
    /* Rotation */
    old2[1]=rot[1][1]*old1[1] + rot[1][2]*old1[2] + rot[1][3]*old1[3];
    old2[2]=rot[2][1]*old1[1] + rot[2][2]*old1[2] + rot[2][3]*old1[3];
    old2[3]=rot[3][1]*old1[1] + rot[3][2]*old1[2] + rot[3][3]*old1[3];
    /* back translation */
    xyz[1]=old2[1]+center_of_rotation[1];
    xyz[2]=old2[2]+center_of_rotation[2];
    xyz[3]=old2[3]+center_of_rotation[3];


}

/* .....................................................*/
/* this routine tests whether the two aminoacids aa_rec */
/* and aa_lig at at least at a distance of dist         */
/*                                                      */

int aatest_contact(Mol *rec, Mol *lig, int aa_rec, int aa_lig, double dist)
{
    int i,j;
    double sqrdist;

    for (i=0; i<(*rec).aa[aa_rec][2]; i++)
        {for (j=0; j<(*lig).aa[aa_lig][2]; j++)
	     {sqrdist=SqrDistance((*rec).xyz[(*rec).aa[aa_rec][1]+i],
			 	  (*lig).xyz[(*lig).aa[aa_lig][1]+j]);
	      if (sqrdist<=dist)  {return 1;} /* contact true */
	     }
	}
    return 0; /* contact false */
}

/**********************************************************/
int main(int argc, char *argv[])
{
    /* Varialbles */
    CmdPara p;
    Mol *lig, *rec;
    double energy, disthelp;
    double center[4], trans[4];
    double **rot_matrix;
    char line[LINELENGTH], fname[LINELENGTH];
    FILE *fp;
    int n, icount=0;
    iGrid dens;
    int i,k,j;
    long count = 0;
    OpenDX dx;
    int **contact=NULL, **aacontact=NULL;
    /* modifications due to mode 9 */
    gzFile structdist=NULL;
   
    if (argc<=2)
       {fprintf(stderr,USAGE);
        exit(10);
       }
    version(stderr, "print-coor", "1.0", "Johannes M. Foerster, Ina Poehner, G. Matthias Ullmann");

    InterpretCommand(&p, argc, argv);

    rot_matrix=dmatrix(1,3,1,3);

    if ((p.mode==5) || (p.mode==6) || (p.mode==7) || (p.mode==9))
        {SetupGrid(&dens, &p);
        }
    
    if (p.mode==9)
       {structdist= gzopen("mod_structures.gz","w");
	if(structdist==NULL) printf("Writing of new orientation file not achieved\n");
       }
    lig = &p.lig;
    rec = &p.rec;
    disthelp= p.dist*p.dist;

    if (p.mode==8) 
       {contact = imatrix(1,(*rec).natoms, 1, (*lig).natoms);
	for (i=1; i<=(*rec).natoms; i++)
	    for (j=1; j<=(*lig).natoms; j++)
	        contact[i][j]=0;
       }
                  
    if (p.mode==10) 
       {aacontact = imatrix(1,(*rec).aminoacids, 1, (*lig).aminoacids);
	for (i=1; i<=(*rec).aminoacids; i++)
	    for (j=1; j<=(*lig).aminoacids; j++)
	        aacontact[i][j]=0;
       }

    icount=0; n=0;
    while (gzgets(p.orient,line,LINELENGTH)!=NULL)
        {n++;
         if (0!=strncmp(line,"ORIENTATION",11)) muerror("file not ok");
         icount++;
         if (icount==p.nprint)
             {sscanf(line,"%*s %lf %lf %lf %lf %lf %lf %lf %lf %lf\
                          %lf %lf %lf %lf %lf %lf %lf\n",
                    &energy,
                    &trans[1],&trans[2],&trans[3], 
                    &center[1],&center[2],&center[3], 
                    &rot_matrix[1][1],&rot_matrix[1][2],&rot_matrix[1][3], 
                    &rot_matrix[2][1],&rot_matrix[2][2],&rot_matrix[2][3], 
                    &rot_matrix[3][1],&rot_matrix[3][2],&rot_matrix[3][3]);
             recover_initial_coordinates(lig);
             /* do the transformation for the ligand */
             rotate((*lig).xyz, (*lig).natoms, center, rot_matrix);
             translation((*lig).xyz, (*lig).natoms, trans);
             /* and if required also for the receptor */
             if ((p.mode==2) || (p.mode==4) || (p.mode==6))
                 {recover_initial_coordinates(rec);
                  /* to save CPU time use only a few atoms for "kabsching" */
                  kabsch(rot_matrix, center, trans, (*lig).xyz,
                         (*lig).save_xyz, 10, (*lig).bval);
                  rotate((*rec).xyz, (*rec).natoms, center, 
			 rot_matrix);
                  translation((*rec).xyz, (*rec).natoms, trans);
                 }
             /* now the output */
             if (p.mode==1)
                 {/* print coordinates of the ligand */
                  sprintf(fname, "Struct/lig.%d.pqr",n);
                  fp = file_open(fname, "w");
                  fprintf(fp,"REMARK Energy: %e kT\n",energy);
                  fprint_PQRM(fp,lig);
                  fclose(fp);
                 }
             else if (p.mode==2)
                 {/* print coordinates of the receptor */
                  sprintf(fname, "Struct/rec.%d.pqr",n);
                  fp = file_open(fname, "w");
                  fprintf(fp,"REMARK Energy: %e kT\n",energy);
                  fprint_PQRM(fp,rec);
                  fclose(fp);
                 }
             else if (p.mode==3)
                 {(*lig).total_mass = center_of_mass((*lig).xyz, (*lig).mass,
                                            (*lig).natoms,
                                            (*lig).cm);
                  fprint_DUMMYatom(stdout, 1, " CM ", "CM ",(*lig).cm, 0.0, 0.0);
                 }
             else if (p.mode==4)
                 {(*rec).total_mass = center_of_mass((*rec).xyz, (*rec).mass,
                                            (*rec).natoms,
                                            (*rec).cm);
		   fprint_DUMMYatom(stdout, 1, " CM ", "CM ", (*rec).cm, 0.0, 0.0);
                 }
             else if (p.mode==5) /* generate ligand dx file */
	         {(*lig).cm[1]=(*lig).save_cm[1];
                  (*lig).cm[2]=(*lig).save_cm[2];
                  (*lig).cm[3]=(*lig).save_cm[3];
                  /* do the transformation */
                  RotateVector((*lig).cm, center, rot_matrix);
                  TranslateVector((*lig).cm, trans);
                  /* determine grid cell and increase counter */
                  i=(int)rint((*lig).cm[1]*dens.inv_spacing);
                  j=(int)rint((*lig).cm[2]*dens.inv_spacing);
                  k=(int)rint((*lig).cm[3]*dens.inv_spacing);
                  if (  (i<dens.imin[1]) || (i>dens.imax[1])
                      ||(j<dens.imin[2]) || (j>dens.imax[2])
                      ||(k<dens.imin[3]) || (k>dens.imax[3]))
		    {if (p.blab>=5) muwarn("CM outside the box");
                      }
                  else {if (dens.grid[i][j][k]<INT_MAX) 
                            dens.grid[i][j][k]++;
                        else muerror("too many structures, counter problem");

                        if (count<LONG_MAX) count++;
                        else muerror("too many structures");
                       }
                 }
             else if (p.mode==6) /* generate receptor dx file */
                 {//CHANGE very slow in this way
                  (*rec).cm[1]=(*rec).save_cm[1];
                  (*rec).cm[2]=(*rec).save_cm[2];
                  (*rec).cm[3]=(*rec).save_cm[3];
                  /* do the transformation */
                  RotateVector((*lig).cm, center, rot_matrix);
                  TranslateVector((*lig).cm, trans);
                  (*rec).total_mass = center_of_mass((*rec).xyz, (*rec).mass,
                                            (*rec).natoms,
                                            (*rec).cm);
                  /* determine grid cell and increase counter */
                  i=(int)rint((*rec).cm[1]*dens.inv_spacing);
                  j=(int)rint((*rec).cm[2]*dens.inv_spacing);
                  k=(int)rint((*rec).cm[3]*dens.inv_spacing);
                  if (  (i<dens.imin[1]) || (i>dens.imax[1])
                      ||(j<dens.imin[2]) || (j>dens.imax[2])
                      ||(k<dens.imin[3]) || (k>dens.imax[3]))
		      {if (p.blab>=5) muwarn("CM outside the box");
                      }
                  else {if (dens.grid[i][j][k]<INT_MAX) 
                            dens.grid[i][j][k]++;
                        else muerror("too many structures, counter problem");

                        if (count<LONG_MAX) count++;
                        else muerror("too many structures");
                       }
                 }
             /*********************************************************************/
             else if (p.mode==7) /* generate minimal approach dx file */
	         {double dist, mindist=DBL_MAX;
		  int  ilig=0; //irec,
		  /* determine the minimal distance */
                  for (i=1; i<=(*rec).natoms; i++)
		      {for (j=1; j<=(*lig).natoms; j++)
			  {dist=Distance((*rec).xyz[i],(*lig).xyz[j]);
                            if (dist<mindist)
                               {mindist=dist;
				ilig=j; //irec=i; 
                               }
			   }
		      }
                  //fprint_DUMMYatomPQR(stdout, 1, " XX ", (*lig).xyz[ilig], 0.0);
		  /* write the result to a file */
                  //fp = file_open("mindist-ene.dat", "a");
                  //fprintf(fp,"%e %e\n", mindist, energy);
                  //fclose(fp);
                  /* determine grid cell and increase counter */
                  i=(int)rint((*lig).xyz[ilig][1]*dens.inv_spacing);
                  j=(int)rint((*lig).xyz[ilig][2]*dens.inv_spacing);
                  k=(int)rint((*lig).xyz[ilig][3]*dens.inv_spacing);
                  if (  (i<dens.imin[1]) || (i>dens.imax[1])
                      ||(j<dens.imin[2]) || (j>dens.imax[2])
                      ||(k<dens.imin[3]) || (k>dens.imax[3]))
		      {if (p.blab>=5) muwarn("Atom outside the box");
                      }
                  else {if (dens.grid[i][j][k]<INT_MAX) 
                            dens.grid[i][j][k]++;
                        else muerror("too many structures, counter problem");

                        if (count<LONG_MAX) count++;
                        else muerror("too many structures");
                       }
                 }
             /*********************************************************************/
             else if (p.mode==8) /* generate a contact map histogram */
	       {double sqrdist;
                for (i=1; i<=(*rec).natoms; i++)
		      {for (j=1; j<=(*lig).natoms; j++)
			   {sqrdist=SqrDistance((*rec).xyz[i],(*lig).xyz[j]);
			     if (sqrdist<=disthelp) 
			        {contact[i][j]++;}
			   }
		      }
	         }
             /*********************************************************************/
             else if (p.mode==9) /* generate ligand dx & orientation file */
	       {if (exclude(&p.incl, (*lig).xyz, (*lig).onsurf, (*lig).n_onsurf)!=0)
		   {(*lig).cm[1]=(*lig).save_cm[1];
		    (*lig).cm[2]=(*lig).save_cm[2];
		    (*lig).cm[3]=(*lig).save_cm[3];
		    /* do the transformation */
		    RotateVector((*lig).cm, center, rot_matrix);
		    TranslateVector((*lig).cm, trans);
		    /* determine grid cell and increase counter */
		    i=(int)rint((*lig).cm[1]*dens.inv_spacing);
		    j=(int)rint((*lig).cm[2]*dens.inv_spacing);
		    k=(int)rint((*lig).cm[3]*dens.inv_spacing);

		    if (  (i<dens.imin[1]) || (i>dens.imax[1])
			||(j<dens.imin[2]) || (j>dens.imax[2])
			||(k<dens.imin[3]) || (k>dens.imax[3]))
			{muwarn("CM outside the box");
			}
		    else {if (dens.grid[i][j][k]<INT_MAX)
			     {dens.grid[i][j][k]++; 
			      gzprintf(structdist,"%s", line);
			     }
			  else muerror("too many structures, counter problem");

			  if (count<LONG_MAX) count++;
			  else muerror("too many structures");
		         }
		   }
	       }
             /*********************************************************************/
             else if (p.mode==10) /* generate a contact map histogram */
	       {for (i=1; i<=(*rec).aminoacids; i++)
		      {for (j=1; j<=(*lig).aminoacids; j++)
			  {if (1==aatest_contact(rec, lig, i, j, disthelp))
			      {aacontact[i][j]++;
			      }
			  }
		      }
	       }
             /*********************************************************************/
             else {muerror("unknown mode");
	          }
             /* reset count */
             icount=0;
            }
        }
    free_dmatrix(rot_matrix,1,3,1,3);

    /* write the grid to an openDX file */
    if ((p.mode==5)||(p.mode==6)||(p.mode==7)||(p.mode==9))
       {iGrd2DX(&dx, &dens);
        if (p.mode==5) sprintf(line,"Density Plot: lig");
        if (p.mode==6) sprintf(line,"Density Plot: rec");
        if (p.mode==7) sprintf(line,"Density Plot: min dist");
        if (p.mode==9) sprintf(line,"Density Plot: encounter to surface");
        WriteDX(stdout,&dx, line);
       }

    if (p.mode==9) 
       {printf("Changed orientation-file saved to mod_structures.gz\n");
        gzclose(structdist);
       }

    /* write a contact map histogram */
    if (p.mode==8)
      {char help1[30], help2[30];
       for (i=1; i<=(*rec).natoms; i++)
	 {Atom2Str(help1, rec, i);
            for (j=1; j<=(*lig).natoms; j++)
	        {Atom2Str(help2, lig, j);
                 fprintf(stdout, "%d  %-20s %d %-20s %d\n", i, help1, j, help2, contact[i][j]);
	        }
	   }
      }

    /* write a aa-contact map histogram */
    if (p.mode==10)
      {char help1[30], help2[30];
       for (i=1; i<=(*rec).aminoacids; i++)
	 {Atom2Str(help1, rec, (*rec).aa[i][1]);
            for (j=1; j<=(*lig).aminoacids; j++)
	        {Atom2Str(help2, lig, (*lig).aa[j][1]);
                 fprintf(stdout, "%d  %-20s %d %-20s %d\n", i, help1, j, help2, aacontact[i][j]);
	        }
	   }
      }

    return 0;

}

