'''

plotHistogram is a module to prepare and plot contact count histograms.

Module to create NumPy arrays from preprocessed input data and plot histograms
with pyplot. Plotting of single standard input data can be done on two different
scales: total counts on standard scale or logarithmic scale. Alternatively, a 
normalization to the absolute total number of captured encouter complexes can be 
done. Subtraction data is always expected to be normalized and will be scaled
accordingly.

plotHistogram is part of PyCoALA.

Copyright (C) Heidelberg, Bayreuth 2017 Ina Poehner, Johannes Foerster

This program is free software: you can redistribute it and/or modify it under the terms of 
the GNU General Public License as published by the Free Software Foundation, either version 3 
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program.  
If not, see <https://www.gnu.org/licenses/>.

'''

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm,SymLogNorm
import matplotlib.colors as colors
import matplotlib.ticker 
from matplotlib import rc
from pylab import *

def createNumPyArray(dataDict, receptorAAList, ligandAAList):
    '''
    Function to collect the receptor and ligand amino acids as well as the contact 
    counts in NumPy arrays. Amino acids with 0 counts were not stored earlier. 
    Counts of '0' are added back in before creating the NumPy array. 
    '''
    receptorAA = extractAminoAcidNumbers(receptorAAList)
    # Append additional value for plotting with pcolormesh: The last data field needs a start and an end value
    receptorAA.append((receptorAA[-1])+1)
    ligandAA = extractAminoAcidNumbers(ligandAAList)
    ligandAA.append((ligandAA[-1])+1)
    
    countData = []
    for ligandTuple in ligandAAList:
        for receptorTuple in receptorAAList:
            if (receptorTuple, ligandTuple) in dataDict.keys():
                countData.append(dataDict.get((receptorTuple, ligandTuple)))
            else: # Count was dropped for being zero
                countData.append(0)
    
    receptorNumbers = convertToListOfNumbers(receptorAA)
    receptorArray = np.array(receptorNumbers)
    ligandNumbers = convertToListOfNumbers(ligandAA)
    ligandArray = np.array(ligandNumbers)
    countArray = np.array(countData)
    countArray = countArray.reshape((len(ligandArray)-1),(len(receptorArray)-1)) 
    return countArray, receptorArray, ligandArray, receptorAA, ligandAA

def convertToListOfNumbers(aminoAcidList):
    '''
    This function provides a list of numbers starting from 1 and reaching a total length similar to the 
    list of amino acids and, if present, hetero groups. For each amino acid or hetero group the number is 
    increased by one and this list of numbers is used for plotting to ensure that no distortion of the 
    plot occurs if the protein amino acids do not start with 1 or if the different residues are not entirely
    listed with subsequent numbers (e.g., because a hetero group following amino acid 294 gets the number 2296
    for easier distinction). Labels are later generated from the original data.
    '''
    sequenceOfNumbers = []
    counter = 1
    
    for entry in aminoAcidList:
        sequenceOfNumbers.append(counter)
        counter += 1
    return sequenceOfNumbers

def extractAminoAcidNumbers(listOfAATuples):
    '''
    Function to return only the amino acid or residue numbers from a tuple consisting of number and 3-letter
    code of the amino acid or residue. This reduced list will be used for label construction for spatial reasons.
    '''
    AAList = []
    for entry in listOfAATuples:
        AAList.append(entry[0])
    return AAList

def prepareTicksLabel(proteinList):
    '''
    Reduce the number of ticks to about 10 intersections and label accordingly using the respective 
    original amino acid/atom/residue numbering.
    '''
    totalTicks = len(proteinList)
    tickfrequency = int(round(totalTicks/10))

    
    labelList = []
    
    for i, entry in enumerate(proteinList):
        if i%tickfrequency == 0:
            labelList.append(entry)
    
    return tickfrequency, labelList

def cbarForm(x, pos):
    '''
    Helper function for plot color bar label formatting when using logarithmic scale
    '''
    a, b = '{:.1e}'.format(x).split('e')
    b = int(b)
    if x != 0:
        return  r'$\mathregular{10^{%d}}$' %(b)
    else:
        return r'$\mathregular{%d}$' %(0)

def plotHistogram(histogramData, receptor, ligand, receptorList, ligandList, outputFile, switchParameters, labelDict):
    '''
    Plot the data as a heatmap. Contact counts must be presented as a NumPy array
    with receptor amino acids in columns and ligand amino acids in rows.
    Lists of receptor amino acids/atoms are used to build column labels and lists of the
    respective ligand amino acids/atoms are used to build row labels.
    '''    
    fig = plt.figure(figsize=(32, 12))
    ax=plt.gca()
    plt.rc('font',**{'family':'sans-serif','sans-serif':['Arial']})

    # Preparation of meshgrid from receptor and ligand number data
    xi, yi = np.linspace(receptor.min(), receptor.max(), len(receptor)), np.linspace(ligand.min(), ligand.max(), len(ligand))
    xi, yi = np.meshgrid(xi, yi)
    
    if switchParameters.get('logscale'): # If logscale was set for standard input
        logthresh = 0
        logPcm = plt.pcolormesh(xi, yi, histogramData, norm=SymLogNorm(10**0, vmin=histogramData.min(), vmax=100000), cmap=cm.jet)

        # Placing tick labels for the logarithmic color bar
        vmin=1
        vmax=100000
        maxLog=int(np.ceil(np.log10(vmax)))
        minLog=int(np.ceil(np.log10(vmin)))
        tickLocationsPlot=([vmin] + [-(10.0**x) for x in range(minLog-1, -logthresh-1, -1)] + [0.0]
                             + [(10.0**x) for x in range(vmin, maxLog)] + [vmax])
    
        # Plot the color bar
        cbar = plt.colorbar(logPcm, ticks=tickLocationsPlot, format=FuncFormatter(cbarForm), fraction=0.0174)
        cbar.set_label('# of counts', fontsize=40,rotation=270, labelpad=40, fontweight='bold')
        plt.title(labelDict.get('title'), fontname="Arial", fontweight="bold", y='1.04', fontsize=40)
    elif switchParameters.get('normalized'):
        # Normalization to absolute maximal value:
        histmax = abs(max(histogramData.min(), histogramData.max(), key=abs))
        invhistmax = int(1/histmax)
        histogramData = histogramData/histmax
                        
        if switchParameters.get('subtraction'): # If subtraction mode use red - blue colors between -1.0 and 1.0
            plt.pcolormesh(xi, yi, histogramData, cmap='seismic_r', vmin=histogramData.min(), vmax=-histogramData.min())
            cbar = plt.colorbar(fraction=0.0174)
            cbar.set_label('Scaled normalized difference', fontsize=40,rotation=270, labelpad=40, fontweight='bold')
        else: # If normalized standard mode use standard color scheme
            plt.pcolormesh(xi, yi, histogramData, cmap=cm.jet, vmin=histogramData.min(), vmax=histogramData.max())
            cbar = plt.colorbar(fraction=0.0174)
            cbar.set_label('Scaled normalized population', fontsize=40,rotation=270, labelpad=40, fontweight='bold')
                    
        # Modify title by scaling factor
        ModTitle = labelDict.get('title') + ' (Scaling factor: ' + str(invhistmax) + 'x)'
        plt.title(ModTitle, fontname="Arial", fontweight="bold", y='1.04', fontsize=40)
    
    else: # Standard input, total count numbers on standard scale
        plt.pcolormesh(xi, yi, histogramData, cmap=cm.jet, vmin=histogramData.min(), vmax=histogramData.max())
        cbar = plt.colorbar(fraction=0.0174)    
        cbar.set_label('# of counts', fontsize=40,rotation=270, labelpad=40, fontweight='bold')
        plt.title(labelDict.get('title'), fontname="Arial", fontweight="bold", y='1.04', fontsize=40)

    cbar.ax.tick_params(labelsize=32)
    cbar.update_ticks()
   
    # Minimal and maximal values of x and y parameters for axis scaling 
    xmin=receptor.min()
    xmax=receptor.max()
    ymin=ligand.min()
    ymax=ligand.max()
    
    plt.xlim([xmin, xmax])
    plt.ylim([ymin, ymax])
    plt.xlabel(labelDict.get('xlabel'), fontsize=40, labelpad=4, fontname="Arial", fontweight='bold' )
    plt.ylabel(labelDict.get('ylabel'), fontsize=40, fontname="Arial", labelpad=8, fontweight='bold')
    
    plt.tick_params(axis='both', which='minor', labelsize=0, direction='out', size=8)
    plt.tick_params(axis='both', which='major', labelsize=32, direction='out', size=16, pad=14)
    
    stepsize, modifiedLabels = prepareTicksLabel(receptorList) # Set about 10 labels using original receptor/ligand data
    plt.xticks(np.arange(xmin+0.5, xmax, stepsize), modifiedLabels, fontsize=32)
    ligstep, modifiedLigLabels = prepareTicksLabel(ligandList)
    plt.yticks(np.arange(ymin+0.5, ymax, ligstep), modifiedLigLabels, fontsize=32)
    
    plt.minorticks_on()
    plt.gca().set_aspect(1)

    plt.savefig(outputFile)
    #plt.show() # Optionally comment in to see plot at runtime
    return