'''

processPrintCoorHistogram is a module to read and preprocess output of print-coor
contact or aa-contact mode.

processPrintCoorHistogram is part of PyCoALA.

Copyright (C) Heidelberg, 2017 Ina Poehner

This program is free software: you can redistribute it and/or modify it under the terms of 
the GNU General Public License as published by the Free Software Foundation, either version 3 
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program.  
If not, see <https://www.gnu.org/licenses/>.

'''

def readInput(inputFile):
    '''
    Function to read in input data. 
    '''
    fobj_input = open(inputFile, 'r') 
    rawData = fobj_input.readlines()
    fobj_input.close()
    return splitAndConvertInput(rawData)

def splitAndConvertInput(readData):
    '''
    Function for initial splitting of the input data. Data is typically in the following format:
        Atom number receptor
        chain (if applicable, or - for no chain):Amino acid number:amino acid:atom type
        Atom number ligand
        chain (if applicable, or - for no chain):Amino acid number:amino acid:atom type
        Contact count
    e.g.
        1  -:34:SER:N           1 -:1012:GLY:N         248
    '''
    # Find out which input data is used: amino acid or atom resolution
    atomPosition = analyzeInputData(readData)
    #atomPosition is None unless the data is at atomic resolution
    
    contactDict = dict()
    receptorSequence = []
    ligandSequence = []
    
    for line in readData:
        rawDataList = line.split()
        if len(rawDataList) != 0: # Ignore empty lines, if any 
            if isNumber(rawDataList[-1]):
                currentCount = int(rawDataList[-1]) # Last entry in list is expected to be the contact count 
            else:
                raise ValueError('Last column is expected to hold contact counts but seems not to be composed of numbers.')
            receptorList = rawDataList[1].split(':') # Expect the receptor atom details to be written in second column
            ligandList = rawDataList[3].split(':') # Expect the ligand atom details to be written in fourth column
        
            receptorTuple = classifyInput(receptorList, atomPosition)
            ligandTuple = classifyInput(ligandList, atomPosition)
            receptorSequence = createSequence(receptorSequence, receptorTuple)
            ligandSequence = createSequence(ligandSequence, ligandTuple)
            contactDict = createContactDict(contactDict, (receptorTuple, ligandTuple), currentCount)
            
    return contactDict, receptorSequence, ligandSequence

def analyzeInputData(listOfData):
    '''
    Function tries to guess whether the input is at atomic resolution or already 
    collapsed to amino acid level. In the latter case, it is expected that all counts
    were collapsed to the N atom of the amino acid and therefore, no other atoms should
    appear.
    '''
    # TRP is the amino acid with the highest heavy atom count: 14. 
    # If no CA is detected in the first 15 lines, assume aa resolution
    for entry in listOfData[0:15]:
        if ':CA' in entry: # Contains CA, so is expected to be atomic detail
            print('C alpha detected. Suspect atom-wise input.')
            splittedDataList = entry.split()
            if 'CA' in splittedDataList[1]: # A CA was detected in the receptor input
                atomicLocation = splittedDataList[1].split(':').index('CA')
            else: # CA was detected in the ligand input
                atomicLocation = splittedDataList[3].split(':').index('CA')
            return atomicLocation
    
    print('Suspect input at amino acid resolution')
    return None # No location for atoms in file needed

def classifyInput(listOfIdentifiers, atomPosition):
    '''
    Function to determine which part of the input data is the amino acid or residue
    number, the corresponding three-letter code, and, if present, the atom.
    Returns a tuple with the corresponding information. If the data is at atomic 
    resolution and contains hydrogen atoms, these and their contact counts are dropped
    in this function.
    '''
    aminoAcid = ''
    aminoAcidNumber = 0
    for descriptor in listOfIdentifiers:
        if isNumber(descriptor):
            if aminoAcidNumber != 0:
                raise Exception('Problem with the input data. More than one number detected.')
            else:
                aminoAcidNumber = int(descriptor)
        elif len(descriptor) == 3: # Amino acids are always given as three letter code
            if atomPosition == None:
                aminoAcid = descriptor
            else:
                if descriptor != listOfIdentifiers[atomPosition]: 
                    # If atoms with length of three could exist as well: Check if it is a three-letter atom code
                    aminoAcid = descriptor
    if atomPosition != None: # In this case, no compression is done.
        atomCode = listOfIdentifiers[atomPosition]
        if not atomCode[0] == 'H': # if the atom is not hydrogen
            return (aminoAcidNumber, aminoAcid, atomCode)
    else:
        return (aminoAcidNumber, aminoAcid)

def isNumber(testString):
    '''
    Function to decide whether a given string is a number by attempting to cast the string 
    to float.
    '''
    try:
        float(testString)
        return True
    except ValueError:
        pass
    return False
    
def createSequence(sequenceList, identifierTuple):
    '''
    Function to generate the amino acid / atom sequence from the prepared input.
    '''
    if identifierTuple not in sequenceList:
        sequenceList.append(identifierTuple)
    return sequenceList

def createContactDict(contactDictionary, contactTuple, contactCount):
    '''
    Function to store the contact counts of atom/amino acid pairs, if the count 
    is not zero.
    '''
    if contactCount != 0: # Only add data to the dictionary if the contact count is non-zero
        if contactTuple not in contactDictionary.keys(): # Contact pair was not found previously
            contactDictionary[contactTuple] = contactCount
        else:
            raise ValueError('Duplication of contact tuple detected for ' + contactTuple)
    return contactDictionary

def normalizeContactDict(originalCountDict, totalEncounter):
    '''
    Function updates the count dictionary to values between 0 and 1 by normalizing
    with the total number of occurring encounter complexes prior to subtraction
    or, if desired, for standard input, when the total number of orientations was provided.
    '''
    normalizedCountDict = dict()
    for contactPair in originalCountDict.keys():
        contactCount = originalCountDict.get(contactPair)
        normalizedCount = float(contactCount)/totalEncounter
        normalizedCountDict[contactPair] = normalizedCount
    return normalizedCountDict

def subtractTwoInputs(firstDict, secondDict, firstReceptor, secondReceptor, firstLigand, secondLigand):
    '''
    Function to subtract two sets of input data. Expects identical sequences for receptor and ligand.
    '''
    combinedDict = dict()
    for aminoAcid in firstReceptor:
        if aminoAcid not in secondReceptor:
            raise Exception('Receptor sequences are not identical!')
        for ligandAcid in firstLigand:
            if ligandAcid not in secondLigand:
                raise Exception('Ligand sequences are not identical!') 
            # check if contact tuple exists in first dictionary
            if (aminoAcid, ligandAcid) in firstDict.keys(): # Get the first count
                firstCount = firstDict.get((aminoAcid, ligandAcid))
            else:
                firstCount = 0 # Missing, so contact count was zero
            if (aminoAcid, ligandAcid) in secondDict.keys():
                secondCount = secondDict.get((aminoAcid, ligandAcid))
            else:
                secondCount = 0
            # Subtract first from second count
            combinedDict[(aminoAcid, ligandAcid)] = firstCount - secondCount
    return combinedDict