'''

Module to read the config file of PyCoALA, do some basic exception handling and check
and collect the chosen parameters.

readConfig is part of PyCoALA.

Copyright (C) Heidelberg, 2017 Ina Poehner

This program is free software: you can redistribute it and/or modify it under the terms of 
the GNU General Public License as published by the Free Software Foundation, either version 3 
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program.  
If not, see <https://www.gnu.org/licenses/>.


'''

import os
import gzip
from processPrintCoorHistogram import isNumber

def readConfigFile(fileName):
    '''
    Function to read the config file and collect all parameters.
    '''
    configFile = open(fileName,"r")
    chosenParameters = configFile.readlines()
    configFile.close()
    return extractRunParameters(chosenParameters)

def removeWhitespaces(inputString):
    '''
    Function to remove whitespaces from config file input.
    '''
    cleanString = ''.join(inputString.split())
    return cleanString

def fileCheck(fileName):
    '''
    Function to check if a specified file exists, otherwise raises exception.
    '''
    if not os.path.exists(fileName):
        raise FileNotFoundError(fileName + ' could not be found.')
    return

def lenientFileCheck(fileName):
    '''
    Function to check if a specified file exists and only displays a warning if not, used if defaults can be used instead.
    '''
    try:
        if os.path.exists(fileName):
            return True
    except FileNotFoundError:
        pass
    return False

def overWriteWarning(currentKey, parameterDict):
    '''
    Function to display a warning in case a variable is specified more than once in the config file. The last entry
    in the config file will be used and a warning will inform the user.
    '''
    if currentKey in parameterDict.keys():
        print('Warning: The following variable was assigned previously: ' + currentKey + '. Previous settings will be overwritten.')
    return

def generateInputList(dataString, parameterDict):
    '''
    Function to generate a list of input files and set the subtraction mode on basis of the number of input
    file(s) given. Assumes comma-separated input, if more than one input file is given.
    '''
    listOfInputFiles = []
    
    dataString = removeWhitespaces(dataString)
    
    if ',' in dataString: # Two input files
        for inFile in dataString.split(','):
            fileCheck(inFile)
            listOfInputFiles.append(inFile)
        if len(listOfInputFiles) != 2: # If not exactly two files are given, raise some error
            raise ValueError('Something wrong with the number of input files. Assume subtraction mode. Expect two input files separated by comma.')
        parameterDict['subtraction'] = True # More than one input, files will be subtracted
    else:
        fileCheck(dataString)
        listOfInputFiles.append(dataString)
        parameterDict['subtraction'] = False # Single input file, no subtraction to be done
    overWriteWarning('input', parameterDict)
    parameterDict['input'] = listOfInputFiles
    return 

def getTotalEncounter(orientations, parameterDict):
    '''
    Function to either count the total number of encounters from given MCMap output (in gzipped format, standard
    structdist.gz) or convert given numbers from strings.
    '''
    encounterCounts = []
    orientationData = []
    
    orientations = removeWhitespaces(orientations)

    if ',' in orientations:
        orientationData = orientations.split(',')
        if len(orientationData) > 2: 
            raise ValueError('Too many orientations specified. A maximum number of two files or orientations can be given.')
    else:
        orientationData.append(orientations)
        
    for entry in orientationData:
        if isNumber(entry): # Function to decide whether string can be casted to float
            encounterCounts.append(int(entry))
        else:
            fileCheck(entry)
            with gzip.open(entry,'r') as orientFile:
                for lineCount, lineContent in enumerate(orientFile, 1):
                    pass
                encounterCounts.append(lineCount)
                print('Found ' + str(lineCount) + ' orientations.')
    overWriteWarning('orientations', parameterDict)
    parameterDict['orientations'] = encounterCounts
    return 

def preparePlottingOutput(plotFile, parameterDict):
    '''
    Function to assign the output file for plotting and add a file extention, if required. Default output
    are PNG files.
    '''
    plotFile = removeWhitespaces(plotFile)
    
    if '.' in plotFile: # Assume a given file extension
        if plotFile.split('.')[-1] != 'png': # If the given file does not already end with png
            print('Default plotting output is in PNG format. It seems like no file type or a different extension was given.')
            plotFile = plotFile + '.png'
    else: # no file type extension at all
        print('Default plotting output is in PNG format. It seems like no file type was given.')
        plotFile = plotFile + '.png'
    overWriteWarning('plot', parameterDict)
    parameterDict['plot'] = plotFile
    return 

def constructLabelsForPlot(rawLabelText, labelKey, plotLabels):
    '''
    Function to prepare plot title and axis labels as specified in the config file and collect those in a label dictionary.
    Parameters are expected to be given as strings enclosed by "".
    '''
    overWriteWarning(labelKey, plotLabels)
    plotLabels[labelKey] = rawLabelText.lstrip().rstrip().lstrip('"').rstrip('"')
    return

def logScaleSwitch(logScale, parameterDict):
    '''
    Function to assign the switch to decide whether plotting will be done on logarithmic scale or not.
    Parameter is expected to be boolean.
    '''
    logScale = removeWhitespaces(logScale)
    overWriteWarning('logscale', parameterDict)
    
    if logScale.title() == 'True': # capitalize first letter only
        parameterDict['logscale'] = True
    elif logScale.title() == 'False':
        parameterDict['logscale'] = False
    else:
        raise ValueError('Some unknown option was given for the logscale. Please specify only boolean values: True or False.')
    return

def getProteinPQRM(proteinFile, typeOfProtein, parameterDict):
    '''
    Function to get the path to a protein pqrm file and the corresponding protein name. Protein is expected to be
    given in pqrm format.
    '''
    proteinFile = removeWhitespaces(proteinFile)
    
    pathToProtein, proteinName = os.path.split(proteinFile)
    
    if '.' in proteinName: # Assume a given file extension. Remove extension from proteinName and add to path if not pqrm
        if proteinName.split('.')[-1] != 'pqrm': # If the given file does not already end with pqrm
            print('Protein is expected to be in pqrm format but for ' + proteinFile + ' a different type seems to be given!')
        else: # The file type is pqrm as expected, just remove from proteinName
            proteinName = proteinName.split('.')[0] # drop the file extension for the protein name
            if pathToProtein:
                pathToProtein = os.path.join(pathToProtein, (proteinName + '.pqrm'))
            else:
                pathToProtein = proteinName + '.pqrm'
    else: # no file type extension at all
        print('Protein ' + proteinFile + ' is assumed to be a pqrm file (file type not given).')
        if pathToProtein:
            pathToProtein = os.path.join(pathToProtein, (proteinName + '.pqrm'))
        else:
            pathToProtein = proteinName + '.pqrm'
            
    if not lenientFileCheck(pathToProtein):
        print('The protein file ' + pathToProtein + ' could not be found. PyMOL script may not be operational or needs to be copied to the same location as the protein file.')
    
    overWriteWarning(typeOfProtein, parameterDict)
    parameterDict[typeOfProtein] = [pathToProtein, proteinName]
    return
    
def displayPercentage(percentageString, parameterDict):
    '''
    Function to evaluate whether the given percentage for displaying is a number and saving the number. 
    Percentage values are typically expected to be given as int or float.
    '''
    percentageString = removeWhitespaces(percentageString)
    
    overWriteWarning('display', parameterDict)

    if isNumber(percentageString):
        parameterDict['display'] = float(percentageString)
    else:
        raise ValueError('The display percentage can only be given as a number. Something is wrong here.')
    return

def topContactReport(reportFile, parameterDict):
    '''
    Function to collect the file name for the text report of top contacting pairs.
    '''
    overWriteWarning('maxcontacts', parameterDict)
    parameterDict['maxcontacts'] = removeWhitespaces(reportFile)
    return

def getProteinPML(proteinFile, typeOfProtein, parameterDict):
    '''
    Function to collect file names for the PyMOL scripts written as output. Default extension '.pml' is
    expected and appended to the given output, if required.
    '''
    proteinFile = removeWhitespaces(proteinFile)
    
    if '.' in proteinFile: # Assume a given file extension
        if proteinFile.split('.')[-1] != 'pml': # If the given file does not already end with pml
            print('Default output for PyMOL script is a pml format. It seems like no file type or a different extension was given.')
            proteinFile = proteinFile + '.pml'
    else: # no file type extension at all
        print('Default output for PyMOL script is pml format. It seems like no file type was given.')
        proteinFile = proteinFile + '.pml'
    
    overWriteWarning(typeOfProtein, parameterDict)
    parameterDict[typeOfProtein] = proteinFile
    return

def extractRunParameters(parameterInput):
    '''
    Function to call various specialized functions for extracting and preparing raw input data from the config file
    and performing some exception handling.
    '''
    # Dictionary with calls to the specific processing functions for config file entries
    configDict = {'input': generateInputList, 'orientations': getTotalEncounter, 'plot': preparePlottingOutput, 
                  'title': constructLabelsForPlot, 'xlabel': constructLabelsForPlot, 'ylabel': constructLabelsForPlot, 
                  'logscale': logScaleSwitch, 'receptor': getProteinPQRM, 'ligand': getProteinPQRM, 'display': displayPercentage, 
                  'maxcontacts': topContactReport, 'receptorpml': getProteinPML, 'ligandpml': getProteinPML}
    
    plotLabels = dict()
    parameterDict = dict()
    
    for entry in parameterInput:
        inLabel = False # Flag to mark # symbols being found in labels which should not be treated as comments
        if '#' in entry and removeWhitespaces(entry)[0] != '#': # The line contains some comment but is not fully commented out
            if '"' in entry: # Symbol for comment can be within a label text
                startOfLabel = entry.index('"') # identify position where label text starts
                modifiedEntry = entry[startOfLabel + 1:] #remove the first "
                endOfLabel = modifiedEntry.index('"') # get position where label text ends
                modifiedEntry = modifiedEntry[:endOfLabel]
                if '#' in modifiedEntry:
                    inLabel = True # Set this flag to process line despite of # symbol
                else:
                    entry = entry[:entry.index('#')] # comment starts after label, remove comment
            elif '"' not in entry:
                entry = entry[:entry.index('#')]

        if '#' not in entry or inLabel: # Line is not a comment line in the config file or # is part of a label text
            if len(entry) > 1: # Line was not empty
                activeParameter = entry.split('=')
                if len(removeWhitespaces(activeParameter[1])) >= 1: # There is at least something written for the variable
                    if removeWhitespaces(activeParameter[0].lower()) in configDict.keys():
                        if removeWhitespaces(activeParameter[0].lower()) in ['title', 'xlabel', 'ylabel']: #List of label types for plot
                            configDict.get(removeWhitespaces(activeParameter[0].lower()))(activeParameter[1], removeWhitespaces(activeParameter[0].lower()), plotLabels)
                        elif removeWhitespaces(activeParameter[0].lower()) in ['receptor', 'ligand', 'receptorpml', 'ligandpml']:
                            configDict.get(removeWhitespaces(activeParameter[0].lower()))(activeParameter[1], removeWhitespaces(activeParameter[0].lower()), parameterDict)
                        else:
                            configDict.get(removeWhitespaces(activeParameter[0].lower()))(activeParameter[1], parameterDict)
                    else:
                        print('Some unknown parameter was used in the config file: ' + removeWhitespaces(activeParameter[0]))
    
    parameterDict['labels'] = plotLabels
    
    return checkInputAndAssignDefaults(parameterDict)
    
def checkInputAndAssignDefaults(parameterDict): 
    '''
    Function to test if all required parameters were given, assign default values where possible if parameters are 
    missing and do a few additional checks.
    '''
    # Dictionary with default plot labels
    defaultLabels = {'title': 'This is my plot title', 'xlabel': 'This is my x-axis label', 'ylabel': 'This is my y-axis label'}
    # Parameter dictionary filled with defaults
    defaultParameters = {'input': None, 'subtraction': False, 'orientations': None, 'plot': 'countHistogramPlot.png',
                     'logscale': False, 'receptor': ['dummyreceptor.pqrm', 'dummyreceptor'], 'ligand': ['dummyligand.pqrm', 'dummyligand'], 
                     'display': 0.05, 'maxcontacts': 'topContactPairs.txt', 'receptorpml': 'dummyreceptor.pml', 
                     'ligandpml': 'dummyligand.pml'}
    
    for defaultSetting in defaultParameters.keys():
        if defaultSetting not in parameterDict.keys(): # The respective parameter was not collected from the config file
            if defaultSetting == 'input' or defaultSetting == 'subtraction':
                raise FileNotFoundError('No input file(s) found in the config file. Aborting.')
            elif defaultSetting == 'orientations': # Orientations can only be omitted for single input files, then, no normalization is done
                if 'input' in parameterDict.keys() and parameterDict.get('subtraction') == True: # Orientations required
                    raise ValueError('No total number of orientations provided for subtraction mode. Orientations are required for normalization. Aborting.')
                elif 'input' in parameterDict.keys() and parameterDict.get('subtraction') == False: # Single input present, no normalization to be done.
                    print('No total number of orientations given. No normalization will be performed.')
                    parameterDict[defaultSetting] = defaultParameters.get(defaultSetting) # Assign None to orientations
            else:
                print('Attention: Parameter ' + defaultSetting + ' was not specified.')
                print('Setting default value ' + str(defaultParameters.get(defaultSetting)) + '!')
                parameterDict[defaultSetting] = defaultParameters.get(defaultSetting)
    
    if parameterDict.get('subtraction'): # If subtraction mode is used, make sure two orientations are present
        if len(parameterDict.get('orientations')) != 2:
            raise ValueError('Wrong number of orientations given for subtraction mode. Please provide the number of encounters for each input data set.')
        
    if parameterDict.get('logscale'): #If logscale set to True, check if subtraction mode is used. If so, reset to False
        if parameterDict.get('subtraction'): # Subtraction is used, so no logscale will be used for plotting
            print('In subtraction mode, the default plotting scale will be used instead of logarithmic scaling.')
            parameterDict['logscale'] = False
        else: # If no subtraction used, check if orientations are given and hence normalization will be done. 
            if parameterDict.get('orientations') is not None:
                print('When using normalization, the default plotting scale will be used instead of logarithmic scaling.')
                parameterDict['logscale'] = False
    
    currentLabelDict = parameterDict.get('labels')
    for defaultLabelEntry in defaultLabels.keys():
        if defaultLabelEntry not in currentLabelDict.keys():
            print('No label for plot ' + defaultLabelEntry +  ' was provided.')
            print('Setting to default label: ' + defaultLabels.get(defaultLabelEntry))
            currentLabelDict[defaultLabelEntry] = defaultLabels.get(defaultLabelEntry)
            
    parameterDict['labels'] = currentLabelDict
    
    return parameterDict          
